

const TYPE_ANY = "*/*";

const CLASSID = Components.ID("{8667f289-ed59-4aaf-9111-b7aa88468cb5}");
const CLASSNAME = "DAISY Book Helper";
const CONTRACTID = "@benetech.org/daisy-helper";

Console = {
   preferencesService: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("daisy.helper."),
   _initialized: false,
   _debug: false,
   service: Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
   log: function(message) {
      this.service.logStringMessage(message);
   },
   debug: function(message) {
      if (!this._initialized) {
         if (Console.preferencesService) {
            try {
               //Console.log(Console.preferencesService.prefHasUserValue("debug"));
               Console._debug = Console.preferencesService.getBoolPref("debug");
            } catch (ex) {
               // no preference
            }
         }
         this._initialized = true;
      }
      if (this._debug) {
         this.service.logStringMessage(message);
      }
   }
}
function findYPos(obj) {
   var curtop = 0;
   if (obj.offsetParent) {
      while (obj.offsetParent) {
         curtop += obj.offsetTop;
         obj = obj.offsetParent;
      }
   } else if (obj.y) {
      curtop += obj.y;
   }
   return curtop;
}

function findPos(obj) {
   var curleft = 0;
   var curtop = 0;
   if (obj.offsetParent) {
      while (obj.offsetParent) {
         curleft += obj.offsetLeft;
         curtop += obj.offsetTop;
         obj = obj.offsetParent;
      }
   }
   return [curleft,curtop];
}


function setTimeout(handler,delay) {
    var timer = Components.classes["@mozilla.org/timer;1"].createInstance().QueryInterface(Components.interfaces.nsITimer);
    timer.initWithCallback(handler,delay,0);
    return timer;
}

function clearTimeout(timer) {
   timer.cancel();
}

function DAISYHelper() {
   this.wrappedJSObject = this;
   var theClass = Components.classes["@benetech.org/daisy;1"];
   var theComponent = theClass.getService(Components.interfaces.nsISupports);
   this.service = theComponent.wrappedJSObject;
   this.service.active = true;
}

DAISYHelper.hackClassName = "stupid-hack-for-crappy-software";
DAISYHelper.testPositionOffset = 100;
DAISYHelper.positionOffset = 80;
DAISYHelper.imagePadding = 10;
DAISYHelper.ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
DAISYHelper.fileURL = function(file) {
   return DAISYHelper.ios.newFileURI(file);
}

DAISYHelper.prototype.forChild = function(e,name,handler) {
   var current = e ? e.firstChild : null;
   while (current) {
      if (current.nodeType!=1) {
         current = current.nextSibling;
         continue;
      }
      //alert("{"+current.namespaceURI+"}"+current.localName+" vs {"+namespace+"}"+name);
      if (current.tagName.toLowerCase()==name) {
         handler(current);
      }
      current = current.nextSibling;
   }
}

DAISYHelper.prototype.init = function(window) {
   this.scrollCounter = 0;
   this.OS = Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULRuntime).OS;
   this.noteManagerId = "daisy-manage-notes-"+(new Date()).getTime();
   this.noteDialogs = [];
   this.window = window.wrappedJSObject;
   this.url = window.location.href;
   var hashPos = this.url.indexOf("#");
   if (hashPos>0) {
      this.hash = this.url.substring(hashPos)
      this.url = this.url.substring(0,hashPos);
   } else {
      this.hash = null;
   }

   this.startLoadingIndicator();

   this.service.currentHelper = this;
   /*
    * get the book from the URI without the hash and reference it.
    */
   Console.debug("Locating book for location "+this.url);
   this.book = this.service.getBook(this.url);
   this.book.ref();

   this.renderer = this.service.renderer;


   var current = this;

   /*
    * the note comparator compares a navigation point's position to
    * the notes position in the document so that a note can inserted
    * into the navigation tree
    */
   this.noteComparator = function(point,note) {
      var pointTarget = current.window.document.getElementById(point.content.id);
      var noteTarget = current.window.document.getElementById("daisy-user-note-ref-"+note.id);
      if (!noteTarget) {
         return 1;
      }
      if (!pointTarget) {
         return -1;
      }
      return pointTarget.offsetTop-noteTarget.offsetTop;
   };
   this._windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
   this.mainWindow = this._windowMediator.getMostRecentWindow("navigator:browser");
   this.initSoundManager();
   var sidebar  = this.mainWindow.document.getElementById("sidebar");
   var tabBrowser = this.mainWindow.getBrowser();

   this.updateNavigationPosition = function() {
      if (sidebar.contentWindow.theSidebar) {
         sidebar.contentWindow.theSidebar.updateNavigationPosition();
      }
   };

   /*
    * the sidebar notification function for attaching the current document to the sidebar
    */
   this.notifySidebar = function() {
      var counter = 0;
      var go = function() {
         if (sidebar.contentWindow.theSidebar && current.context) {
            Console.debug("Notifying sidebar to display TOC.");
            try {
               sidebar.contentWindow.theSidebar.attach(current.url,current.context);
               if (current.context.isRendered) {
                  sidebar.contentWindow.theSidebar.onAfterRender();
               }
            } catch (ex) {
               Console.log(ex);
            }
         } else if (counter<600) {
            counter++;
            setTimeout(go,100);
         } else {
            Console.log("Cannot notify sidebar to display TOC.");
         }
      };
      go();
   }

   /*
    * the document switch notification when the sidebar is already open
    */
   this.notifyDocumentSwitch = function() {
      var counter = 0;
      var go = function() {
         if (sidebar.contentWindow.theSidebar && current.context) {
            Console.debug("Notifying sidebar of document switch");
            try {
               sidebar.contentWindow.theSidebar.switchTo(current.url,current.context);
            } catch (ex) {
               Console.log(ex);
            }
         } else if (counter<20) {
            counter++;
            setTimeout(go,100);
         }
      };
      go();
   }

   this.tab = null;
   for (var i=0; i<tabBrowser.tabContainer.itemCount; i++) {
      var tab = tabBrowser.tabContainer.getItemAtIndex(i);
      if (tab.linkedBrowser.contentWindow==this.window) {
         this.tab = tab;
         break;
      }
   }
   if (this.tab==null) {
      Console.log("ERROR: Defaulting to selected tab.");
      this.tab = tabBrowser.selectedTab;
   }

   /*
    * when the tab is selected, this notifies the sidebar that
    * the document has changed.
    */
   this.tab.addEventListener('TabSelect',function() {
      Console.debug("Selected tab "+current.window.location.href);
      current.mainWindow.toggleSidebar('openDaisySidebar',true);
      current.notifyDocumentSwitch();
      current.service.currentHelper = current;
   },false);

   /*
    * self-voicing of the tab
    */
   this.tab.onfocus = function() {
      if (current.service.readUI) {
         var text = "Tab, DAISY book "+current.book.title;
         current.speak(text,null);
      }
   }

   /**
    * the load listener fires when the current tab is reloaded or
    * a file->open affects the document display (possibly a regular
    * web page and not a DAISY book)
    */
   this.loadListener = function(event) {
      var url = current.tab.linkedBrowser.contentWindow.location.href;
      var hashPos = url.indexOf("#");
      if (hashPos>0) {
         url = url.substring(0,hashPos);
      }
      Console.debug("Tab load: "+url);
      var book = current.service.getBook(url);
      if (!book) {
         current.tab.removeEventListener('load',current.loadListener,false);
         current.mainWindow.toggleSidebar('openDaisySidebar',false);
         var validPanel  = current.mainWindow.document.getElementById("daisy-status-bar-valid");
         var quirkyPanel  = current.mainWindow.document.getElementById("daisy-status-bar-quirky");
         validPanel.style.display = "none";
         quirkyPanel.style.display = "none";
      }
   };
   this.tab.addEventListener('load',this.loadListener,false);

   /*
    * force the sidebar open to our sidebar if it isn't already
    */
   if (sidebar.contentWindow.location.href!="chrome://daisy/content/sidebar.xul") {
      this.mainWindow.toggleSidebar('openDaisySidebar',true);
      this.openedSidebar = true;
   }

   /**
    * window level keyboard handling
    */
   this.window.onkeypress = function(event) {
      //Console.debug("DAISYHelper: "+event.altKey+" "+event.metaKey+" "+event.ctrlKey+" "+event.shiftKey+" "+event.charCode+" "+event.keyCode);
      if (!event.altKey && !event.ctrlKey && !event.metaKey) {
         switch (event.charCode) {
            case 43: //'+'

               if (current.window.focusTimer) {
                  current.window.focusTimer.cancel();
                  current.window.focusTimer = null;
               }
               current.service.tts.cancel();

               setTimeout(function() {
                  current.onNext();
               },250);
               break;
            case 45: //'-'

               if (current.window.focusTimer) {
                  current.window.focusTimer.cancel();
                  current.window.focusTimer = null;
               }
               current.service.tts.cancel();

               setTimeout(function() {
                  current.onPrevious();
               },250);
               break;
            case 112: // 'p'

               if (current.window.focusTimer) {
                  current.window.focusTimer.cancel();
                  current.window.focusTimer = null;
               }
               current.service.tts.cancel();

               setTimeout(function() {
                  current.onPlayCurrent();
               },100);
               break;
         }
         switch (event.keyCode) {
            case 27: //escape
               if (current.bookContext.playing) {
                  current.bookContext.stop();
               } else {
                  current.bookContext.skip({ next: true, play: event.shiftKey, outOfContainer: true});
               }
               break;
         }
      } else if ((event.altKey && event.ctrlKey && event.shiftKey) || (current.OS=="Darwin" && event.metaKey && event.ctrlKey && event.shiftKey)) {
         switch (event.charCode) {
            case 66: // 'B'
               // previous page
               current.bookContext.stop();
               current.onPreviousNote();
               break;
            case 71: // 'G'
               // next page
               current.bookContext.stop();
               current.onNextNote();
               break;
            case 78: // 'N'
               // next page
               current.bookContext.stop();
               current.onNextPage();
               break;
            case 80: // 'P'
               // previous page
               current.bookContext.stop();
               current.onPreviousPage();
               break;
            case 83: // 'S'
               // skip structural
               current.bookContext.skip({next: true, play: true});
               break;
         }
      } else if ((event.altKey && event.ctrlKey) || (current.OS=="Darwin" && event.metaKey && event.ctrlKey)) {
         //Console.log("charCode="+event.charCode);
         switch (event.charCode) {
            case 27: // '[' mac
            case 91: // windows an linux
               current.bookContext.previousLink();
               break;
            case 29: // ']' mac
            case 93: // windows and linux
               current.bookContext.nextLink();
               break;
            case 97: // 'a'
               // jumps to the action bar
               current.onMoveToActionBar();
               break;
            case 104: // 'h'
               // previous word
               current.bookContext.previousWord();
               break;
            case 105: // 'i'
               // insert note
               current.bookContext.stop();
               current.onAddNote();
               break;
            case 106: // 'j'
               // down structural
               current.bookContext.nextStructural();
               break;
            case 107: // 'k'
               // up structural
               current.bookContext.previousStructural();
               break;
            case 108: // 'l'
               // next word
               current.bookContext.nextWord();
               break;
            case 110: // 'n'
               // next navigation item
               current.bookContext.nextNavigation();
               break;
            case 112: // 'p'
               // previous navigation item
               current.bookContext.previousNavigation();
               break;
            case 114: // 'r'
               // navigates to the next note
               if (current.notes.length>0) {
                  current.bookContext.stop();
                  current.bookContext.noteIndex++;
                  if (current.bookContext.noteIndex>=current.notes.length) {
                     current.bookContext.noteIndex = 0;
                  }
                  Console.debug("Moving to note, index="+current.bookContext.noteIndex);
                  var id = "daisy-user-note-ref-"+current.notes[current.bookContext.noteIndex].id;
                  var target = current.window.document.getElementById(id);
                  if (target) {
                     current.bookContext.setElement(target,{ scroll: true, invoke: true });
                  } else {
                     Console.log("Cannot get note rendering for id "+id);
                  }
               }
               break;
            case 115: // 's'
               // skip structural
               current.bookContext.skip({ next: true});
               break;
            case 116: // 't'
               // jumps to toc sidebar
               if (sidebar.contentWindow.theSidebar) {
                  sidebar.contentWindow.theSidebar.focus();
               }
               break;
            case 120: // 'x'
               // skip previous structural
               current.bookContext.skip({ previous: true});
               break;

         }
      } else if (event.altKey && !event.ctrlKey && !event.shiftKey) {
         switch (event.keyCode) {
            case 13:
               Console.debug("alt-enter");
               current.bookContext.currentLocation();
               break;
            case 40:
               Console.debug("alt-down");
               current.bookContext.beneathTableCell();
               break;
            case 39:
               Console.debug("alt-right");
               current.bookContext.nextTableCell();
               break;
            case 38:
               Console.debug("alt-up");
               current.bookContext.aboveTableCell();
               break;
            case 37:
               Console.debug("alt-left");
               current.bookContext.previousTableCell();
               break;
         }
      }
   };

   /*
    * setup the notes menu
    */
   this.notesMenu = this.window.document.getElementById("daisy-internal-notes");
   this.notes = [];
   try {
      this.notes = this.service.database.getNotes(this.url);
   } catch (ex) {
      Console.log(ex);
   }

   this.notesMenu.onchange = function() {
      var option = current.notesMenu.options[current.notesMenu.selectedIndex];
      if (option.value=="id:add-note") {
         current.onAddNote();
      } else if (option.value=="id:manage-notes") {
         current.onManageNotes();
      } else {
         for (var i=0; i<current.renderedNotes.length; i++) {
            if (current.renderedNotes[i].note.id==option.value) {
               current.moveToNote(current.renderedNotes[i]);
               break;
            }
         }
      }
      current.notesMenu.selectedIndex = 0;
   };
   /* track keypresses to voice changes */
   this.notesMenu.onkeypress = function(event) {
      switch (event.keyCode) {
         case 37: // left
         case 38: // up
         case 39: // right
         case 40: // down
            if (current.service.readUI) {
               var option = current.notesMenu.options[current.notesMenu.selectedIndex];
               current.speak("item, "+option.textContent,null);
            }
            break;
         case 13:
            if (current.service.readUI) {
               current.speak("Note selected.",null);
            }
      }
   };
   this.notesMenu.onfocus = function() {
      if (current.service.readUI) {
         current.speak("list box, My notes",null);
      }
   };

   var buttonTTS = function() {
      var text = "button, "+this.getAttribute("title");
      if (current.service.readUI) {
         if (current.context.playContext && this.getAttribute("id")!="daisy-internal-play-button") {
            current.onStop();
         }
         var button = this;
         button.focusTimer = setTimeout(function() {
            button.focusTimer = null;
            current.speak(text,null);
         },250);

      }
   };
   var noteButton = this.window.document.getElementById("daisy-internal-add-note-button");
   noteButton.onclick = function() {
      current.onAddNote();
   };
   noteButton.onfocus = buttonTTS;

   /**
    * setup page number handling in action bar
    */
   var pageValue = this.window.document.getElementById("daisy-internal-page-no");
   pageValue.onkeypress = function(event) {
      if (event.keyCode==13) {
         current.onGotoPage(pageValue.value);
      }
   };
   pageValue.onfocus = function() {
      if (current.service.readUI) {
         current.speak(pageValue.getAttribute("title")+", current value "+pageValue.value,null);
      }
   }
   var pageLink = this.window.document.getElementById("daisy-internal-page-link");
   pageLink.onclick = function() {
      current.onGotoPage(pageValue.value);
   };

   /* setup playback widget */
   var playButton = this.window.document.getElementById("daisy-internal-play-button");
   playButton.onclick = function() {
      if (this.focusTimer) {
         this.focusTimer.cancel();
         this.focusTimer = null;
      }
      if (current.window.focusTimer) {
         current.window.focusTimer.cancel();
         current.window.focusTimer = null;
      }
      if (!current.bookContext.playing) {
         current.service.tts.cancel();
      }
      setTimeout(function() {
         current.onPlayCurrent();
      },100);
   };
   playButton.onfocus = buttonTTS;
   var nextButton = this.window.document.getElementById("daisy-internal-next-button");
   nextButton.onclick = function() {
      current.onNext();
   }
   nextButton.onfocus = buttonTTS;
   var prevButton = this.window.document.getElementById("daisy-internal-prev-button");
   prevButton.onclick = function() {
      current.onPrevious();
   };
   prevButton.onfocus = buttonTTS;

   /*
    * determine if the book has audio overlays
    */
   var isAudioTest = /^audio/;
   this.isAudioBook = false;
   for (var id in this.book.manifest) {
      if (this.book.manifest[id].mediaType.match(isAudioTest)) {
         this.isAudioBook = true;
         break;
      }
   }

   /**
    * toggle playback widget depending on whether TTS is enabled or if the book has audio overlays
    */
   var playbackWidget = this.window.document.getElementById("daisy-internal-playback-widget");
   if (playbackWidget) {
      this.tts = null;
      if (current.service.readContent && !current.isAudioBook || current.service.forceReadContent) {
         this.tts = this.service.tts;
      }
      if (this.isAudioBook || ((this.service.readContent || current.service.forceReadContent) && this.tts)) {
         playbackWidget.style.display = "block";
      } else {
         playbackWidget.style.display = "none";
         if (!this.isAudioBook && !this.tts) {
            var noPlaybackWidget = this.window.document.getElementById("daisy-internal-no-playback");
            noPlaybackWidget.style.display = "block";
            this.window.document.getElementById("daisy-internal-no-playback-link").onclick = function() {
               var moreWindow = window.openDialog("chrome://daisy/content/no-playback.xul",'daisy-no-playback',"chrome,resizable",{ OS: current.OS});
               moreWindow.focus();
               return false;
            }
         }
      }
   }

   /**
    * initialize self-voicing via TTS
    */
   this.window.onfocus = function() {
      if (!current.bookContext || !current.bookContext.playing) {
         this.focusTimer = setTimeout(function() {
            current.window.focusTimer = null;
            if (current.service.readUI) {
               Console.debug("window focus, reading title");
               current.speak("DAISY Book, "+current.book.title,null);
            }
            var id = current.window.location.hash.substring(1);
            var target = current.window.document.getElementById(id);
            if (target && target.firstChild && target.firstChild.className && target.firstChild.className.indexOf(DAISYHelper.hackClassName)>=0) {
               /**
                * a gigantic hack for crappy software like JAWS
                */
               target.firstChild.focus();
            }
         },250);
      }
   };

   // Initialize tracking of location
   this.locationListener = {
     QueryInterface: function(aIID)
     {
      if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
          aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
          aIID.equals(Components.interfaces.nsISupports))
        return this;
      throw Components.results.NS_NOINTERFACE;
     },

     onLocationChange: function(aProgress, aRequest, aURI)
     {
        var url = aURI.spec;
        var hashPos = url.indexOf("#");
        if (hashPos>0) {
           url = url.substring(0,hashPos);
        }
        //Console.debug("current.url="+current.url+", url="+url);
        if (url!=current.url) {
           return;
        }
        if (current.window.location.hash!=current.hash) {
           Console.debug("location changed to "+current.window.location.hash+", playing="+current.bookContext.playing);
           current.hash = current.window.location.hash;
           if (!current.bookContext.playing) {
              var id = current.hash.substring(1);
              var target = current.window.document.getElementById(id);
              if (target) {
                 current.bookContext.setElement(target);
              }
              if (current.needLocationScroll) {
                 current.needLocationScroll = false;
                 this.locationScrollTimer = setTimeout(function() {
                    current.window.scrollOwner = this;
                    current.window.scroll(0,current.window.scrollY-DAISYHelper.positionOffset);
                 },100);
              }
           }
        }
     },

     onStateChange: function(a, b, c, d) {},
     onProgressChange: function(a, b, c, d, e, f) {},
     onStatusChange: function(a, b, c, d) {},
     onSecurityChange: function(a, b, c) {}
   };

   this.tab.linkedBrowser.addProgressListener(this.locationListener,Components.interfaces.nsIWebProgress.NOTIFY_LOCATION);

}

DAISYHelper.prototype.initSoundManager = function(force) {

   /**
    * The sound manager is loaded into a hidden browser instance located in
    * the status bar.  On windows, this browser instance must be partially
    * visable for the flash to load properly.
    *
    * Once the document is loaded for the browser window and initialized, the
    * same sound manager instance is used for all books.  This keeps books
    * from playing audio over each other.
    */
   var extManager = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
   var id = "{6be5ef88-4289-44d1-81f2-097313ed640b}";
   var soundFile = extManager.getInstallLocation(id).getItemFile(id,"content/sound.xhtml");
   var soundContext = this.mainWindow.document.getElementById("daisy-sound-context");
   var soundDocument = DAISYHelper.fileURL(soundFile).spec;
   if (soundContext.contentWindow.location.href!=soundDocument || force) {
      if (this.OS=="WINNT" && soundContext.boxObject.height==0) {
         soundContext.style.height = "10px";
      }
      this.inSoundManagerInit = true;
      Console.debug("Loading soundManager2");
      /**
       * flash block is evil, so we have to disable it and restore it
       * after we're done.
       */
      FlashBlockHack.save();

      /**
       * we track the progress of the document loading via a WebProgressListener
       * instance.
       */
      var current = this;
      var myListener = {
         QueryInterface: function(aIID) {
            if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
               aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
               aIID.equals(Components.interfaces.nsISupports)) {
               return this;
            }
            throw Components.results.NS_NOINTERFACE;
         },

         onStateChange: function(aWebProgress, aRequest, aFlag, aStatus) {
            if(aFlag & Components.interfaces.nsIWebProgressListener.STATE_STOP && aFlag & Components.interfaces.nsIWebProgressListener.STATE_IS_DOCUMENT && aStatus==0) {
                // This fires when the load finishes

               /**
                * The document is now loaded.  We now need to wait for the
                * sound manger to initialize.  Once that is complete, we're done.
                */
               soundContext.removeProgressListener(myListener);
               var tryCount = 0;
               var waitForInit = function() {
                  if (soundContext.contentWindow.soundManager && soundContext.contentWindow.soundManager._didInit) {
                     current.inSoundManagerInit = false;
                     FlashBlockHack.restore();
                  } else if (tryCount<30) {
                     tryCount++;
                     setTimeout(waitForInit,100);
                  } else {
                     current.inSoundManagerInit = false;
                     FlashBlockHack.restore();
                     Console.log("Sound manager did not initialize.");
                  }
               }
               waitForInit();
            }
            return 0;
         },

         onLocationChange: function(aProgress, aRequest, aURI) { return 0; },
         onProgressChange: function() {return 0;},
         onStatusChange: function() {return 0;},
         onSecurityChange: function() {return 0;},
         onLinkIconAvailable: function() {return 0;}
      };
      soundContext.addProgressListener(myListener,Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
      soundContext.loadURI(soundDocument,null,null);
   }

}

/**
 * This method calculates the current page position
 */
DAISYHelper.prototype.calculatePage = function() {
   if (!this.window.pagelist || this.window.pagelist.length==0) {
      this.pageIndex = -1;
      this.onUpdatePage();
      return;
   }
   try {
      var last = null;
      var i=0;
      var pageIndex = -1;
      for (; i<this.window.pagelist.length; i++) {
         if (!this.window.pagelist[i].element) {
            continue;
         }
         if (this.window.pagelist[i].element.offsetTop<(this.window.scrollY+DAISYHelper.testPositionOffset)) {
            last = this.window.pagelist[i];
            pageIndex = i;
         } else {
            break;
         }
      }
      if (!last) {
         last = this.window.pagelist[0];
         pageIndex = 0;
      }
      this.pageIndex = pageIndex;
      /**
       * Call onUpdatePage to update the page position in the display
       */
      //Console.log("last.element="+last.element.localName+" "+last.element.getAttribute("id"));
      this.onUpdatePage(last ? last.element.getAttribute("id") : null,last ? last.number : 0,this.window.pagelist[this.window.pagelist.length-1].number);
   } catch (ex) {
      Console.log(ex);
   }
}

/**
 * Update the page position display
 */
DAISYHelper.prototype.onUpdatePage = function(pageId,page,total) {
   var pageNo = this.window.document.getElementById("daisy-internal-page-no");
   var hiddenPageNo = this.window.document.getElementById("daisy-internal-page-no-hidden");
   var pageTotal = this.window.document.getElementById("daisy-internal-page-total");
   // clear the pageTotal element
   var node = pageTotal.firstChild;
   while (node) {
      pageTotal.removeChild(node);
      node = pageTotal.firstChild;
   }
   // clear the hiddenPageNo element
   node = hiddenPageNo.firstChild;
   while (node) {
      hiddenPageNo.removeChild(node);
      node = hiddenPageNo.firstChild;
   }
   if (!page || page<0) {
      pageNo.value = "";
   } else {
      /**
       * display the page information including the hidden information for screen readers
       */
      var pos = page.indexOf("Page ");
      if (pos==0) {
         this.pagePrefix = "Page ";
         page = page.substring(pos+5);
      } else {
         this.pagePrefix = null;
      }
      pos = total.indexOf("Page ");
      if (pos==0) {
         total = total.substring(pos+5);
      }
      pageNo.value = page+"";
      pageTotal.appendChild(this.window.document.createTextNode(" of "+total));
      hiddenPageNo.appendChild(this.window.document.createTextNode(" "+page+" "));
   }
}

/**
 * Adds a note to the notes menu
 */
DAISYHelper.prototype.addNoteToMenu = function(note) {
   var option = this.window.document.createElement("option");
   option.appendChild(this.window.document.createTextNode(note.name));
   option.setAttribute("value",note.id+"");
   var notesOptGroup = this.notesMenu.firstChild.nextSibling;
   notesOptGroup.appendChild(option);
}

/**
 * Updates the note name in the notes menu
 */
DAISYHelper.prototype.updateNoteInMenu = function(note) {
   var notesOptGroup = this.notesMenu.firstChild.nextSibling;
   var option = notesOptGroup.firstChild;
   var sid = note.id+"";
   while (option) {
      if (option.value==sid) {
         var node = option.firstChild;
         while (node) {
            option.removeChild(node);
            node = option.firstChild;
         }
         option.appendChild(this.window.document.createTextNode(note.name));
         break;
      }
      option = option.nextSibling;
   }
}

/**
 * removes a note from the notes menu
 */
DAISYHelper.prototype.removeNoteFromMenu = function(note) {
   var notesOptGroup = this.notesMenu.firstChild.nextSibling;
   var option = notesOptGroup.firstChild;
   var sid = note.id+"";
   while (option) {
      if (option.value==sid) {
         notesOptGroup.removeChild(option);
         break;
      }
      option = option.nextSibling;
   }
}

/**
 * Renders a note in the document and returns the anchor
 */
DAISYHelper.prototype.renderNote = function(note) {
   Console.debug("Rendering note: "+note.id);
   var target = this.window.document.evaluate(note.xpath,this.window.document,null,8,null).singleNodeValue;
   if (target) {
      var noteRefContainer = target.ownerDocument.createElement("div");
      noteRefContainer.className = "daisy-user-note-ref";
      noteRefContainer.setAttribute("id","daisy-user-note-ref-"+note.id);

      var noteImage = target.ownerDocument.createElement("img");
      noteImage.setAttribute("src",this.images["info"]);
      noteImage.setAttribute("alt","User note icon");
      noteImage.className = "daisy-skip";
      noteRefContainer.appendChild(noteImage);

      var noteRef = target.ownerDocument.createElement("a");
      noteRef.className = "daisy-user-note-ref";
      noteRef.setAttribute("href","#daisy-user-note-"+note.id)
      noteRef.setAttribute("role","note");
      noteRef.setAttribute("title","user note");
      noteRef.appendChild(target.ownerDocument.createTextNode(note.name));
      noteRefContainer.appendChild(noteRef);

      target.parentNode.insertBefore(noteRefContainer,target);

      var noteDiv = target.ownerDocument.createElement("div");
      noteRefContainer.appendChild(noteDiv);
      noteDiv.style.display = "none";
      noteDiv.className = "daisy-user-note";
      noteDiv.setAttribute("id","daisy-user-note-"+note.id)
      var closer = target.ownerDocument.createElement("span");
      closer.className = "daisy-closer daisy-skip";
      closer.appendChild(target.ownerDocument.createTextNode("X"));
      noteDiv.appendChild(closer);
      /*
      var heading = target.ownerDocument.createElement("h2");
      heading.appendChild(target.ownerDocument.createTextNode(note.name));
      noteDiv.appendChild(heading);
      */
      var p = target.ownerDocument.createElement("p");
      p.appendChild(target.ownerDocument.createTextNode(note.note));
      noteDiv.appendChild(p);

      this.processReference(noteRef);
      
      return noteRef;
   } else {
      Console.log("Cannot find note target: "+note.xpath);
      return null;
   }
}

/**
 * Moves the current scroll position to a note
 */
DAISYHelper.prototype.moveToNote = function(renderedNote) {
   var id = "daisy-user-note-ref-"+renderedNote.note.id;
   this.moveToLocation(id);
   this.currentNote = renderedNote;
   var current = this;
   setTimeout(function() {
      current.notifyTOCChange(true);
   },100);
}

DAISYHelper.prototype.moveToLocation = function(id) {
   this.needLocationScroll = true;
   this.window.location.hash = "#"+id;
   this.linkHackForTarget(id);
}

DAISYHelper.prototype.linkHackForTarget = function(id) {
   var target = this.window.document.getElementById(id);
   if (target && (!target.firstChild || !target.firstChild.className || target.firstChild.className.indexOf(DAISYHelper.hackClassName)<0)) {
      var link = this.window.document.createElement("a");
      link.setAttribute("href","#"+id);
      link.setAttribute("role","heading");
      link.className = DAISYHelper.hackClassName+" daisy-skip";
      if (target.firstChild) {
         target.insertBefore(link,target.firstChild);
      } else {
         target.appendChild(link);
      }
   }
}

/**
 * Moves to the next note
 */
DAISYHelper.prototype.onNextNote = function() {
   this.onStop();
   var scrollY = this.window.scrollY;
   var pageBottom = this.window.scrollY+this.window.innerHeight;

   // if there is a current note and it is within the viewport, search from the current position
   var currentY = 0;
   var currentX = 0;
   if (this.currentNote) {
      var yPos = findYPos(this.currentNote.ref);
      if (yPos>scrollY && yPos<pageBottom) {
         scrollY = yPos;
         currentY = yPos;
         currentX = this.currentNote.ref.offsetLeft;
      }
   }

   //Console.debug("next note, scrollY="+scrollY+", window.scrollY="+this.window.scrollY);

   var nextNote = null;
   var nextYPos = 0;

   // TODO: rendered notes should be kept in sorted order
   var sorted = [];
   for (var i=0; i<this.renderedNotes.length; i++) {
      sorted.push(this.renderedNotes[i]);
   }
   sorted.sort(function(noteA,noteB) {
      var targetA = noteA.ref;
      var targetB = noteB.ref;
      return targetA.offsetTop==targetB.offsetTop ? targetA.offsetLeft-targetB.offsetLeft : targetA.offsetTop-targetB.offsetTop;
   });
   for (var i=0; i<sorted.length; i++) {
      if (sorted[i]==this.currentNote) {
         continue;
      }
      var yPos = findYPos(sorted[i].ref);
      if (yPos>=scrollY) {
         //Console.log("name="+sorted[i].note.name+", next.yPos="+nextYPos+", yPos="+yPos+", next.offsetLeft="+(nextNote ? nextNote.ref.offsetLeft : 0)+", offsetLeft="+sorted[i].ref.offsetLeft+", currentY="+currentY+", currentX="+currentX);
         //Console.debug("Note "+note.id+", "+note.name+" qualifies.");
         // either we have no next note or
         // the nextNote is further down than the current note
         if ((!nextNote && (currentY>0 ? (yPos==currentY ? sorted[i].ref.offsetLeft>currentX : true) : true)) ||
             (nextNote && nextYPos>yPos) ||
             (nextNote && nextYPos==yPos && nextNote.ref.offsetLeft>sorted[i].ref.offsetLeft && (yPos==currentY ? sorted[i].ref.offsetLeft>currentX : true))) {
            //Console.debug("Note "+note.id+", "+note.name+" set as next note.");
            // set the next note to the current note
            nextNote = sorted[i];
            nextYPos = yPos;
         }
      }
   }
   if (nextNote) {
      this.moveToNote(nextNote);
   }
   return nextNote ? true : false;
}

/**
 * Moves to the previous note
 */
DAISYHelper.prototype.onPreviousNote = function() {
   this.onStop();
   var scrollY = this.window.scrollY;
   var pageBottom = this.window.scrollY+this.window.innerHeight;

   // if there is a current note and it is within the viewport, search from the current position
   var currentY = 0;
   var currentX = 0;
   if (this.currentNote) {
      var yPos = findYPos(this.currentNote.ref);
      if (yPos>scrollY && yPos<pageBottom) {
         scrollY = yPos;
         currentY = yPos;
         currentX = this.currentNote.ref.offsetLeft;
      }
   }

   var prevNote = null;
   var prevYPos = 0;

   // TODO: rendered notes should be kept in sorted order
   var sorted = [];
   for (var i=0; i<this.renderedNotes.length; i++) {
      sorted.push(this.renderedNotes[i]);
   }
   sorted.sort(function(noteA,noteB) {
      var targetA = noteA.ref;
      var targetB = noteB.ref;
      return targetA.offsetTop==targetB.offsetTop ? targetA.offsetLeft-targetB.offsetLeft : targetA.offsetTop-targetB.offsetTop;
   });

   for (var i=0; i<sorted.length; i++) {
      if (sorted[i]==this.currentNote) {
         continue;
      }
      var yPos = findYPos(sorted[i].ref);
      if (yPos<=scrollY) {
         Console.debug("name="+sorted[i].note.name+", next.yPos="+prevYPos+", yPos="+yPos+", next.offsetLeft="+(prevNote ? prevNote.ref.offsetLeft : 0)+", offsetLeft="+sorted[i].ref.offsetLeft+", currentY="+currentY+", currentX="+currentX);
         // either we have no previous note or
         // the prevNote is closer to the top than the current note
         if ((!prevNote && (currentY>0 ? (yPos==currentY ? sorted[i].ref.offsetLeft<currentX : true) : true)) ||
            (prevNote && prevYPos<yPos && (yPos!=currentY || (yPos==currentY && sorted[i].ref.offsetLeft<currentX) ) ) ||
            (prevNote && prevYPos==yPos && prevNote.ref.offsetLeft<sorted[i].ref.offsetLeft && (yPos==currentY ? sorted[i].ref.offsetLeft<currentX : true))) {
            // set the next note to the current note
            prevNote = sorted[i];
            prevYPos = yPos;
         }
      }
   }
   if (prevNote) {
      this.moveToNote(prevNote);
   }
   return prevNote ? true : false;
}


/**
 * Updates the note represenation in the document when the note changes
 */
DAISYHelper.prototype.updateNoteRendering = function(note) {
   var noteRef = this.window.document.getElementById("daisy-user-note-ref-"+note.id);
   if (noteRef) {
      noteRef.setAttribute("title","Note "+note.name);
      noteRef.firstChild.setAttribute("title","Note "+note.name);
   }
}

/**
 * Removes a note rendering from the document
 */
DAISYHelper.prototype.removeNoteRendering = function(note) {
   var noteRef = this.window.document.getElementById("daisy-user-note-ref-"+note.id);
   if (noteRef) {
      if (this.bookContext.element==noteRef) {
         var next = noteRef.nextSibling;
         while (next && next.nodeType!=1) {
            next = next.nextSibling;
         }
         this.bookContext.setElement(next ? next : noteRef.parentNode);
      }
      noteRef.parentNode.removeChild(noteRef);
   }
}

/**
 * The main callback to render a book in the current document.
 */
DAISYHelper.prototype.writeContent = function() {

   var current = this;
   var soundContext = this.mainWindow.document.getElementById("daisy-sound-context");
   var soundWindow = soundContext.contentWindow;
   var soundManager  = soundWindow.soundManager;
   var extManager = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
   var extid = "{6be5ef88-4289-44d1-81f2-097313ed640b}"
   var autoPauseSound =  DAISYHelper.fileURL(extManager.getInstallLocation(extid).getItemFile(extid,"content/auto-pause.mp3")).spec;
   /*
    * make sure we have a sound manager.  If not, wait till we do (within a timeout) via polling
    */
   if (!this.skipSoundManager && (this.inSoundManagerInit || !soundManager || !soundManager._didInit || !soundManager.supported())) {
      Console.debug("Waiting for soundManager");
      var soundCount = 0;
      var tryCount = 0;
      var waitForSoundManager = function() {
         if (soundWindow.soundManager && soundWindow.soundManager._didInit && !current.inSoundManagerInit) {
            if (soundWindow.soundManager.supported()) {
               current.writeContent();
            } else if (tryCount<2) {
               Console.log("soundManager didn't load, retrying...");
               tryCount++;
               soundCount = 0;
               current.initSoundManager(true);
               setTimeout(waitForSoundManager,100);
            } else {
               Console.log("soundManager is not supported.");
               current.skipSoundManager = true;
               current.writeContent();
            }
         } else if (soundCount<30) {
            soundCount++;
            setTimeout(waitForSoundManager,100);
         } else {
            Console.log("No soundManager instance could be found in sound context.");
            current.skipSoundManager = true;
            current.writeContent();
         }
      };
      waitForSoundManager();
      return;
   }
   /**
    * toggle the soundmanager browser visuable off if we're on windows
    */
   if (this.OS=="WINNT" && soundContext.boxObject.height==10) {
      soundContext.style.height = "0px";
   }
   if (soundManager && !soundManager.supported()) {
      Console.log("soundManager is not supported.");
      soundManager = null;
   }

   /**
    * update the images in the document with the local content
    * for button images, etc.
    */
   this.images = {};
   var imageNames = ["info","play","pause","prev","next","add-note"];
   for (var i=0; i<imageNames.length; i++) {
      this.images[imageNames[i]] = "chrome://daisy-public/content/"+imageNames[i]+"-button.png";
   }

   /**
    * Change the tab title/etc. to the book title
    */
   this.window.document.title = this.book.title;

   /*
    * add the local stylesheet for DAISY books
    */
   this.forChild(this.window.document.documentElement,"head",function(head) {
      var styleE = head.ownerDocument.createElement("link");
      styleE.setAttribute("rel","stylesheet");
      styleE.setAttribute("type","text/css");
      styleE.setAttribute("href","chrome://daisy-public/content/rendered-"+current.OS.toLowerCase()+".css");
      head.appendChild(styleE);
   });

   /**
    * If the book is severely broken, give up
    */
   if (!this.book.isValid()) {
      var p =this.window.document.createElement("p");
      p.appendChild(this.window.document.createTextNode("This document is not a valid DAISY book packaging."));
      this.window.document.body.appendChild(p);
      p.className = "error"
      current.window.document.body.style.color = "rgb(0,0,0)";
      return;
   }

   var doShow = function() {

      /**
       *  The book is rendered as a child of 'daisy-internal-book'.
       */
      var bookTarget = current.window.document.getElementById("daisy-internal-book");

      Console.debug("Playback pause: "+current.service.playbackPause);

      /**
       * create the navigation context
       */
      current.context = current.renderer.createNavigationContext(current.book,current.book.loadNavigation(),{
         tts: current.tts,
         playbackPause: current.service.playbackPause,
         soundManager: soundManager,
         sound: current.service.sound && !current.service.forceReadContent,
         region: bookTarget,
         focusFollow: current.service.focusFollow,
         highlight: current.service.highlight,
         onFinish: function() {
            Console.debug("onFinish() called on navigation context in helper.");
            if (current.context.next()==0) {
               current.togglePlayButton(true);
            } else {
               if (this.context.element) {
                  current.bookContext.setElement(this.context.element);
               }
            }
         },
         context: {}
      });

      /**
       * listen for changes from the settings dialog (or other preference changes)
       */
      current.prefListener = {
         tts: current.tts,
         onChange: function(name) {
            if (name=="readContent" || name=="forceReadContent") {
               if (current.service.readContent && !current.isAudioBook || current.service.forceReadContent) {
                  if (!this.tts) {
                     this.tts = current.service.tts;
                  }
                  current.context.options.tts = this.tts;
               } else {
                  current.context.options.tts = null;
               }
               current.context.options.sound = current.service.sound && !current.service.forceReadContent;
            } else if (name=="playbackPause") {
               current.context.options.playbackPause = current.service.playbackPause;
            } else if (name=="sound") {
               current.context.options.sound = current.service.sound && !current.service.forceReadContent;
            } else if (name=="focusFollow") {
               current.context.options.focusFollow = current.service.focusFollow;
            } else if (name=="highlight") {
               current.context.options.highlight = current.service.highlight;
            }
         }
      };
      current.service.addPreferenceObserver(current.prefListener);
      
      current.context.owner = current;
      current.context.document = current.window.document;

      /**
       * make the content visable
       */
      current.window.document.body.style.color = "rgb(0,0,0)";

      /**
       * if we opened the sidebar, notify it and attach
       */
      if (!current.openedSidebar) {
         current.notifySidebar();
      }

      /**
       * Create an XHTML renderer context
       */
      var renderContext = current.renderer.xhtmlRender(current.book,bookTarget);

      /*
       * afterRender is called to notify the sidebar that we're done with the transform and setup.  This
       * may need to retry because the sidebar is loaded asynchronously
       */
      var trySidebar = 0;
      var afterRender = function() {
         var sidebar  = current.mainWindow.document.getElementById("sidebar");
         if (sidebar && sidebar.contentWindow.theSidebar && sidebar.contentWindow.theSidebar.attached) {
            Console.debug("onAfterRender called from helper")
            sidebar.contentWindow.theSidebar.onAfterRender();
         } else if (trySidebar<50) {
            trySidebar++;
            setTimeout(afterRender,100);
         } else {
            Console.log("Cannot find sidebar application to notify onAfterRender(): sidebar="+sidebar+" sidebar.contentWindow.theSidebar="+(sidebar ? sidebar.contentWindow.theSidebar : null)+" attached="+(sidebar && sidebar.contentWindow.theSidebar ? sidebar.contentWindow.theSidebar.attached : null));
         }
      };

      /*
       * onload is called after the content has been processed and rendered into the target element
       */
      renderContext.onload = function() {
         var smdebug = current.window.document.getElementById("soundmanager-debug")
         if (smdebug) {
            smdebug.style.display = "none";
         }

         /**
          * Render the notes
          */
         current.renderedNotes = [];
         current.currentNote = null;
         for (var i=0; i<current.notes.length; i++) {
            var noteRef = current.renderNote(current.notes[i]);
            current.renderedNotes.push({ note: current.notes[i], ref: noteRef});
         }

         /**
          * sort the notes in rendered order
          */
         // sort the notes according to position
         current.notes.sort(function(noteA,noteB) {
            var targetA = current.window.document.getElementById("daisy-user-note-ref-"+noteA.id);
            var targetB = current.window.document.getElementById("daisy-user-note-ref-"+noteB.id);
            if (!targetA) {
               return 1;
            }
            if (!targetB) {
               return -1;
            }
            return targetA.offsetTop==targetB.offsetTop ? targetA.offsetLeft-targetB.offsetLeft : targetA.offsetTop-targetB.offsetTop;
         });
         // add the document's notes in rendered order
         for (var i=0; i<current.notes.length; i++) {
            current.addNoteToMenu(current.notes[i]);
         }
         /**
          * add the notes to the navigation
          */
         current.book.addNotes(current.notes,current.noteComparator);

         /**
          * calculate the page and/or toggle pages off if there are no pages
          */
         current.calculatePage();
         if (!current.window.pagelist || current.window.pagelist.length==0) {
            current.hasPages = false;
            current.window.document.getElementById("daisy-internal-page-widget").style.display = "none";
            current.window.document.getElementById("daisy-internal-page-description").style.display = "none";
            current.window.document.getElementById("daisy-internal-page-no").style.display = "none";
         } else {
            current.hasPages = true;
            current.window.document.getElementById("daisy-internal-page-widget").style.display = "inline";
         }

         /**
          * add a resize event listener to pass on resize events to the sidebar
          */
         var sidebar  = current.mainWindow.document.getElementById("sidebar");
         current.window.addEventListener('resize',function() {
            Console.debug("resize, scrollY="+current.window.scrollY);
            if (current.resizeTimer) {
               clearTimeout(current.resizeTimer);
               current.resizeTimer = null;
            }
            current.resizeTimer = setTimeout(function() {
               Console.debug("resize timeout, scrollY="+current.window.scrollY);
               current.resizeTimer = null;
               if (sidebar.contentWindow.theSidebar) {
                  sidebar.contentWindow.theSidebar.resize();
               }
               Console.debug("resize timeout, scrollY="+current.window.scrollY);
               current.resize();
               Console.debug("resize timeout, scrollY="+current.window.scrollY);
            },100);
         },true);

         /**
          * add a scroll listener to track the navigation and page position in the document
          */
         var actionBar = current.window.document.getElementById("daisy-internal-controls");
         current.window.onscroll = function(event) {

            Console.debug("DAISYHelper: window.onscroll, event phase="+event.eventPhase+", scrollY="+current.window.scrollY);
            if (current.scrollTimer) {
               clearTimeout(current.scrollTimer);
               current.scrollTimer = null;
            }

            current.scrollTimer = setTimeout(function() {
               Console.debug("scrollY="+current.window.scrollY);
               current.scrollTimer = null;
               if (!current.window.scrollOwner && current.bookContext && current.bookContext.playing) {
                  Console.debug("Stopping from manual scroll on #"+current.bookContext.element.getAttribute("id"));
                  current.onStop();
               } else {
                  current.window.scrollOwner = null;
               }
               Console.debug("scrollY="+current.window.scrollY);

               actionBar.style.left = "-"+current.window.scrollX+"px";
               current.pageTop = current.window.scrollY;
               if (current.pageTop<0) {
                  current.pageTop = 0;
               }
               current.pageBottom = current.window.scrollY+current.window.innerHeight;

               Console.debug("scrollY="+current.window.scrollY);
               /**
                * track hash position changes so that we know when we change TOC or navigation item
                */
               var hashChanged = false;
               if (!current.hash && current.window.location.hash || current.hash!=current.window.location.hash) {
                  current.changeLocationAt = (new Date()).getTime();
                  hashChanged = true;
               }
               current.scrollCounter++;
               current.hash = current.window.location.hash;

               Console.debug("scrollY="+current.window.scrollY);
               Console.debug("hash: "+current.hash+", changed="+hashChanged);

               /**
                * recalculate the page
                */
               current.calculatePage();

               Console.debug("scrollY="+current.window.scrollY);
               /**
                * Check to see if we need to load more images
                */
               if (current.updateImages()) {
                  // we've loaded images, so notify the sidebar that sizing and context has changed
                  if (sidebar.contentWindow.theSidebar) {
                     setTimeout(function() {
                        sidebar.contentWindow.theSidebar.resize();
                        if (sidebar.contentWindow.theSidebar) {
                           sidebar.contentWindow.theSidebar.updateContext(current.window.scrollY+100);
                        }
                     },100);
                  }
               } else {
                  // No images were loaded, so just notify the sidebar the context changed
                  if (sidebar.contentWindow.theSidebar) {
                     sidebar.contentWindow.theSidebar.updateContext(current.window.scrollY+100);
                     if (hashChanged) {
                        var yPos = current.window.scrollY-DAISYHelper.positionOffset;
                        if (yPos<0) {
                           yPos = 0;
                        }
                        current.window.scrollOwner = current;
                        current.window.scroll(0,yPos);
                     }
                  }
               }
               Console.debug("scrollY="+current.window.scrollY);
               if (current.context.playContext) {
                  // we're in playback so we need to track the navigation position
                  // we need to find the current navigation item from the current position forward
                  current.updatePlaybackPosition();
               }
            },250);
         };

         var contextTTS = current.tts;
         if (!contextTTS && current.service.readContent) {
            contextTTS = current.service.tts;
         }
         var useStructural = (current.service.navigateByStructure && !current.isAudioBook ||
                                current.service.navigateByStructure && current.isAudioBook && current.service.forceReadContent);
         current.bookContext =  useStructural ? new StructuralPlayback() : new SMILPlayback();
         Console.debug(useStructural ? "Using structural playback" : "Using SMIL playback");
         current.bookContext.init({
            owner: current,
            document: current.window.document,
            tts: contextTTS,
            soundManager: soundManager,
            sounds: {
               'auto-pause': autoPauseSound,
               'table-boundary': autoPauseSound,
               'no-link': autoPauseSound
            },
            locator: function(element) {
               return current.getElementXPath(element);
            }
         });
         Console.debug("playback.ttsCallback="+current.bookContext.ttsCallback);

         current.bookContext.element = current.bookContext.top;
         var node = current.bookContext.element.firstChild;
         while (node && node.nodeType!=1 && node.nextSibling) {
            node = node.nextSibling;
         }
         if (node.nodeType==1) {
            current.bookContext.element = node;
            current.bookContext.newPosition = true;
         }


         current.context.isRendered = true;
         current.window.daisyRendered = true;

         /**
          * Load the first set of images from the current document position
          */
         current.window.loadedImages = [];
         current.updateImagePlaceHolders();
         current.updateImages();

         /**
          * if there is a fragment identifier, find the correct position in the navigation
          */
         if (current.window.location.hash.length>0) {
            Console.debug("Updating position to fragment "+current.window.location.hash);
            var id = current.window.location.hash.substring(1);
            if (id.length>0) {
               var target = current.window.document.getElementById(id);
               if (target) {
                  var ypos = findYPos(target)-DAISYHelper.positionOffset;
                  if (ypos<0) {
                     ypos = 0;
                  }
                  current.window.scroll(0,ypos);
                  //current.bookContext.element = target;
                  current.notifyTOCChange();
               }
               current.linkHackForTarget(id);
            }
         }

         /**
          * Change the status bar indicator for the documemnt
          */
         var validPanel  = current.mainWindow.document.getElementById("daisy-status-bar-valid");
         var quirkyPanel  = current.mainWindow.document.getElementById("daisy-status-bar-quirky");
         if (current.book.quirky) {
            validPanel.style.display = "none";
            quirkyPanel.style.display = "block";
         } else {
            validPanel.style.display = "block";
            quirkyPanel.style.display = "none";
         }

         /*
          * notify the sidebar that we're done with rendering and setup
          */
         afterRender();
         if (current.context.onload) {
            current.context.onload();
         }

         current.processReferences();

      }
      // showing the book initiates the render
      renderContext.show(current.book.dtbook);
   };
   doShow();
}

DAISYHelper.prototype.processReferences = function() {
   var links = this.window.document.links;
   for (var i=0; i<links.length; i++) {
      var names = links[i].className.split(/\s+/);
      for (var j=0; j<names.length; j++) {
         if (names[j]=="daisy-noteref" || names[j]=="daisy-annoref" || names[j]=="daisy-user-note-ref") {
            this.processReference(links[i]);
            break;
         }
      }
   }
}

DAISYHelper.prototype.processReference = function(link) {
   var current = this;
   var id = link.getAttribute("href");
   if (id.charAt(0)!='#') {
      return;
   }
   id = id.substring(1);
   var target = this.window.document.getElementById(id);
   if (!target) {
      return;
   }
   target.style.display = "none";
   var closer = target.firstChild;
   while (closer && closer.className.indexOf("daisy-closer")<0) {
      closer = closer.nextSibling;
   }
   if (closer) {
      closer.onclick = function() {
         target.style.display = "none";
      }
   }
   link.onclick = function() {
      var pos = findPos(link);
      target.style.position = "absolute";
      target.style.left = pos[0]+"px";
      target.style.top = (pos[1]+link.offsetHeight+3)+"px";
      target.style.display = "block";
      current.bookContext.setElement(target, { referer: link, invoke: true });
      target.focus();
      return false;
   }
}

/**
 * as the document scrolls during playback, this method tracks
 * the current navigation item regardless of whether the sidebar is open
 */
DAISYHelper.prototype.updatePlaybackPosition = function(scrollY) {
   var point = this.context.currentPoint;
   var last = point;
   var top = this.window.scrollY+100;
   do {
      var target = this.window.document.getElementById(point.content.id);
      if (target && target.offsetTop<top) {
         last = point;
         while (point && !point.next) {
            point = point.parent;
         }
         if (point) {
            point = point.next;
         }
      }
   } while (point && (!target || target.offsetTop<top));
   if (last) {
      this.context.currentPoint = last;
   }
}

/**
 * Move to page
 */
DAISYHelper.prototype.onGotoPage = function(page,skipPrefix) {
   this.onStop();
   var id = this.context.moveToPage(this.pagePrefix && !skipPrefix ? this.pagePrefix+page : page,true,true);
   var target = this.window.document.getElementById(id);
   if (target) {
      // change the page by changing the fragment identifier
      var current = this;
      setTimeout(function() {
         current.bookContext.setElement(target,{invoke: true});
      },250);
   }
}

/**
 * Move to the next page
 */
DAISYHelper.prototype.onNextPage = function() {
   this.onStop();
   Console.debug("Next page, pageIndex="+this.pageIndex);
   if (this.pageIndex<(this.window.pagelist.length-1)) {
      this.pageIndex++;
      this.onGotoPage(this.window.pagelist[this.pageIndex].number,true);
      return true;
   } else {
      return false;
   }
}

/**
 * Move to the previous page
 */
DAISYHelper.prototype.onPreviousPage = function() {
   this.onStop();
   Console.debug("Previous page, pageIndex="+this.pageIndex);
   if (this.pageIndex>0) {
      this.pageIndex--;
      this.onGotoPage(this.window.pagelist[this.pageIndex].number,true);
      return true;
   } else {
      return false;
   }
}


/**
 * launch the manage notes dialog
 */
DAISYHelper.prototype.onManageNotes = function(note) {
   if (this.closed) {
      return;
   }
   var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
   this.noteManager = ww.openWindow(null,"chrome://daisy/content/manage-notes.xul",this.noteManagerId,"chrome,resizable",{});
   this.noteManager.focus();
   var current = this;
   var context = {
      note: note,
      database: this.service.database,
      notes: this.notes,
      updateNote: function(note) {
         this.database.updateNote(note);
         current.updateNoteRendering(note);
         current.updateNoteInMenu(note);
         note.point.labels[0].text = "Note: "+note.name;
         note.point.item.firstChild.firstChild.setAttribute("label",note.point.labels[0].text);
      },
      removeNote: function(note) {
         for (var i=0; i<this.notes.length; i++) {
            if (this.notes[i].id==note.id) {
               this.database.removeNote(note);
               current.book.removeNote(note);
               note.point.item.parentNode.removeChild(note.point.item);
               this.notes.splice(i,1);
               current.removeNoteRendering(note);
               current.removeNoteFromMenu(note);
               current.updateNavigationPosition();
               break;
            }
         }
      }
   };
   var tryCount = 0;
   var waitForWindow = function() {
      if (current.noteManager.ManageNotes && current.noteManager.ManageNotes.loaded) {
         current.noteManager.ManageNotes.getInstance().init(context);
      } else if (tryCount<20) {
         tryCount++;
         setTimeout(waitForWindow,100);
      } else {
         Console.log("Cannot initialize manage notes window.");
      }
   }
   waitForWindow();
}

DAISYHelper.prototype.onMoveToActionBar = function() {
   if (this.hasPages) {
      this.window.document.getElementById("daisy-internal-page-no").focus();
   } else {
      this.notesMenu.focus();
   }
}

/**
 * launch the add note dialog
 */
DAISYHelper.prototype.onAddNote = function() {
   var location = this.getBookLocation();
   if (!location) {
      Console.log("Cannot find location target for note.");
      return;
   }
   Console.debug("note target: "+location.xpath);
   var current = this;
   var heading = "";
   var node = location.target;
   while (node) {
      if (node.localName=="h1" || node.localName=="h2" || node.localName=="h3" || node.localName=="h4" || node.localName=="h5") {
         heading = node.textContent;
         break;
      }
      node = node.previousSibling;
   }
   var data = {
      location: location,
      name: this.book.title+" / "+heading+" / "+(this.notes.length+1),
      onClose: function() {
         current.noteDialogs[this.dialogIndex] = null;
         if (this.success) {
            var note = current.service.database.addNote(current.url,this.location.xpath,this.name,this.note);
            current.notes.push(note);
            current.addNoteToMenu(note);
            var noteRef = current.renderNote(note);
            current.renderedNotes.push({ note: note, ref: noteRef});
            current.book.addNote(note,current.noteComparator);
            var sidebar  = current.mainWindow.document.getElementById("sidebar");
            if (sidebar && sidebar.contentWindow.theSidebar) {
               sidebar.contentWindow.theSidebar.notifyAddNote(note);
            }
         }
      }
   };
   data.wrappedJSObject = data;
   var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
   var noteWindow = ww.openWindow(null,"chrome://daisy/content/add-note.xul",'daisy-add-note-'+(new Date()).getTime(),"chrome,resizable",data);
   noteWindow.focus();
   data.dialogIndex = this.noteDialogs.length;
   this.noteDialogs.push(noteWindow);
}

/**
 * Calculate the current position and return an XPath
 */
DAISYHelper.prototype.getElementXPath = function(node) {
   var concat = function(pre,post) {
      return post.length==0 ? pre : pre+"/"+post;
   }
   var xpath = "";
   while (node) {
      if (node.nodeType==1) {
         var id = node.getAttribute("id");
         if (id) {
            return concat("id('"+id+"')",xpath);
         }
         var count = 0;
         var current = node;
         while (current) {
            if (current.nodeType==1) {
               count++;
            }
            current = current.previousSibling;
         }
         xpath = concat("*["+count+"]",xpath);
      } else {
         xpath = "/"+xpath;
      }
      node = node.parentNode;
   }
   return xpath;
}


/**
 * find the book location from the scroll position
 */
DAISYHelper.prototype.getBookLocation = function() {
   var target = this.bookContext.element;
   if (!target) {
      Console.debug("No book context, calculating by position.");
      if (!this.window.pagelist || this.window.pagelist.length==0) {
         // no pages
         var book = this.window.document.getElementById("daisy-internal-book");
         var child = book.firstChild;
         while (child && child.className!="daisy-book") {
            child = child.nextSibling;
         }
         if (child) {
            var top = child.firstChild;
            var node = top;
            Console.debug("node: "+this.getElementXPath(node));
            while (node && (node.nodeType!=1 || (node.nodeType==1 && node.offsetTop<this.window.scrollY))) {
               if (node.nodeType==1 && node.firstChild) {
                  node = node.firstChild;
               } else if (node.nextSibling) {
                  node = node.nextSibling;
               } else {
                  while (!node.nextSibling) {
                     node = node.parentNode;
                     if (node==top) {
                        break;
                     }
                  }
                  node = node.nextSibling;
               }
            }
            target = node;
         }
      } else {
         target = this.window.pagelist[this.pageIndex].element;
         if (target.nextSibling) {
            var seek = target.nextSibling;
            while (seek && seek.nodeType!=1) {
               seek = seek.nextSibling;
            }
            if (seek) {
               target = seek;
            }
         }

      }

   } else if (this.window.pagelist && this.window.pagelist.length>0){
      /*
       * we have a book context but we also have pages, so check to see if
       * the current page is closer to the viewport
       */
      Console.log("ypos="+this.bookContext.yPos+", top="+this.pageTop+", bottom="+this.pageBottom);
      if (this.bookContext.yPos<this.pageTop || this.bookContext.yPos>this.pageBottom) {
         /**
          * the context is not in the viewport, so use the current page
          */
         target = this.window.pagelist[this.pageIndex].element;

         Console.log("page: "+target.getAttribute("id")+", "+target.textContent);
         /**
          * seek to the sibling of the page
          */
         if (target.nextSibling) {
            var seek = target.nextSibling;
            while (seek && seek.nodeType!=1) {
               seek = seek.nextSibling;
            }
            if (seek) {
               target = seek;
            }
         }
      }
   }
   if (target) {
      return {
         id: target.getAttribute("id"),
         target: target,
         xpath: this.getElementXPath(target)
      };
   } else {
      return null;
   }
}

/**
 * stop playback
 */
DAISYHelper.prototype.onStop = function() {
   if (!this.bookContext) {
      return;
   }
   this.bookContext.stop();
}

/**
 * previous action in playback
 */
DAISYHelper.prototype.onPrevious = function() {
   this.bookContext.previous();
}

/**
 * next button in playback
 */
DAISYHelper.prototype.onNext = function() {
   this.bookContext.next();
}

/**
 * play in playback
 */
DAISYHelper.prototype.onPlayCurrent = function() {
   this.bookContext.togglePlay();
}

/**
 * toggle the play button icon
 */
DAISYHelper.prototype.togglePlayButton = function(flag) {
   if (!this.window.document) {
      return;
   }
   if (flag) {
      this.window.document.getElementById("daisy-internal-play-button").firstChild.setAttribute("src",this.images["play"]);
   } else {
      this.window.document.getElementById("daisy-internal-play-button").firstChild.setAttribute("src",this.images["pause"]);
   }
}

DAISYHelper.prototype.updateImagePlaceHolders = function() {
   // the unactivated images are stored in the window
   var images = this.window.imagelist;

   // enumerate the unactivated images and look for images that need to be shown
   for (var i=0; i<images.length; i++) {

      // a span is stored in the document by id
      var target = this.window.document.getElementById(images[i].id);

      // replace the span with an image element to load the image
      var altSpan = this.window.document.getElementById("daisy-image-alt-"+images[i].id);

      altSpan.style.display = "inline-block";
      if (images[i].height.length>0) {
         altSpan.style.height = images[i].height;
      }
      if (images[i].width.length>0) {
         altSpan.style.width = images[i].width;
      }
   }
}

DAISYHelper.imageResize = function(img) {
   var pos = findPos(img);
   var rightEdge = pos[0]+img.originalSize.width;

   var padding = (img.offsetLeft-img.parentNode.offsetLeft)*2 + 2*DAISYHelper.imagePadding;  // + 20px padding

   if (rightEdge>img.ownerDocument.defaultView.innerWidth) {
      // resize image
      var delta = rightEdge-img.ownerDocument.defaultView.innerWidth;
      var width = img.originalSize.width-delta-20-padding; // 2*padding + 20 px of scrollbars
      if (width>40) {
         if (width>img.parentNode.clientWidth) {
            // adjust so that it is not larger than parent
            width = img.parentNode.clientWidth-padding;
         }
         if (img.getAttribute("height")) {
            img.removeAttribute("height");
         }
         Console.debug("Resizing image width for "+img.getAttribute("src")+" from "+img.clientWidth+"px to "+width+"px");
         img.setAttribute("width",width+"");
         img.originalSize.resized = true;

      } else {
         Console.log("Failed to resize image "+img.getAttribute("src")+" as it would be too small.");
      }
   } else if (img.originalSize.resized) {
      Console.log("Restoring image "+img.getAttribute("src")+" size.");
      img.originalSize.resized = false;
      img.setAttribute("width",img.originalSize.width+"");
      img.setAttribute("height",img.originalSize.height+"");
      if (img.parentNode.clientWidth<img.clientWidth) {
         // resize to parent width minus padding
         Console.log("Image overflowed, resizing image "+img.getAttribute("src")+" to parent.");
         img.setAttribute("width",(img.parentNode.clientWidth-padding)+"");
         img.removeAttribute("height");
         img.originalSize.resized = true;
      }
   } else if (img.parentNode.clientWidth<img.clientWidth) {
      // bad CSS rendering, force a resize
      // resize to parent width minus padding
      Console.log("Resizing image "+img.getAttribute("src")+" to parent.");
      img.setAttribute("width",(img.parentNode.clientWidth-padding)+"");
      img.removeAttribute("height");
      img.originalSize.resized = true;
   }
}
/**
 * Calculates if new images need to be loaded based on scroll position
 */
DAISYHelper.prototype.updateImages = function() {
   var current = this;
   // Set the image active region to be viewport + 500px
   var pageTop = this.window.scrollY;
   if (pageTop<0) {
      pageTop = 0;
   }
   var pageBottom = this.window.scrollY+this.window.innerHeight+500;

   // the unactivated images are stored in the window
   var images = this.window.imagelist;

   var loaded = false;

   var loadCounter = this.scrollCounter+0;

   // enumerate the unactivated images and look for images that need to be shown
   for (var i=0; i<images.length; i++) {

      // a span is stored in the document by id
      var target = this.window.document.getElementById(images[i].id);

      // the span has to be in the image active region
      if (target.offsetTop>=pageTop && target.offsetTop<pageBottom && target.localName=="span") {
         // replace the span with an image element to load the image
         var altSpan = this.window.document.getElementById("daisy-image-alt-"+images[i].id);
         var img = this.window.document.createElementNS("http://www.w3.org/1999/xhtml","img");
         img.onload = function() {
            current.window.loadedImages.push(this);
            this.originalSize = {
               height: img.naturalHeight,
               width: img.naturalWidth
            };
            var h = this.getAttribute("height");
            if (h) {
               try {
                  this.originalSize.height = parseInt(h);
               } catch (ex) {
                  Console.log("Bad image height value "+h+" on "+this.getAttribute("src"));
               }
            }
            var w = this.getAttribute("width");
            if (w) {
               try {
                  this.originalSize.width = parseInt(w);
               } catch (ex) {
                  Console.log("Bad image width value "+w+" on "+this.getAttribute("src"));
               }
            }
            Console.debug("Image "+this.getAttribute("src")+" loaded, scrollCounter="+current.scrollCounter+", loadCounter="+loadCounter);
            DAISYHelper.imageResize(this);
            var now = (new Date()).getTime();
            if (current.hash && current.scrollCounter==loadCounter && (now-current.changeLocationAt)<2000) {
               var id = current.hash.substring(1);
               var target = current.window.document.getElementById(id);
               if (target) {
                  current.window.scroll(0,findYPos(target)-DAISYHelper.positionOffset);
                  loadCounter++;
               }
            }
         }
         img.setAttribute("src",images[i].src);
         img.setAttribute("id",images[i].id);
         if (images[i].height.length>0) {
            img.setAttribute("height",images[i].height);
         }
         if (images[i].width.length>0) {
            img.setAttribute("width",images[i].width);
         }
         if (altSpan) {
            img.setAttribute("alt",altSpan.textContent);
            altSpan.parentNode.removeChild(altSpan);
         }
         var nextSibling = target.nextSibling;
         var parent = target.parentNode;
         parent.removeChild(target);
         if (nextSibling) {
            parent.insertBefore(img,nextSibling);
         } else {
            parent.appendChild(img);
         }

         // remove the image from the unactivated images
         images.splice(i,1);
         i--;

         // change the loaded status
         loaded = true;

         // update context if image is current context (happens on document load)
         if (this.bookContext.element==target) {
            this.bookContext.element = img;
            this.bookContext.element.className = target.className;
         }
      }
   }
   
   return loaded;
}

DAISYHelper.prototype.resize = function() {
   Console.debug("DAISYHelper.resize()");
   this.resizeImages();
}

DAISYHelper.prototype.resizeImages = function() {
   for (var i=0; i<this.window.loadedImages.length; i++) {
      DAISYHelper.imageResize(this.window.loadedImages[i]);
   }
}

/**
 * changes position to nearest navigation item so next works.
 * This method is no longer used
 */
DAISYHelper.prototype.seekNavigationStart = function(target) {
   if (target.localName=="h1" || target.localName=="h2" || target.localName=="h3" || target.localName=="h4" || target.localName=="h5" || target.localName=="h6") {
      Console.debug("Seeking previous for heading.");
      // attempt to seek to previous element so navigation starts at header
      var previous = target.previousSibling;
      while (previous && previous.nodeType!=1) {
         previous = previous.previousSibling;
      }
      if (previous) {
         // just set the context to the previous sibling and "next" will do the right thing
         target = previous;
         target = this.bookContext._findLastDescendant(target);
      } else {
         // set the context to the parent node and next will descend
         target = target.parentNode;
      }
   }
   return target;
}

/*
 * Notification method for when the fragment identifier is changed via the sidebar or otherwise
 */
DAISYHelper.prototype.notifyTOCChange = function(invokeNote) {
   var id = this.window.location.hash.length>0 ? this.window.location.hash.substring(1) : null;
   if (!id) {
      return;
   }
   var target = this.window.document.getElementById(id);
   if (target) {
      //target = this.seekNavigationStart(target);
      this.bookContext.setElement(target,{ newPosition: true, invoke: invokeNote });
   }
}


/**
 * Called by the containing document when the tab or document is closed
 */
DAISYHelper.prototype.close = function() {
   this.closed = true;
   this.window.onfocus = null;

   this.onStop();
   this.service.tts.cancel();

   //this.tab.linkedBrowser.removeProgressListener(this.locationListener);

   /*
    * unreference the book so that it can be unloaded from the cache
    */
   this.book.unref();
   Console.debug("Closed book "+this.url);

   /**
    * close dialogs
    */
   for (var i=0; i<this.noteDialogs.length; i++) {
      if (this.noteDialogs[i]) {
         this.noteDialogs[i].close();
      }
   }
   if (this.noteManager) {
      this.noteManager.close();
   }

   /**
    * remove preference listener
    */
   this.service.removePreferenceObserver(this.prefListener);

   /**
    * calculate book position
    */

   var url = this.window.location.href;

   /**
    * first we'll try to preserve the keyboard navigation context
    * if it is on the page
    */
   var contextOnPage = false;
   var id = null;
   if (this.bookContext.element) {
      
      Console.debug("yPos="+this.bookContext.yPos+", top="+this.pageTop+", bottom="+this.pageBottom);

      if (this.bookContext.yPos>=this.pageTop && this.bookContext.yPos<=this.pageBottom) {

         // If there is a book context element and it has an 'id'
         // attribute, replace the fragment identifier (the TOC item)
         // with the element's id
         id = this.bookContext.element.getAttribute("id");
         if (id) {
            contextOnPage = true;
         }
      }
   }

   /**
    * if the context wasn't in the viewport or we can't construct
    * a url reference, we'll try to calculate the page.
    */
   if (!contextOnPage && this.pageIndex>=0) {
      id = this.window.pagelist[this.pageIndex].element.getAttribute("id");
      Console.debug("Using page number instead: "+id);
   }
   
   /**
    * update the fragment identifier
    */
   if (id) {
      // remove query
      var q = url.indexOf("?");
      if (q>0) {
         url = url.substring(0,q);
      }
      // replace has with book context element position
      var pos = url.indexOf("#");
      if (pos>0) {
         url = url.substring(0,pos)+"#"+id;
      } else {
         url = url+"#"+id;
      }
   }
   
   /**
    * update book position
    */
   Console.debug("Updating book viewed: "+url);
   this.service.database.bookViewed(this.book.title=="" ? null : this.book.title,url);
}

DAISYHelper.prototype.startLoadingIndicator = function() {
   var container = this.window.document.getElementById("daisy-internal-loading");
   var square = container.firstChild.nextSibling;
   var current = this;
   var direction = 1;
   var pos = 0;
   var move = function() {
      if (current.closed) {
         return;
      }
      if (current.context && current.context.isRendered) {
         container.style.display = "none";
         return;
      }
      Console.debug("tick");
      try {
         if (direction>0 && (pos+50)>container.offsetWidth) {
            direction = -1;
         } else if (direction<0 && pos==0) {
            direction = 1;
         }
         if (direction>0) {
            pos += 50;
         } else {
            pos -= 50;
         }
         square.style.left = pos+"px";
         current.loadingTimer = setTimeout(move,500);
      } catch (ex) {
         Console.log("Error in loading timer: "+ex);
      }
   };
   this.loadingTimer = setTimeout(move,500);
}

DAISYHelper.prototype.speak = function(text,callback) {
   var lock = this.service.getLockFor("tts");
   var current = this;
   this.ownedLock = lock.acquire({
      exclusive: true,
      onRelinquish: function() {
         current.service.tts.cancel();
      }
   });
   this.service.tts.speak(text,{
      QueryInterface: function(aIID) {
         if (aIID.equals(Components.interfaces.nsITTSCallback) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports)) {
            return this;
         }
         throw Components.results.NS_NOINTERFACE;
      },
      onFinish: function() {
         if (current.ownedLock) {
            current.ownedLock.release();
            current.ownedLock = null;
         }
         if (callback) {
            callback();
         }
      }
   });

}

/**
 * XPCOM information to integrate into Javascript
 */
DAISYHelper.prototype.QueryInterface = function(iid) {
    if (iid.equals(Components.interfaces.nsISupports) || 
        iid.equals(Components.interfaces.nsIDAISYHelper) ||
        iid.equals(Components.interfaces.nsIClassInfo)) {
       return this;
    }
    throw Components.results.NS_ERROR_NO_INTERFACE;
};

DAISYHelper.prototype.classDescription = "A DAISY Reader Helper Class";
DAISYHelper.prototype.classID = CLASSID;
DAISYHelper.prototype.contractID = CONTRACTID;
DAISYHelper.prototype.implementationLanguage = Components.interfaces.nsIProgrammingLanguage.JAVASCRIPT;
DAISYHelper.prototype.flags = Components.interfaces.nsIClassInfo.DOM_OBJECT;
DAISYHelper.prototype.getHelperForLanguage = function() { return null; }
DAISYHelper.prototype.getInterfaces = function(countRef) {
   var interfaces = [Components.interfaces.nsIDAISYHelper, Components.interfaces.nsIClassInfo, Components.interfaces.nsISupports];
   countRef.value = interfaces.length;
   return interfaces;
}

/**
 * Flash block enable/disable code
 */

var FlashBlockHack = {
  prefs: Components.classes["@mozilla.org/preferences-service;1"]
         .getService(Components.interfaces.nsIPrefBranch)
         .QueryInterface(Components.interfaces.nsIPrefBranchInternal),
  valueSet: false,
  value: false,

  // Returns the value of the flashblock.enabled pref
  save: function() {
     if (this.prefs.prefHasUserValue("flashblock.enabled")) {
        this.value = this.prefs.getBoolPref("flashblock.enabled");
        this.valueSet = true;
        Console.debug("User value: "+this.value);
     } else {
        this.valueSet = false;
        Console.debug("flashblock.enabled is not set");
     }
     this.prefs.setBoolPref("flashblock.enabled", false);
  },


  // Sets the flashblock.enabled pref to the given boolean value
  restore: function() {
     Console.debug("Restoring flashblock.enabled");
     try {
        if (this.valueSet) {
           this.prefs.setBoolPref("flashblock.enabled", this.value);
        } else {
           this.prefs.clearUserPref("flashblock.enabled");
        }
     } catch (ex) {
     }
  }

}


function GenericComponentFactory(ctor, params) {
  this._ctor = ctor;
  this._params = params;
}
GenericComponentFactory.prototype = {
  _ctor: null,
  _params: null,

  createInstance: function GCF_createInstance(outer, iid) {
    if (outer != null)
      throw Cr.NS_ERROR_NO_AGGREGATION;
    return (new this._ctor(this._params)).QueryInterface(iid);
  },

  QueryInterface: function GCF_QueryInterface(iid) {
    if (iid.equals(Components.interfaces.nsIFactory) ||
        iid.equals(Components.interfaces.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  }
};


var Module = {
  QueryInterface: function M_QueryInterface(iid) {
    if (iid.equals(Components.interfaces.nsIModule) ||
        iid.equals(Components.interfaces.nsISupports))
      return this;
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  getClassObject: function M_getClassObject(cm, cid, iid) {
    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    return new GenericComponentFactory(DAISYHelper);
  },

  registerSelf: function M_registerSelf(cm, file, location, type) {
    var cr = cm.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    cr.registerFactoryLocation(CLASSID, CLASSNAME, CONTRACTID,
                               file, location, type);

    /**
     * Register the global constructor for DAISYHelper
     */
    
    /**
     * In this module we also register the DAISYHelper as a global constructor in
     * javascript.  This allows any document to instantiate the DAISYHelper object
     * by "new DAISYHelper().
     *
     * This object only works if the URI of the current window points to a book already
     * loaded by one of the converters in this module.
     *
     */
    var catManager = Components.classes["@mozilla.org/categorymanager;1"].getService(Components.interfaces.nsICategoryManager);
    catManager.addCategoryEntry("JavaScript global constructor","DAISYHelper","@benetech.org/daisy-helper",true,true);
  },

  unregisterSelf: function M_unregisterSelf(cm, location, type) {
     var catManager = Components.classes["@mozilla.org/categorymanager;1"].getService(Components.interfaces.nsICategoryManager);
     catManager.deleteCategoryEntry("JavaScript global constructor","DAISYHelper",true);
     var cr = cm.QueryInterface(Components.interfaces.nsIComponentRegistrar);

     cr.registerFactoryLocation(CLASSID, location);
  },

  canUnload: function M_canUnload(cm) {
    return true;
  }
};

function NSGetModule(cm, file) {
  return Module;
}


function BookContext()
{
}

BookContext.prototype.init = function(options) {
   this.owner = options.owner;
   this.tts = options.tts;
   this.noteIndex = -1;
   this.top = options.document.getElementById("daisy-internal-book").firstChild;
   this.something = false;
   this.descend = true;
   this.newPosition = false;
   this.revert = null;
   this.locator = options.locator;
   if (!this.locator) {
      this.locator = function(element) { return ""; }
   }
   this.soundManager = options.soundManager;
   this.sounds = options.sounds;
   this.boundaryCount = 0;

}

// we need to run TTS is a callback because it can
// potentially block the main thread depending on how
// the TTS engine is implemented.
BookContext.speak = function(context,text,callback) {
   if (context.tts) {
      // cleanup text:
      var parts = text.split(/\n/);
      var spoken = "";
      for (var i=0; i<parts.length; i++) {
         if (i>0 && parts[i-1].length>0) {
            spoken += " ";
         }
         spoken += parts[i];
      }

      context.timer = setTimeout(function() {
         context.tts.speak(spoken,callback);
      },10);
   }
};

BookContext.skipDescendants = {
   process: function(element,context) {
      context.revert = context._makeRestore(element);
      context._makeMark(element).exec();
      var text = BookContext.inlineText(element);
      if (text=="") {
         text = "empty";
      }
      BookContext.speak(context,text,context.ttsCallback);
      return false;
   },
   canDescend: function(e) { return false }
};

BookContext.canMixedDescend = function(element) {

   // only desend if there are sentences
   var sent = false;
   var child = element.firstChild;
   while (child && !sent) {
      if (child.nodeType==1 && child.localName=="span" && child.className.indexOf("daisy-sent")>=0) {
         sent = true;
      }
      child = child.nextSibling;
   }

   return sent;

/*
   var child = element.firstChild;
   var mixed = false;
   while (child && !mixed) {
      if (child.nodeType==3 || child.nodeType==4) {
         var parts = child.textContent.split(/\s+/);
         for (var i=0; i<parts.length && !mixed; i++) {
            if (parts[i].length>0) {
               mixed = true;
            }
         }
      }
      child = child.nextSibling;
   }
   if (!mixed) {
      // if do not have mixed content, check for sent
      child = element.firstChild;
      var sent = false;
      while (child && !sent) {
         if (child.nodeType==1 && child.localName=="span" && child.className.indexOf("daisy-sent")>=0) {
            sent = true;
         }
         child = child.nextSibling;
      }
      if (!sent) {
         // no sentences, switch to mixed
         mixed = true;
      }
   }
   return !mixed;
   */
};

BookContext.inlineText = function(element) {
   var text = "";
   var child = element.firstChild;
   while (child) {
      var suffix = "";
      var skip = false;
      var recurse = true;
      if (child.nodeType==1 && child.localName=="a") {
         if (child.className==DAISYHelper.hackClassName) {
            child = child.nextSibling;
            continue;
         }
         var role = child.getAttribute("role");
         if (!role || role.length==0) {
            role = "link";
         }
         var title = child.getAttribute("title");
         if (role!="note" || !title || title.length==0) {
            text += ", "+role;
         }
         if (title && title.length>0) {
            text += ", "+title;
         }
         if ((title && title.length>0) || (role && role.length>0)) {
            text += ", ";
         }
         suffix = ", ";
      } else if (child.nodeType==1 && child.localName=="img") {
         suffix = child.getAttribute("alt");
         if (!suffix || suffix=="") {
            suffix = ", Image (no description), image complete, ";
         } else {
            suffix = ", Image "+suffix+" , image complete, ";
         }
      } else if (child.nodeType==1 && child.localName=="math") {
         var alt = child.getAttribute("alttext");
         if (alt.length>0) {
            skip = true;
            text += " "+alt+" ";
         }
      }
      if (!skip) {
         if (child.nodeType==1) {
            text += " "+BookContext.inlineText(child)+suffix;
         } else {
            text += " "+child.textContent+suffix;
         }
      }
      child = child.nextSibling;
   }
   //Console.log("text: "+text);
   return text;
}

BookContext.speakAndDescend = function(text,end) {
   return {
      _: {
            process: function(element,context) {
               context.revert = context._makeRestore(element);
               context._makeMark(element).exec();
               BookContext.speak(context,text,context.ttsCallback);
               return true;
            },
            onEnd: end ? ( (typeof end=='function') ? end : function() { return end; }) : null,
            canDescend: function(e) { return true; }
         }
   };
};

BookContext.inlines = [
   "a",
   "em",
   "strong",
   "img",
   "sub",
   "sup",
   "span",
   "tt",
   "i",
   "b",
   "big",
   "small",
   "dfn",
   "code",
   "samp",
   "kbd",
   "var",
   "cite",
   "abbr",
   "acronym",
   "br",
   "q",
   "bdo"
];

BookContext.isInline = function(element) {
   var tagName = element.localName;
   for (var i=0; i<BookContext.inlines.length; i++) {
      if (tagName==BookContext.inlines[i]) {
         return true;
      }
   }
   return false;
}


BookContext.prototype.setElement = function(element,options) {
   if (this.revert) {
      this.revert.exec();
      this.revert = null;
   }
   this.referer = options ? options.referer : null;
   this.element = element;
   this.yPos = findYPos(this.element);
   this.descend = this._canDescend(element);
   if (options && options.scroll) {
      var window = this.element.ownerDocument.defaultView;
      window.scrollOwner = this;
      window.scroll(0,this.yPos-DAISYHelper.positionOffset);
   }
   if (options && options.invoke) {
      this._handle();
   }
   this.newPosition = options && options.newPosition ? true : false;
}

BookContext.prototype._canDescend = function(element) {
   var descend = true;
   var handler = this.handlers[element.localName];
   if (handler) {
      var className = element.getAttribute("class");
      var names = className==null ? [] : className.split(/\s+/);
      var classHandler  = null;
      for (var i=0; i<names.length && !classHandler; i++) {
         Console.debug("Class: "+names[i]);
         classHandler = handler[names[i]];
      }
      if (!classHandler) {
         var role = this.element.getAttribute("role");
         Console.debug("Role: "+role);
         classHandler = handler[role];
      }
      if (!classHandler && handler._) {
         classHandler = handler._;
      }
      if (classHandler) {
         descend = classHandler.canDescend(element);
      }
   }
   return descend;
}

BookContext.prototype.handlers = {
   img: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            if (context.tts) {
               var caption = null;
               /* caption handling is disabled per Reuben's request

               var caption = element.nextSibling;
               while (caption && caption.nodeType!=1) {
                  caption = caption.nextSibling;
               }
               if (caption && element.parentNode.className=="daisy-imggroup") {
                  var imgCount = 0;
                  var child = element.parentNode.firstChild;
                  while (child) {
                     if (child.nodeType==1 && child.localName=="img") {
                        imgCount++;
                     }
                     child = child.nextSibling;
                  }
                  if (imgCount>1) {
                     caption = null;
                  }
               }
               var isCaption = false;
               if (caption && caption.localName=="p") {
                  var className = caption.getAttribute("class");
                  var names = className==null ? [] : className.split(" ");
                  var hasSkip = false;
                  for (var i=0; i<names.length; i++) {
                     if (names[i]=="daisy-caption") {
                        isCaption = true;
                     } else if (names[i]=="daisy-skip") {
                        hasSkip = true;
                     }
                  }
                  if (isCaption && !hasSkip) {
                     caption.className = className+" daisy-skip";
                  }
                  if (!isCaption) {
                     caption = null;
                  }
               } else {
                  caption = null;
               }
               */
               if (caption) {
                  Console.debug("caption.localName="+caption.localName+" "+caption.className);
                  BookContext.speak(context,"Image, "+BookContext.inlineText(caption),context.ttsCallback);
               } else {
                  var text = element.getAttribute("alt");
                  if (!text || text=="") {
                     BookContext.speak(context,"Image, no description, image complete",context.ttsCallback);
                  } else {
                     BookContext.speak(context,"Image, "+text+" , image complete",context.ttsCallback);
                  }
               }
            }
            return false;
         },
         canDescend: function(e) { return false }
      }
   },
   math: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            if (context.tts) {
               var text = element.getAttribute("alttext");
               if (!text || text=="") {
                  BookContext.speak(context,"mathematical expression",context.ttsCallback);
               } else {
                  BookContext.speak(context,text,context.ttsCallback);
               }
            }
            return false;
         },
         canDescend: function(e) { return false }
      }
   },
   p: {
      "daisy-pagenum": {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            if (context.tts) {
               var text = element.textContent;
               if (text.indexOf("page")>=0 || text.indexOf("Page")>=0) {
                  BookContext.speak(context,text,context.ttsCallback);
               } else {
                  BookContext.speak(context,"page "+text,context.ttsCallback);
               }
            }
            return false;
         },
         canDescend: function(e) { return false }
      },
      "daisy-p": {
         process: function(element,context) {
            var descend = this.canDescend(element);
            if (descend) {
               var component = null;
               var child = element.firstChild;
               while (child && !component) {
                  if (child.nodeType==1) {
                     component = child;
                  }
                  child = child.nextSibling;
               }
               context.element = component;
               context._handle();
               return context.descend;
            } else {
               context.revert = context._makeRestore(element);
               context._makeMark(element).exec();
               var text = BookContext.inlineText(element);
               BookContext.speak(context,text,context.ttsCallback);
               return false;
            }
         },
         skipOnPrevious: BookContext.canMixedDescend,
         canDescend: BookContext.canMixedDescend
      },
      "daisy-li": {
         process: function(element,context) {
            var component = null;
            var child = element.firstChild;
            while (child && !component) {
               if (child.nodeType==1 && (child.getAttribute("class") || child.localName!="span" && child.getAttribute("class")!="daisy-w")) {
                  component = child;
               }
               child = child.nextSibling;
            }
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            if (component) {
               BookContext.speak(context,"Item",context.ttsCallback);
               return true;
            } else {
               var text = BookContext.inlineText(element);
               BookContext.speak(context,"Item, "+text,context.ttsCallback);
               return false;
            }
         },
         canDescend: function(element) {
            var component = null;
            var child = element.firstChild;
            while (child && !component) {
               if (child.nodeType==1 && (child.getAttribute("class")=="daisy-lic" || child.localName!="span" && child.getAttribute("class")!="daisy-w")) {
                  component = child;
               }
               child = child.nextSibling;
            }
            return component ? true : false
         }
      },
      "daisy-caption": {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            BookContext.speak(context,"Caption, "+BookContext.inlineText(element),context.ttsCallback);
            return false;
         },
         canDescend: function(e) { return false }
      },
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            BookContext.speak(context,BookContext.inlineText(element),context.ttsCallback);
            return false;
         },
         canDescend: function(e) { return false }
      }
   },
   a: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var role = element.getAttribute("role");
            if (!role || role.length==0) {
               role = "link";
            }
            var text = role+", ";

            var title = element.getAttribute("title");
            if (title && title.length>0) {
               if (role=="note") {
                  text = title+", ";
               } else {
                  text += title+", ";
               }
            }
            text += element.textContent;
            BookContext.speak(context,text,context.ttsCallback);
            return false;
         },
         canDescend: function(e) { return false }
      }
   },
   span: {
      "math": {
         process: function(element,context) {
            var math = null;
            var child = element.firstChild;
            while (child && !math) {
               if (child.nodeType==1 && child.localName=="math") {
                  math = child;
               }
            }
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            if (math) {
               var text = math.getAttribute("alttext");
               if (!text || text=="") {
                  BookContext.speak(context,"mathematical expression",context.ttsCallback);
               } else {
                  BookContext.speak(context,text,context.ttsCallback);
               }
            } else {
               BookContext.speak(context,element.textContent,context.ttsCallback);
            }
            return false;
         },
         canDescend: function(element) { return false }

      },
      _: BookContext.skipDescendants
   },
   h1: {
      _: BookContext.skipDescendants
   },
   h2: {
      _: BookContext.skipDescendants
   },
   h3: {
      _: BookContext.skipDescendants
   },
   h4: {
      _: BookContext.skipDescendants
   },
   h5: {
      _: BookContext.skipDescendants
   },
   h6: {
      _: BookContext.skipDescendants
   },
   table: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var caption = null;
            var child = element.firstChild;
            while (!caption && child) {
               if (child.nodeType==1 && child.localName=="caption") {
                  caption = child;
               }
               child = child.nextSibling;
            }
            var text = "table, ";
            if (caption) {
               text += caption.textContent+", ";
            }
            text += "in navigation mode";
            BookContext.speak(context,text,context.ttsCallback);
            context.boundaryCount = 0;
            return true;
         },
         onEnd: function() { return "table complete" },
         pause: function(e,context) { return true},
         afterPause: function(e,context) {
            var cell = context._getTableCell(e);
            if (!cell) {
               Console.debug("Cannot find table cell to announce.");
               return;
            }
            var prefix = cell.localName=="td" ? "cell, " : "column heading, ";
            // TODO: this flattens the cell
            BookContext.speak(context,prefix+cell.textContent,context.ttsCallback);
         },
         canDescend: function(e) { return true; }
      }
   },
   caption: {
      _: BookContext.skipDescendants
   },
   thead: BookContext.speakAndDescend("header","header complete"),
   tfoot: BookContext.speakAndDescend("footer","footer complete"),
   tbody: BookContext.speakAndDescend("body","body complete"),
   tr: BookContext.speakAndDescend("row"),
   th: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            BookContext.speak(context,"column heading, "+element.textContent,context.ttsCallback);
            return false;
         },
         canDescend: function(e) { return false; }
      }
   },
   td: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var descend = this.canDescend(element);
            if (context.tts) {
               if (descend) {
                  BookContext.speak(context,"cell, multiple items",context.ttsCallback);
               } else {
                  BookContext.speak(context,"cell, "+BookContext.inlineText(element),context.ttsCallback);
               }
            }
            return descend;
         },
         canDescend: function(element) {
            // if the table cell is mixed, then we have simple content (inlines)
            // and so we don't descend
            var mixed = false;
            var child = element.firstChild;
            while (child && !mixed) {
               if (child.nodeType!=1) {
                  var parts = child.textContent.split(/\s+/);
                  for (var i=0; !mixed && i<parts.length; i++) {
                     if (parts[i].length>0) {
                        mixed = true;
                     }
                  }
               }
               child = child.nextSibling;
            }
            if (!mixed) {
               // check for all inlines and return true only when there is a block element.
               var child = element.firstChild;
               while (child) {
                  if (child.nodeType==1 && !BookContext.isInline(child)) {
                     return true;
                  }
                  child = child.nextSibling;
               }
               return false;
            }
            return !mixed;
         }
      }
   },
   dl: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            BookContext.speak(context,"definition list",context.ttsCallback);
            return true;
         },
         canDescend: function(e) { return true; }
      }
   },
   dt: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            BookContext.speak(context,"term: "+element.textContent,context.ttsCallback);
            return false;
         },
         canDescend: function(e) { return false; }
      }
   },
   dd: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            BookContext.speak(context,"definition, "+element.textContent,context.ttsCallback);
            return false;
         },
         canDescend: function(e) { return false; }
      }
   },
   ul: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var suffix = "";
            if (element.parentNode.localName=="li") {
               var list = element;
               var level = 0;
               while (list.parentNode.localName=="li") {
                  level++;
                  list = list.parentNode.parentNode;
               }
               suffix = ", level "+level;
            }
            BookContext.speak(context,"list"+suffix,context.ttsCallback);
            return true;
         },
         onEnd: function() { return "list complete" },
         canDescend: function(e) { return true; }
      }
   },
   ol: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var suffix = "";
            if (element.parentNode.localName=="li") {
               var list = element;
               var level = 0;
               while (list.parentNode.localName=="li") {
                  level++;
                  list = list.parentNode.parentNode;
               }
               suffix = ", level "+level;
            }
            BookContext.speak(context,"list"+suffix,context.ttsCallback);
            return true;
         },
         onEnd: function() { return "list complete" },
         canDescend: function(e) { return true; }
      }
   },
   li: {
      _: {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var descend = this.canDescend(element);
            var text = "item";
            if (element.parentNode.localName=="ol") {
               var count = 1;
               var li = element.previousSibling;
               while (li) {
                  if (li.nodeType==1) {
                     count++;
                  }
                  li = li.previousSibling;
               }
               text = count+"";
            }
            if (context.tts) {
               if (descend) {
                  BookContext.speak(context,text,context.ttsCallback);
               } else {
                  BookContext.speak(context,text+", "+BookContext.inlineText(element),context.ttsCallback);
               }
            }
            return descend;
         },
         canDescend: BookContext.canMixedDescend
      }
   },
   blockquote: BookContext.speakAndDescend("block quote","block quote complete"),
   div: {
      "daisy-sidebar": {
         process: function(element,context) {
            context.revert = context._makeRestore(element);
            context._makeMark(element).exec();
            var heading = element.firstChild;
            while (heading && heading.nodeType!=1) {
               heading = heading.nextSibling;
            }
            var suffix = "";
            if (heading) {
               var className = heading.getAttribute("class");
               var names = className==null ? [] : className.split(" ");
               var isHeading = false;
               var hasSkip = false;
               for (var i=0; heading && i<names.length; i++) {
                  if (names[i]=="daisy-hd") {
                     isHeading = true;
                  } else if (names[i]=="daisy-skip") {
                     hasSkip = true;
                  }
               }
               if (isHeading) {
                  suffix = ", "+heading.textContent;
                  if (!hasSkip) {
                     heading.className = className+" daisy-skip";
                  }
               }
            }

            BookContext.speak(context,"sidebar"+suffix,context.ttsCallback);
            return true;
         },
         onEnd: function() { return "sidebar complete" },
         pause: function(e,context) { return context.owner.service.sidebarPause; },
         canDescend: function(e) { return true; }
      },
      "daisy-note": BookContext.speakAndDescend("note",
          function(context,element) {
             if (context.referer) {
                element.style.display = "none";
                context.element = context.referer;
                context.referer = null;
             }
             return "note complete";
          }
       )._,
      "daisy-annotation": BookContext.speakAndDescend("annotation",
          function(context,element) {
             if (context.referer) {
                element.style.display = "none";
                context.element = context.referer;
                context.referer = null;
             }
             return "annotation complete";
          }
      )._,
      "daisy-user-note-ref": BookContext.speakAndDescend("user note reference","user note reference complete")._,
      "daisy-user-note": BookContext.speakAndDescend("user note",
          function(context,element) {
             if (context.referer) {
                element.style.display = "none";
                context.element = context.referer;
                context.referer = null;
             }
             return "user note complete";
          }
      )._,
      "daisy-imggroup": {
         process: function(element,context) {
            var child = element.firstChild;
            var imgCount = 0;
            while (child) {
               if (child.nodeType==1 && child.localName=="img") {
                  imgCount++;
               }
               child = child.nextSibling;
            }
            if (imgCount!=1) {
               context.revert = context._makeRestore(element);
               context._makeMark(element).exec();
               BookContext.speak(context,"image group",context.ttsCallback);
               return true;
            } else {
               context._nextStructural();
               return context.descend;
            }
         },
         onEnd: function(context,e) {
            var child = e.firstChild;
            var imgCount = 0;
            while (child) {
               if (child.nodeType==1 && child.localName=="img") {
                  imgCount++;
               }
               child = child.nextSibling;
            }
            return imgCount==1 ? "image complete" : "image group complete";
         },
         canDescend: function(e) { return true; }
      },
      "daisy-prodnote": BookContext.speakAndDescend("producer's note")._,
      "daisy-list-pl": BookContext.speakAndDescend("list")._,
      _: {
         process: function(element,context) {
            if (element.firstChild) {
               context._nextStructural();
               return context.descend;
            } else {
               BookContext.speak(context,"empty",context.ttsCallback);
               return false;
            }
         },
         skipOnPrevious: function() { return true },
         canDescend: function(e) { return e.firstChild ? true : false; }
      }
   }
}

BookContext.prototype._makeRestore = function(element,action) {
   return {
      className: element.className+"",
      element: element,
      exec: function() {
         Console.debug("Reverting to "+this.className);
         this.element.className = this.className;
         if (action) {
            action();
         }
         return 1;
      }
   };
};


BookContext.prototype._makeMark = function(element) {
   return {
      element: element,
      isNavigation: this.isNavigation,
      exec: function() {
         var window = this.element.ownerDocument.defaultView;
         var pageTop = window.scrollY+DAISYHelper.positionOffset-10;
         if (pageTop<0) {
            pageTop = 0;
         }
         var pageBottom = window.scrollY+window.innerHeight;
         var diff = 0;
         var yPos = findYPos(this.element);
         if (yPos>0) {
            if (this.isNavigation) {
               window.scrollOwner = this;
               window.location.hash = "#"+this.element.getAttribute("id");
               /*
               */
               var current = this;
               setTimeout(function() {
                  yPos -= DAISYHelper.positionOffset;
                  if (yPos<0) {
                     yPos = 0;
                  }
                  window.scrollOwner = current;
                  window.scroll(0,yPos);
               },250);
            } else {
               if (yPos<pageTop) {
                  Console.debug("Scrolling due to keyboard navigation")
                  /*
                  diff = yPos-pageTop;
                  var newScrollY = window.scrollY+diff-DAISYHelper.positionOffset;
                  if (newScrollY<0) {
                     newScrollY = 0;
                  }*/
                  yPos -= DAISYHelper.positionOffset;
                  window.scrollOwner = this;
                  window.scroll(0,yPos);
               } else if ((yPos+this.element.offsetHeight+30)>pageBottom) {
                  Console.debug("Scrolling as content is below vierport bottom")
                  diff = yPos+this.element.offsetHeight-pageBottom;
                  window.scrollOwner = this;
                  window.scroll(0,window.scrollY+diff+DAISYHelper.positionOffset);
               }
            }
         }
         //Console.log(this.element.className);
         this.element.className = this.element.className + " daisy-highlight";
         return -1;
      }
   };
};

BookContext.prototype._getHandler = function(e) {
   var handler = this.handlers[e.localName];
   if (handler) {
      var className = e.getAttribute("class");
      var names = className==null ? [] : className.split(" ");
      var classHandler  = null;
      for (var i=0; i<names.length && !classHandler; i++) {
         Console.debug("Class: "+names[i]);
         classHandler = handler[names[i]];
      }
      if (!classHandler) {
         var role = this.element.getAttribute("role");
         Console.debug("Role: "+role);
         classHandler = handler[role];
      }
      if (!classHandler && handler._) {
         classHandler = handler._;
      }
      return classHandler;
   }
   return null;
}

BookContext.prototype._handle = function() {
   this.yPos = findYPos(this.element);
   this.element.focus();
   var classHandler = this._getHandler(this.element);
   var handled = false;
   if (classHandler) {
      Console.debug("Using class handler for "+this.element.localName);
      handled = true;
      this.descend = classHandler.process(this.element,this);
   } else {
      Console.debug("No handler for "+this.element.localName);
   }
   if (!handled) {
      Console.debug("Using default semantics for "+this.element.localName);
      this.revert = this._makeRestore(this.element);
      this._makeMark(this.element).exec();
      BookContext.speak(this,this.element.textContent,this.ttsCallback);
   }
   this.handler = classHandler;
   Console.debug("descend="+this.descend);
}

BookContext.prototype._canDisplay = function(element) {
   return (element.nodeType==1 && element.offsetHeight>0 && element.localName!="br" && element.className.indexOf("daisy-skip")<0);
}

BookContext.prototype.nextWord = function(nostop) {
   if (!nostop) {
      this.stop();
   }
}

BookContext.prototype.previousWord = function(nostop) {
   if (!nostop) {
      this.stop();
   }
}

BookContext.prototype.reset = function() {
   if (this.revert) {
      this.revert.exec();
      this.revert = null;
   }
}

BookContext.prototype.stop = function() {}

BookContext.prototype._getTableCell = function(e) {
   var cell = null;
   if (e.localName=="table") {
      var row = e.firstChild;
      //while (row && (row.nodeType!=1 || (row.localName!="tr" && row.localName!="thead" && row.localName!="tfoot" && row.localName!="tbody"))) {
      while (row && (row.nodeType!=1 || (row.localName!="tr" && row.localName!="tbody"))) {
         row = row.nextSibling;
      }
      if (row && row.localName!="tr") {
         row = row.firstChild;
         while (row && (row.nodeType!=1 || row.localName!="tr")) {
            row = row.nextSibling;
         }
      }
      if (row) {
         e = row;
      }
   }

   if (e.localName=="tr") {
      cell = e.firstChild;
      while (cell && (cell.nodeType!=1 || (cell.localName!="td" && cell.localName!="th"))) {
         cell = cell.nextSibling;
      }
   } else {
      cell = e;
      while (cell && cell.localName!="td" && cell.localName!="th") {
         //Console.debug(cell.localName);
         cell = cell.parentNode;
         if (cell.nodeType!=1) {
            cell = null;
         }
      }
   }
   return cell;
}

BookContext.prototype._getCellPosition = function(cell) {
   var pos = 0;
   while (cell) {
      var colspan = cell.getAttribute("colspan");
      pos += colspan ? parseInt(colspan) : 1;
      cell = cell.previousSibling;
      while (cell && cell.nodeType!=1) {
         cell = cell.previousSibling;
      }
   }
   return pos;
}

BookContext.prototype.currentLocation = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   if (!this.element) {
      return;
   }

   if (this.tts) {
      this.tts.cancel();
   }
   if (this.revert) {
      this.revert.exec();
      this.revert = null;
   }

   var cell = this._getTableCell(this.element);
   if (cell) {
      this.setElement(cell,{ scroll: true, invoke: true });
   } else {
      this.setElement(this.element,{ scroll: true, invoke: true });
   }

}

BookContext.prototype._playNoLink = function(oncomplete) {
   if (this.soundManager) {
      var sound = this.soundManager.createSound({
         id: 'playback-no-link',
         url: this.sounds['no-link'],
         onfinish: function() {
            if (oncomplete) {
               oncomplete();
            }
         },
         autoPlay: false
      });
      sound.load();
      var waitFor = function() {
         if (sound.readyState==3) {
            sound.play();
         } else {
            setTimeout(waitFor,10);
         }
      }
      waitFor();
   }
}

BookContext.prototype._playTableBoundary = function(oncomplete) {
   if (this.soundManager) {
      var sound = this.soundManager.createSound({
         id: 'playback-table-boundary',
         url: this.sounds['table-boundary'],
         onfinish: function() {
            if (oncomplete) {
               oncomplete();
            }
         },
         autoPlay: false
      });
      sound.load();
      var waitFor = function() {
         if (sound.readyState==3) {
            sound.play();
         } else {
            setTimeout(waitFor,10);
         }
      }
      waitFor();
   }
}

BookContext.makeTableBoundarySkip = function(context) {
   return function() {
      Console.debug("context.boundaryCount="+context.boundaryCount);
      if (context.boundaryCount>=3) {
         context.boundaryCount = 0;
         Console.debug("Skipping table due to boundary hit 3 times.");
         var cell = context._getTableCell(context.element);
         var table = cell.parentNode.parentNode;
         if (table.localName!="table") {
            table = table.parentNode;
         }
         context.element = table;
         BookContext.speak(context,"table complete",{
            QueryInterface: function(aIID) {
               if (aIID.equals(Components.interfaces.nsITTSCallback) ||
                  aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
                  aIID.equals(Components.interfaces.nsISupports)) {
                  return this;
               }
               throw Components.results.NS_NOINTERFACE;
            },
            onFinish: function(cancelled) {
               context._skip({ next: true, outOfContainer: true});
            }
         })
      }
   }
}


BookContext.prototype.nextTableCell = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   if (!this.element) {
      return;
   }

   var cell = this._getTableCell(this.element);
   if (!cell) {
      return;
   }

   Console.debug("In table cell, calculating next cell");

   cell = cell.nextSibling;
   while (cell && cell.nodeType!=1) {
      cell = cell.nextSibling;
   }

   if (cell) {
      this.setElement(cell,{ scroll: true, invoke: true });
   } else {
      this.boundaryCount++;
      this._playTableBoundary(BookContext.makeTableBoundarySkip(this));
   }

}

BookContext.prototype.previousTableCell = function() {
   if (!this.element) {
      return;
   }

   var cell = this._getTableCell(this.element);
   if (!cell) {
      return;
   }

   Console.debug("In table cell, calculating previous cell");

   cell = cell.previousSibling;
   while (cell && cell.nodeType!=1) {
      cell = cell.previousSibling;
   }

   if (cell) {
      this.setElement(cell,{ scroll: true, invoke: true });
   } else {
      this.boundaryCount++;
      this._playTableBoundary(BookContext.makeTableBoundarySkip(this));
   }

}

BookContext.prototype.beneathTableCell = function(nostop) {
   if (!nostop) {
      this.stop();
   }

   if (!this.element) {
      return;
   }

   var cell = this._getTableCell(this.element);
   if (!cell) {
      return;
   }

   Console.debug("In table cell, calculating beneath cell");

   var row = cell.parentNode;

   row = row.nextSibling;
   while (row && row.nodeType!=1) {
      row = row.nextSibling;
   }

   // if there is no next table row, find the next row in the next table part
   if (!row) {
      var seek = null;
      if (cell.parentNode.parentNode.localName=="thead") {
         seek = "tbody";
      } else if (cell.parentNode.parentNode.localName=="tbody") {
         seek = "tfoot";
      }
      if (seek) {
         // find the table part
         var table = cell.parentNode.parentNode.parentNode;
         var child = table.firstChild;
         while (!row && child) {
            if (child.nodeType==1 && child.localName==seek) {
               row = child;
            }
            child = child.nextSibling;
         }
         if (row) {
            // find the first row
            row = row.firstChild;
            while (row && (row.nodeType!=1 || row.localName!="tr")) {
               row = row.nextSibling;
            }
         }
      }
   }
   if (!row) {
      this.boundaryCount++;
      this._playTableBoundary(BookContext.makeTableBoundarySkip(this));
      return;
   }

   var currentPos = this._getCellPosition(cell);
   var pos = 0;
   var beneath = row.firstChild;
   while (beneath && beneath.nodeType!=1) {
      beneath = beneath.nextSibling;
   }
   while (beneath && pos<currentPos) {
      var colspan = beneath.getAttribute("colspan");
      pos += colspan ? parseInt(colspan) : 1;
      if (pos<currentPos) {
         beneath = beneath.nextSibling;
         while (beneath && beneath.nodeType!=1) {
            beneath = beneath.nextSibling;
         }
      }
   }

   if (beneath && pos>=currentPos) {
      this.setElement(beneath,{ scroll: true, invoke: true });
   }

}

BookContext.prototype.aboveTableCell = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   if (!this.element) {
      return;
   }

   var cell = this._getTableCell(this.element);
   if (!cell) {
      return;
   }

   Console.debug("In table cell, calculating above cell");

   var row = cell.parentNode;

   row = row.previousSibling;
   while (row && row.nodeType!=1) {
      row = row.previousSibling;
   }

   if (!row) {
      var seek = null;
      if (cell.parentNode.parentNode.localName=="tbody") {
         seek = "thead";
      } else if (cell.parentNode.parentNode.localName=="tfoot") {
         seek = "tbody";
      }
      if (seek) {
         // find the table part
         var table = cell.parentNode.parentNode.parentNode;
         var child = table.firstChild;
         while (!row && child) {
            if (child.nodeType==1 && child.localName==seek) {
               row = child;
            }
            child = child.nextSibling;
         }
         if (row) {
            // find the last row
            row = row.lastChild;
            while (row && (row.nodeType!=1 || row.localName!="tr")) {
               row = row.previousSibling;
            }
         }
      }
   }

   if (!row) {
      this.boundaryCount++;
      //Console.debug("boundaryCount="+this.boundaryCount);
      this._playTableBoundary(BookContext.makeTableBoundarySkip(this));
      return;
   }

   var currentPos = this._getCellPosition(cell);
   var pos = 0;
   var above = row.firstChild;
   while (above && above.nodeType!=1) {
      above = above.nextSibling;
   }
   while (above && pos<currentPos) {
      var colspan = above.getAttribute("colspan");
      pos += colspan ? parseInt(colspan) : 1;
      if (pos<currentPos) {
         above = above.nextSibling;
         while (above && above.nodeType!=1) {
            above = above.nextSibling;
         }
      }
   }

   if (above && pos>=currentPos) {
      this.setElement(above,{ scroll: true, invoke: true });
   }

}

BookContext.prototype.skip = function(options) {
   var current = this;
   var isPlaying = this.playing;
   this.playing = false;

   if (options && options.nostop) {
      this.stop();
   }
   if (this.tts) {
      Console.debug("Cancelling TTS.");
      this.tts.cancel();
   }
   if (this.revert) {
      this.revert.exec();
      this.revert = null;
   }

   this.skipTimer = setTimeout(function() {
      current.skipTimer = null;
      options.play = options.play || isPlaying;
      Console.debug("skip play: "+options.play);
      current._skip(options);
   },100);
}
BookContext.prototype._skip = function(options) {
   if (!options || options.next) {
      if (options && options.outOfContainer || this.element.localName=="td" || this.element.localName=="th" || this.element.localName=="tr") {
         var container = null;
         if (this.element.parentNode.localName=="p") {
            container = this.element.parentNode;
         } else {
            var current = this.element;
            while (current!=this.top && !container) {
               if (current.localName=="ol" || current.localName=="ul" || current.localName=="table" || current.localName=="div") {
                  container = current;
               } else {
                  current = current.parentNode;
               }
            }
         }
         if (container) {
            this.element = container;
            this.descend = false;
            if (options.play) {
               this.paused = true;
               this.togglePlay();
            } else {
               this._nextStructural();
            }
            return;
         }
      }
      if (this.descend) {
         this.descend = false;
         if (options && options.play) {
            this.togglePlay();
         } else {
            this._nextStructural();
         }
      } else {
         this.element = this.element.parentNode;
         this.descend = false;
         if (options && options.play) {
            this.paused = true;
            this.togglePlay();
         } else {
            this._nextStructural();
         }
      }
   } else if (options.previous) {
      if (this.descend) {
         var prev = this.element.previousSibling;
         while (prev && prev.nodeType!=1) {
            prev = prev.previousSibling;
         }
         this.element = prev ? prev : this.element.parentNode;
         if (prev) {
            this.newPosition = true;
         }
      } else {
         this.element = this.element.parentNode;
      }
      this.previousStructural();
   }
}

BookContext.prototype.nextStructural = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   if (this.tts) {
      Console.debug("Cancelling TTS.");
      this.tts.cancel();
   }
   if (this.revert) {
      this.revert.exec();
      this.revert = null;
   }
   this._nextStructural();
}

BookContext.prototype._nextStructural = function() {
   if (!this.newPosition) {
      Console.debug("nextStructural at "+this.locator(this.element)+", descend="+this.descend);
      // find next element in tree
      this.something = false;
      var next = this.descend ? this.element.firstChild : null;
      var last = this.element;
      while (!next || next.nodeType!=1 || !this._canDisplay(next)) {
         Console.debug("seek at "+this.locator(this.element));
         if (!next) {
            // no children, move to next sibling amongst ancestry
            while (!this.element.nextSibling && this.element!=this.top) {
               this.element = this.element.parentNode;
               // check for on end text
               var classHandler = this._getHandler(this.element);
               if (classHandler.onEnd) {
                  var text = classHandler.onEnd(this,this.element);
                  if (text) {
                     BookContext.speak(this,text,this.ttsCallback);
                     this.descend = false;
                     Console.debug("nextStructural, onEnd at "+this.locator(this.element)+", descend="+this.descend);
                     return;
                  }
               }
            }
            next = this.element.nextSibling;
         }
         while (next && (next.nodeType!=1 || !this._canDisplay(next))) {
            next = next.nextSibling;
         }
         if (!next && this.element!=this.top) {
            Console.debug("No next child element, seeking to sibling");
            // ran out of child siblings (all non-elements) so move to next sibling sibling
            next = this.element.nextSibling;
            while (next && (next.nodeType!=1 || !this._canDisplay(next))) {
               next = next.nextSibling;
            }
         }
         if (!next && this.element!=this.top) {
            Console.debug("No next sibling element, moving to parent");
            // ran out of element siblings so move to parent
            this.element = this.element.parentNode;
            // check for on end text
            var classHandler = this._getHandler(this.element);
            if (classHandler.onEnd) {
               var text = classHandler.onEnd(this,this.element);
               if (text) {
                  BookContext.speak(this,text,this.ttsCallback);
                  this.descend = false;
                  Console.debug("nextStructural, onEnd at "+this.locator(this.element)+", descend="+this.descend);
                  return;
               }
            }
         }
         if (!next && this.element==this.top) {
            // there is no next so we're at the top
            break;
         }
      }
      this.descend = true;
      if (next) {
         Console.debug("current: "+this.element.parentNode.parentNode.localName+" > "+this.element.parentNode.localName+" > "+this.element.localName);
         Console.debug("next: "+next.parentNode.parentNode.localName+" > "+next.parentNode.localName+" > "+next.localName);
         if ((this.element.localName=="thead" || this.element.parentNode.localName=="thead" || this.element.parentNode.parentNode.localName=="thead") && 
             (next.localName=="tfoot" || next.parentNode.localName=="tfoot" || next.parentNode.parentNode.localName=="tfoot")) {
            var footer = next;
            next = next.nextSibling;
            while (next && next.nodeType!=1) {
               next = next.nextSibling;
            }
            this.element = next ? next : footer;
            if (next) {
               this.newPosition = true;
               this._nextStructural()
               return;
            }
         }
         this.something = true;
         this.element = next;
         /* skip this feature till ctrl-alt-k works
      } else if (this.something) {
         // we've hit the end and so start over
         this.element = this.top;
         Console.log("")
         this.nextStructural();
         return;
         */
      } else if (!this.something && this.element==this.top) {
         // do not move if at the end.
         this.element = last;
         BookContext.speak(this,"End of Book",this.ttsCallback);
         this.stop();
         return;
      } else {
         return;
      }
   } else {
      this.newPosition = false;
   }
   var point = this._findNextNavigation();
   this.isNavigation = point && point.content.id==this.element.getAttribute("id");
   Console.debug("nextStructure at "+this.locator(this.element)+", point="+(point ? point.content.id : "")+", isNavigation="+this.isNavigation);
   this._handle();
}

BookContext.prototype.previousStructural = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   if (this.tts) {
      Console.debug("Cancelling TTS.");
      this.tts.cancel();
   }
   if (this.revert) {
      this.revert.exec();
      this.revert = null;
   }
   if (!this.newPosition) {
      Console.debug("previousStructural at "+this.locator(this.element)+", descend="+this.descend);
      // find next element in tree
      var prev = null;
      var movedToParent = false;
      var last = this.element;
      while (!prev || prev.nodeType!=1 || !this._canDisplay(prev)) {
         movedToParent = false;
         Console.debug("seek at "+this.locator(this.element));
         if (!prev) {
            if (!this.element.previousSibling) {
               // no children, move to next sibling amongst ancestry
               while (!this.element.previousSibling && this.element!=this.top) {
                  this.element = this.element.parentNode;
               }
               var classHandler = this._getHandler(this.element);
               Console.debug("Moved to parent "+this.locator(this.element)+", last="+this.locator(last));
               movedToParent = true;

               while (!prev && classHandler.skipOnPrevious && classHandler.skipOnPrevious(this.element) && this.element!=this.top) {
                  prev = this.element.previousSibling;
                  while (prev && prev.nodeType!=1) {
                     prev = prev.previousSibling;
                  }
                  if (prev) {
                     Console.debug("found new previous "+this.locator(prev)+", last="+this.locator(last));
                     movedToParent = false;
                  } else {
                     movedToParent = true;
                     this.element = this.element.parentNode;
                     classHandler = this._getHandler(this.element);
                     Console.debug("Moved to parent "+this.locator(this.element));
                  }
               }
               if (!prev) {
                  prev = this.element;
               }
               if (prev.firstChild==last && this.element!=this.top) {
                  Console.debug("Looping to first child "+this.locator(last)+" detected, moving to ancestor.");
                  prev = prev.previousSibling;
                  while (!prev || prev.nodeType!=1) {
                     if (prev) {
                        prev = prev.previousSibling;
                     } else if (this.element.parentNode!=this.top) {
                        prev = this.element.parentNode;
                     }
                  }
                  movedToParent = false;
               }
            } else {
               prev = this.element.previousSibling;
            }
            Console.debug("No previous, prev set to "+this.locator(prev)+", type="+prev.nodeType+", can display="+this._canDisplay(prev));
         }
         while (prev && (prev.nodeType!=1 || !this._canDisplay(prev))) {
            prev = prev.previousSibling;
         }
         Console.debug("prev set to "+this.locator(prev));
         if (!prev && this.element!=this.top) {
            // no previous element sibling, move to parent
            this.element = this.element.parentNode;
            var classHandler = this._getHandler(this.element);
            Console.debug("Moved to parent "+this.locator(this.element));
            movedToParent = true;

            while (!prev && classHandler.skipOnPrevious && classHandler.skipOnPrevious(this.element) && this.element!=this.top) {
               prev = this.element.previousSibling;
               while (prev && prev.nodeType!=1) {
                  prev = prev.previousSibling;
               }
               if (prev) {
                  movedToParent = false;
               } else {
                  movedToParent = true;
                  this.element = this.element.parentNode;
                  classHandler = this._getHandler(this.element);
                  Console.debug("Moved to parent "+this.locator(this.element));
               }
            }
            if (!prev) {
               prev = this.element;
            }
         }
         if (!prev && this.element==this.top) {
            // there is no next so we're at the top
            break;
         }
      }
      this.descend = true;
      if (prev) {
         this.something = true;
         this.element = prev;
         if (this.element==this.top) {
            BookContext.speak(this,"Start of Book",this.ttsCallback);
            return;
         }
      } else if (this.something) {
         this.element = this.top;
         BookContext.speak(this,"Start of Book",this.ttsCallback);
         this.newPosition = true;
         //this._nextStructural();
         return;
      } else {
         return;
      }
      if (!movedToParent) {
         this.element = this._findLastDescendant(this.element);
      }
   } else {
      this.newPosition = false;
   }
   var point = this._findPreviousNavigation();
   this.isNavigation = point && point.content.id==this.element.getAttribute("id");
   Console.debug("previousStructural at "+this.locator(this.element)+", point="+(point ? point.content.id : "")+", isNavigation="+this.isNavigation);
   this._handle();
}

BookContext.prototype._findLastDescendant = function(element) {
   if (this._canDescend(element)) {
      // move to the last element in the tree
      var right = null;
      var next = element.firstChild;
      var last = null;
      while (next && !right) {
         // find the last element sibling
         var lastSibling = next.nodeType==1 ? next : null;
         while (next.nextSibling) {
            next = next.nextSibling;
            if (next.nodeType==1 && this._canDisplay(next)) {
               lastSibling = next;
            }
         }
         if (lastSibling) {
            if (this._canDescend(lastSibling)) {
               // descend into the last sibling
               last = lastSibling;
               next = lastSibling.firstChild;
            } else {
               // the last sibling is final right most element as we can't descend
               right = lastSibling;
            }
         } else if (last) {
            // set this.owner to the last found element in the tree
            right = last;
         } else {
            // we have no right most element so break
            break;
         }
      }
      return right ? right : element;
   } else {
      return element;
   }
}

BookContext.prototype._seekToLastDescendant = function(element) {
   // move to the last element in the tree
   var right = null;
   var next = element.lastChild;
   var last = null;
   while (next && !right) {
      // find the last element sibling
      var lastSibling = next;
      while (lastSibling && lastSibling.style && lastSibling.style.display=="none") {
         lastSibling = lastSibling.previousSibling;
      }
      if (lastSibling) {
         if (lastSibling.nodeType==1) {
            // descend into the last sibling
            last = lastSibling;
            next = lastSibling.lastChild;
         } else {
            // the last sibling is final right most element as we can't descend
            right = lastSibling;
         }
      } else if (last) {
         // set this.owner to the last found element in the tree
         right = last;
      } else {
         // we have no right most element so break
         break;
      }
   }
   return right ? right : element;
}


BookContext.prototype._findNextNavigation = function() {
   var point = this.owner.context.currentPoint;
   if (point) {
      if (point.children.length>0) {
         // move to first child if there is a child
         point = point.children[0];
      } else {
         // otherwise, move to first next sibling of an ancestor or self
         while (!point.next && point.parent) {
            point = point.parent;
         }
         if (point.next) {
            point = point.next;
         } else {
            Console.debug("End of navigation.");
            return null;
         }
      }
      return point;
   } else {
      return null;
   }
}

BookContext.prototype.nextNavigation = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   this.owner.context.currentPoint = this._findNextNavigation();
   if (this.owner.context.currentPoint) {
      Console.debug("Moved to navigation point "+this.owner.context.currentPoint.content.id);
      this.owner.window.location.hash = "#"+this.owner.context.currentPoint.content.id;
      var target = this.owner.window.document.getElementById(this.owner.context.currentPoint.content.id);
      if (target) {
         //target = this.owner.seekNavigationStart(target);
         this.setElement(target,{ scroll: true, invoke: true });
         //this.nextStructural();
      }
   } else {
      Console.debug("No current point for next.");
   }
}

BookContext.prototype._findPreviousNavigation = function() {
   var point = this.owner.context.currentPoint;
   if (!point) {
      return null;
   }
   // otherwise, move to first previous sibling of an ancestor or self
   while (!point.previous && point.parent) {
      point = point.parent;
   }
   if (point.previous) {
      // move to the previous sibling
      point = point.previous;
      // descend to right most descendant
      while (point.children.length>0) {
         point = point.children[0];
         while (point.next) {
            point = point.next;
         }
      }
   } else {
      point = null;
   }
   return point;

}

BookContext.prototype.previousNavigation = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   this.owner.context.currentPoint = this._findPreviousNavigation();
   if (this.owner.context.currentPoint) {
      Console.debug("Moved to navigation point "+this.owner.context.currentPoint.content.id);
      this.owner.window.location.hash = "#"+this.owner.context.currentPoint.content.id;
      var target = this.owner.window.document.getElementById(this.owner.context.currentPoint.content.id);
      if (target) {
         //target = this.owner.seekNavigationStart(target);
         this.setElement(target,{ scroll: true, invoke: true });
         //this.nextStructural();
      }
   } else {
      Console.debug("No current point for previous.");
   }
}

BookContext.prototype.nextLink = function(nostop) {
   if (!nostop) {
      this.stop();
   }
   if (!this.element) {
      return;
   }
   var current = this.element.firstChild ? this.element.firstChild : this.element.nextSibling;
   if (!current) {
      // there is no next sibling, seek to ancestor's next sibling
      current = this.element.parentNode;
      while (!current.nextSibling && current!=this.top) {
         current = current.parentNode;
      }
      if (current!=this.top) {
         current = current.nextSibling;
      }
   }
   // find the next link from the current position
   while (current && current!=this.top) {
      if (current.nodeType==1 && current.localName=="a") {
         if (current.className!=DAISYHelper.hackClassName) {
            this.setElement(current,{ scroll: true, invoke: true });
            return;
         }
      }
      if (current.firstChild) {
         // descend to first child
         current = current.firstChild;
      } else if (current.nextSibling) {
         // move to sibling
         current = current.nextSibling;
      } else {
         // find ancestor's next sibling
         current = current.parentNode;
         while (!current.nextSibling && current!=this.top) {
            current = current.parentNode;
         }
         if (current!=this.top) {
            current = current.nextSibling;
         }
      }
   }
   this._playNoLink();
}

BookContext.prototype.previousLink = function(nostop) {
   Console.debug("Finding previous link:");
   if (!nostop) {
      this.stop();
   }
   if (!this.element) {
      Console.debug("previousLink: no element");
      return;
   }
   var current = this.element;
   // find ancestor or self previous sibling
   while (!current.previousSibling && current!=this.top) {
      //Console.debug("previousLink at "+this.locator(current)+", moving to parent");
      current = current.parentNode;
   }
   if (current!=this.top) {
      current = current.previousSibling;
      current = this._seekToLastDescendant(current);
   } else {
      //Console.debug("previousLink: at top, no previous link");
      this._playNoLink();
      return;
   }
   //Console.debug("previousLink at "+this.locator(current));

   // find the previous link from the current position
   while (current && current!=this.top) {
      //Console.debug("previousLink at "+this.locator(current));
      if (current.nodeType==1 && current.localName=="a") {
         if (current.className!=DAISYHelper.hackClassName) {
            this.setElement(current,{ scroll: true, invoke: true });
            return;
         }
      }
      if (current.previousSibling) {
         // move to sibling
         //Console.debug("previousLink previous sibling of "+this.locator(current));
         current = current.previousSibling;
         //Console.debug("previousLink finding last descendant of "+this.locator(current));
         current = this._seekToLastDescendant(current);
      } else {
         // find ancestor's next sibling
         current = current.parentNode;
         //Console.debug("previousLink moving to parent, "+this.locator(current));
         if (current.nodeType==1 && current.localName=="a") {
            this.setElement(current,{ scroll: true, invoke: true });
            return;
         }
         while (!current.previousSibling && current!=this.top) {
            current = current.parentNode;
            //Console.debug("previousLink moving to parent, "+this.locator(current));
            if (current.nodeType==1 && current.localName=="a") {
               this.setElement(current,{ scroll: true, invoke: true });
               return;
            }
         }
         if (current!=this.top) {
            //Console.debug("previousLink previous sibling of "+this.locator(current));
            current = current.previousSibling;
            //Console.debug("previousLink finding last descendant of "+this.locator(current));
            current = this._seekToLastDescendant(current);
         }
      }
   }
   this._playNoLink();
}


SMILPlayback.prototype = new BookContext();
SMILPlayback.prototype.constructor = SMILPlayback;
function SMILPlayback() {
   this.tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   var current = this;
   this.ttsCallback = {
      thread: current.tm.currentThread,
      QueryInterface: function(aIID) {
         if (aIID.equals(Components.interfaces.nsITTSCallback) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports)) {
            return this;
         }
         throw Components.results.NS_NOINTERFACE;
      },
      onFinish: function(cancelled) {
         if (Console._debug) {
            this.thread.dispatch({
               run: function() {
                  Console.debug("Speech complete, playing="+current.playing);
               }
            },this.thread.DISPATCH_NORMAL);
         }
         this.thread.dispatch({
            run: function() {
               if (!cancelled) {
                  current.reset();
               }
            }
         },this.thread.DISPATCH_NORMAL);
      }
   };
}

SMILPlayback.prototype.lockTTS = function() {
   var current = this;
   var lock = this.owner.service.getLockFor("tts");
   this.ownedLock = lock.acquire({
      exclusive: true,
      onRelinquish: function() {
         current.ownedLock = null;
         current.stop();
      }
   });
}

SMILPlayback.prototype.togglePlay = function() {
   if (this.owner.context.playContext) {
      this.playing = false;
      this.owner.context.stop();
      this.owner.togglePlayButton(true);
      if (this.owner.context.options.context.element) {
         this.setElement(this.context.options.context.element);
      }
      if (this.ownedLock) {
         this.ownedLock.release();
      }
   } else {
      this.reset();
      this.lockTTS();
      try {
         this.owner.context.seek(this.element.getAttribute("id"));
         if (this.owner.context.playContext) {
            this.playing = true;
            this.owner.togglePlayButton(false);
         } else {
            // The current item failed to play, so play the next item
            this.owner.onNext();
         }
      } catch (ex) {
         Console.log(ex);
      }
   }
}

SMILPlayback.prototype.next = function() {
   if (this.owner.context.currentPoint) {
      Console.debug("Next called at "+this.owner.context.currentPoint.content.id+", next="+this.owner.context.currentPoint.next);
   }
   this.owner.context.stop();
   this.reset();
   this.lockTTS();

   var current = this.owner;
   var tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   var operation = {
      run: function() {
         current.context.next();
         if (current.context.currentPoint) {
            Console.debug("After next at "+current.context.currentPoint.content.id);
         }
         if (current.context.playContext) {
            current.togglePlayButton(false);
         } else {
            current.togglePlayButton(true);
            if (current.context.currentPoint) {
               // the play failed because the target could not be found, since
               // there is a context, try again
               current.timer = setTimeout(function() {
                  tm.currentThread.dispatch(operation,tm.currentThread.DISPATCH_NORMAL);
               },250);
            }
         }
      }
   }
   current.timer = setTimeout(function() {
      tm.currentThread.dispatch(operation,tm.currentThread.DISPATCH_NORMAL);
   },250);
   
}

SMILPlayback.prototype.previous = function() {
   this.owner.context.stop();
   this.reset();
   this.lockTTS();

   var current = this.owner;
   var tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   var operation = {
      run: function() {
         current.context.previous();
         if (current.context.playContext) {
            current.togglePlayButton(false);
         } else {
            current.togglePlayButton(true);
            if (current.context.currentPoint) {
               // the play failed because the target could not be found, since
               // there is a context, try again
               current.timer = setTimeout(function() {
                  tm.currentThread.dispatch(operation,tm.currentThread.DISPATCH_NORMAL);
               },250);
            }
         }
      }
   }
   current.timer = setTimeout(function() {
      tm.currentThread.dispatch(operation,tm.currentThread.DISPATCH_NORMAL);
   },250);
}

SMILPlayback.prototype.stop = function() {
   this.owner.context.stop();
   this.owner.togglePlayButton(true);
}


StructuralPlayback.prototype = new BookContext();
StructuralPlayback.prototype.constructor = StructuralPlayback;
function StructuralPlayback() {
   this.tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   this.playing = false;
   var current = this;
   this.ttsCallback = {
      thread: current.tm.currentThread,
      QueryInterface: function(aIID) {
         if (aIID.equals(Components.interfaces.nsITTSCallback) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports)) {
            return this;
         }
         throw Components.results.NS_NOINTERFACE;
      },
      onFinish: function(cancelled) {
         if (Console._debug) {
            this.thread.dispatch({
               run: function() {
                  Console.debug("Speech complete, playing="+current.playing);
               }
            },this.thread.DISPATCH_NORMAL);
         }
         if (current.playing) {
            var dispatch = function() {
               current.pauseTimer = null;
               current.nextOperation();
            };
            if (current.owner.service.playbackPause>0) {
               try {
                  this.thread.dispatch({
                     run: function() {
                        Console.debug("Pause: "+current.owner.service.playbackPause);
                        current.pauseTimer = setTimeout(dispatch,current.owner.service.playbackPause);
                     }
                  },this.thread.DISPATCH_NORMAL);
               } catch (ex) {
                  dispatch();
               }
            } else {
               this.thread.dispatch({
                  run: function() {
                     dispatch();
                  }
               },this.thread.DISPATCH_NORMAL);
            }
         } else {
            this.thread.dispatch({
               run: function() {
                  if (!cancelled) {
                     current.reset();
                  }
                  Console.debug("No longer playing, playback stopped.");
               }
            },this.thread.DISPATCH_NORMAL);
         }
      }
   };
}

StructuralPlayback.prototype.nextOperation = function() {
   var current = this;
   var operation = {
      run: function() {
         Console.debug("nextOperation(), current.playing="+current.playing);
         if (current.playing) {
            if (!current.paused && current.handler && current.handler.pause && current.handler.pause(current.element,current)) {
               current.reset();
               current.playing = false;
               current.paused = true;
               current.owner.togglePlayButton(true);
               if (current.soundManager) {
                  var sound = current.soundManager.createSound({
                     id: 'playback-auto-pause',
                     url: current.sounds['auto-pause'],
                     onfinish: function() {
                        Console.debug("auto-pause sound finished.");
                        if (current.handler.afterPause) {
                           current.handler.afterPause(current.element,current);
                        }
                     },
                     autoPlay: false
                  });
                  sound.load();
                  var waitFor = function() {
                     if (sound.readyState==3) {
                        sound.play();
                     } else {
                        setTimeout(waitFor,10);
                     }
                  }
                  waitFor();
               } else {
                  if (current.handler.afterPause) {
                     current.handler.afterPause(current.element,current);
                  }
               }
               return;
            }
            current.paused = false;
            current.nextStructural(true);
         } else {
            current.reset();
         }
      }
   };
   this.tm.currentThread.dispatch(operation,this.tm.currentThread.DISPATCH_NORMAL);
}

StructuralPlayback.prototype.togglePlay = function() {
   if (this.playing) {
      Console.debug("Pausing");
      this.playing = false;
      this.tts.cancel();
      this.owner.togglePlayButton(true);
      if (this.ownedLock) {
         this.ownedLock.release();
      }
   } else {
      Console.debug("Resuming");
      var current = this;
      var lock = this.owner.service.getLockFor("tts");
      this.ownedLock = lock.acquire({
         exclusive: true,
         onRelinquish: function() {
            current.stop();
         }
      });
      this.reset();
      this.owner.togglePlayButton(false);
      this.playing = true;
      if (!this.paused) {
         this.newPosition = true;
      }
      this.nextOperation();
   }
}

StructuralPlayback.prototype.next = function() {
   this.stop();
   this.reset();
   var current = this;
   this.nextTimer = setTimeout(function() {
      current.nextTimer = null;
      current.owner.context.currentPoint = current._findNextNavigation();
      if (current.owner.context.currentPoint) {
         Console.debug("Moved to navigation point "+current.owner.context.currentPoint.content.id);
         current.owner.window.location.hash = "#"+current.owner.context.currentPoint.content.id;
         var target = current.owner.window.document.getElementById(current.owner.context.currentPoint.content.id);
         if (target) {
            current.setElement(target,{ scroll: true, newPosition: true });
         }
         current.togglePlay();
      } else {
         Console.debug("No current point for next.");
      }
   },500);
}

StructuralPlayback.prototype.previous = function() {
   this.stop();
   this.reset();
   var current = this;
   this.prevTimer = setTimeout(function() {
      current.prevTimer = null;
      current.owner.context.currentPoint = current._findPreviousNavigation();
      if (current.owner.context.currentPoint) {
         Console.debug("Moved to navigation point "+current.owner.context.currentPoint.content.id);
         current.owner.window.location.hash = "#"+current.owner.context.currentPoint.content.id;
         var target = current.owner.window.document.getElementById(current.owner.context.currentPoint.content.id);
         if (target) {
            current.setElement(target,{ scroll: true, newPosition: true });
         }
         current.togglePlay();
      } else {
         Console.debug("No current point for next.");
      }
   },500);
}

StructuralPlayback.prototype.stop = function() {
   Console.debug("Stopping playback.");
   this.playing = false;
   this.tts.cancel();
   this.reset();
   this.owner.togglePlayButton(true);
   if (this.ownedLock && this.ownedLock.lock.owner) {
      this.ownedLock.release();
   }
   this.ownedLock = null;
}


       