Console = {
   preferencesService: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("espeak.tts."),
   _initialized: false,
   _debug: false,
   service: Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
   log: function(message) {
      this.service.logStringMessage(message);
   },
   debug: function(message) {
      if (!this._initialized) {
         if (Console.preferencesService) {
            try {
               //Console.log(Console.preferencesService.prefHasUserValue("debug"));
               Console._debug = Console.preferencesService.getBoolPref("debug");
            } catch (ex) {
               // no preference
            }
         }
         this._initialized = true;
      }
      if (this._debug) {
         this.service.logStringMessage(message);
      }
   }
}

const CLASSID = Components.ID("{ae9f0e64-9cce-43be-97bd-0cd15e31bb11}");
const CLASSNAME = "A espeak-based TTS Engine Factory";
const CONTRACTID = "@benetech.org/tts-engine-factory;1?name=espeak";

function GenericComponentFactory(ctor, params) {
  this._ctor = ctor;
  this._params = params;
}
GenericComponentFactory.prototype = {
  _ctor: null,
  _params: null,

  createInstance: function GCF_createInstance(outer, iid) {
    if (outer != null)
      throw Cr.NS_ERROR_NO_AGGREGATION;
    return (new this._ctor(this._params)).QueryInterface(iid);
  },

  QueryInterface: function GCF_QueryInterface(iid) {
    if (iid.equals(Components.interfaces.nsIFactory) ||
        iid.equals(Components.interfaces.nsISupports))
      return this;
    throw Cr.NS_ERROR_NO_INTERFACE;
  }
};


var Module = {
  QueryInterface: function M_QueryInterface(iid) {
    if (iid.equals(Components.interfaces.nsIModule) ||
        iid.equals(Components.interfaces.nsISupports))
      return this;
    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  getClassObject: function M_getClassObject(cm, cid, iid) {
    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    return new GenericComponentFactory(ESpeakTTSEngineFactory);
  },

  registerSelf: function M_registerSelf(cm, file, location, type) {
    var cr = cm.QueryInterface(Components.interfaces.nsIComponentRegistrar);

    cr.registerFactoryLocation(CLASSID, CLASSNAME, CONTRACTID,
                               file, location, type);

  },

  unregisterSelf: function M_unregisterSelf(cm, location, type) {
     var cr = cm.QueryInterface(Components.interfaces.nsIComponentRegistrar);

     cr.registerFactoryLocation(CLASSID, location);
  },

  canUnload: function M_canUnload(cm) {
    return true;
  }
};

function NSGetModule(cm, file) {
  return Module;
}

function setTimeout(handler,delay) {
    var timer = Components.classes["@mozilla.org/timer;1"].createInstance().QueryInterface(Components.interfaces.nsITimer);
    timer.initWithCallback(handler,delay,0);
    return timer;
}

function ESpeakTTSEngineFactory() {

}

ESpeakTTSEngineFactory.prototype.QueryInterface = function(iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsITTSEngineFactory) ||
        iid.equals(Components.interfaces.nsIClassInfo)) {
       return this;
    }
    throw Components.results.NS_ERROR_NO_INTERFACE;
};

ESpeakTTSEngineFactory.prototype.classDescription = CLASSNAME;
ESpeakTTSEngineFactory.prototype.classID = CLASSID;
ESpeakTTSEngineFactory.prototype.contractID = CONTRACTID;
ESpeakTTSEngineFactory.prototype.implementationLanguage = Components.interfaces.nsIProgrammingLanguage.JAVASCRIPT;
ESpeakTTSEngineFactory.prototype.flags = Components.interfaces.nsIClassInfo.DOM_OBJECT;
ESpeakTTSEngineFactory.prototype.getHelperForLanguage = function() { return null; }
ESpeakTTSEngineFactory.prototype.getInterfaces = function(countRef) {
   var interfaces = [Components.interfaces.nsITTSEngineFactory, Components.interfaces.nsIClassInfo, Components.interfaces.nsISupports];
   countRef.value = interfaces.length;
   return interfaces;
}

ESpeakTTSEngineFactory.prototype.priority = 10;
ESpeakTTSEngineFactory.prototype.name = "espeak";

ESpeakTTSEngineFactory.prototype.initialize = function() {
   this.espeakBinary =  Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
   this.espeakBinary.initWithPath("/usr/bin/espeak");
   this.speakBinary =  Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
   this.speakBinary.initWithPath("/usr/bin/speak");
   return true;
}

ESpeakTTSEngineFactory.prototype.isAvailable = function() {
   // we don't check for the binary because espeak might be installed after firefox is installed
   //return this.espeakBinary.exists() || this.speakBinary.exists();
   return true;
}

ESpeakTTSEngineFactory.prototype.createEngine = function() {
   if (this.espeakBinary.exists()) {
      return new ESpeakTTSEngine(this.espeakBinary);
   } else if (this.speakBinary.exists()) {
      return new ESpeakTTSEngine(this.speakBinary);
   } else {
      return null;
   }
}

function ESpeakTTSEngine(binary) {
   this.binary = binary;
   this.process = null;
   this.speaking = false;
   this.rate = 170;
   this.minimumRate = 80;
   this.normalRate = 170;
   this.maximumRate = 390;
   this.rateUnit = "wpm";
   var tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   this.currentThread = tm.currentThread;
   this.thread = tm.newThread(0);
}

ESpeakTTSEngine.prototype.QueryInterface = function(iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsITTSEngine)) {
       return this;
    }
    throw Components.results.NS_ERROR_NO_INTERFACE;
};

ESpeakTTSEngine.prototype.name = ESpeakTTSEngineFactory.prototype.name;

ESpeakTTSEngine.prototype.isSpeaking = function() {
   return this.speaking;
}

ESpeakTTSEngine.prototype.cancel = function() {
   if (this.process) {
      this.process.cancel();
      this.process = null;
   }
}

ESpeakTTSEngine.prototype.speak = function(text,callback) {
   if (this.process) {
      this.process.cancel();
      this.process = null;
   }
   var current = this;
   this.process = {
      running: false,
      cancelled: false,
      run: function() {
         this.running = true;
         this.process = Components.classes["@mozilla.org/process/util;1"].createInstance(Components.interfaces.nsIProcess);
         this.process.init(current.binary);
         
         var argv = current.rate!=170 ? ["-s", current.rate+"", text] : [ text ];

         current.speaking = true;
         this.process.run(true, argv, argv.length);
         current.speaking = false;
         this.running = false;
         if (callback) {
            var myProcess = this;
            current.currentThread.dispatch({
               run: function() {
                  callback.onFinish(myProcess.cancelled);
               }
            },current.currentThread.DISPATCH_NORMAL);
         }
         current.process = null;
      },
      cancel: function() {
         this.cancelled = true;
         var proc = this.process;
         if (proc && this.running) {
            try {
               proc.kill();
            } catch (ex) {
               current.currentThread.dispatch({
                  run: function() {
                     Console.log("espeak cancel error: "+ex);
                  }
               },current.currentThread.DISPATCH_NORMAL);
            }
         }
      }
   };
   this.thread.dispatch(this.process,this.thread.DISPATCH_NORMAL);
}




