Console = {
   preferencesService: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("daisy.client."),
   _initialized: false,
   _debug: false,
   service: Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
   log: function(message) {
      this.service.logStringMessage(message);
   },
   debug: function(message) {
      if (!this._initialized) {
         if (Console.preferencesService) {
            try {
               //Console.log(Console.preferencesService.prefHasUserValue("debug"));
               Console._debug = Console.preferencesService.getBoolPref("debug");
            } catch (ex) {
               // no preference
            }
         }
         this._initialized = true;
      }
      if (this._debug) {
         this.service.logStringMessage(message);
      }
   }
}


function ChangeSettings() {
   var theClass = Components.classes["@benetech.org/daisy;1"];
   var theComponent = theClass.getService(Components.interfaces.nsISupports);
   this.service = theComponent.wrappedJSObject;
}

ChangeSettings.getInstance = function() {
   if (!ChangeSettings.instance) {
      ChangeSettings.instance = new ChangeSettings();
   }
   return ChangeSettings.instance;
}

ChangeSettings.prototype.init = function(data) {
   this.data = data;
   var highlightTime = document.getElementById("highlight-time");
   highlightTime.onchange = function() {
      document.getElementById("highlight-time-display").value = " "+highlightTime.value+" seconds";
   }
   var playbackPauseTime = document.getElementById("playback-pause-time");
   playbackPauseTime.onchange = function() {
      var ticks = parseInt(playbackPauseTime.value);
      document.getElementById("playback-pause-display").value = " "+(ticks*0.5)+" seconds";
   }
   this.uivoice = new UIVoice();
   this.uivoice.init();
   this.uivoice.attach(document);
   this.uivoice.speak("DAISY Fox player settings",null);

   var current = this;

   var speechRate = document.getElementById("speech-rate");
   speechRate.onchange = function() {
      var rate = current.calculateSpeechRate();
      if (current.data.speechRate.unit=="wpm") {
         document.getElementById("speech-rate-display").value = rate+" wpm";
      } else {
         document.getElementById("speech-rate-display").value = rate;
      }
   };


   var useSystemDefaults = document.getElementById("voice-system-defaults-checkbox");
   useSystemDefaults.addEventListener("click",function() {
      if (useSystemDefaults.checked) {
         document.getElementById("speech-rate").disabled = true;
         document.getElementById("speech-rate").value = 0;
         if (current.data.speechRate.unit=="wpm") {
            document.getElementById("speech-rate-display").value = current.data.speechRate.rate+" wpm";
         } else {
            document.getElementById("speech-rate-display").value = current.data.speechRate.rate;
         }
      } else {
         document.getElementById("speech-rate").disabled = false;
      }
   },false);
   useSystemDefaults.addEventListener("keypress",function() {
      if (useSystemDefaults.checked) {
         document.getElementById("speech-rate").disabled = true;
      } else {
         document.getElementById("speech-rate").disabled = false;
      }
   },false);

   this.updateUI();


}


ChangeSettings.prototype.updateUI = function() {
   document.getElementById("sound-checkbox").checked = this.data.sound;
   document.getElementById("highlight-checkbox").checked = this.data.highlight;
   //document.getElementById("selecttext-checkbox").checked = this.data.selectText;
   document.getElementById("focusfollow-checkbox").checked = this.data.focusFollow;
   //document.getElementById("debug-checkbox").checked = this.data.debug;
   document.getElementById("show-images-checkbox").checked = this.data.showImages;
   //document.getElementById("sound-context-checkbox").checked = this.data.showSoundContextWindow;
   //document.getElementById("generate-smil-checkbox").checked = this.data.generateSMIL;
   document.getElementById("quirks-checkbox").checked = this.data.quirks;
   document.getElementById("highlight-time").value = this.data.highlightTime+"";
   document.getElementById("playback-pause-time").value = (this.data.playbackPause/500)+"";
   document.getElementById("read-ui-checkbox").checked = this.data.readUI;
   document.getElementById("read-content-checkbox").checked = this.data.readContent;
   document.getElementById("force-read-content-checkbox").checked = this.data.forceReadContent;
   document.getElementById("navigate-by-structure-checkbox").checked = this.data.navigateByStructure;
   document.getElementById("sidebar-pause-checkbox").checked = this.data.sidebarPause;
   //document.getElementById("notes-inline-checkbox").checked = this.data.readNotesInline;
   document.getElementById("cue-sidebar-open-checkbox").checked = this.data.cueSidebarOpen;
   var value = 0;
   if (this.data.speechRate.preferredRate) {
      if (this.data.speechRate.preferredRate!=this.data.speechRate.normal) {
         value = -10;
         while (value<=10 && this.data.speechRate.preferredRate>this.calculateSpeechRate(value)) {
            value++;
         }
      }
      document.getElementById("voice-system-defaults-checkbox").checked = false;
   } else {
      if (this.data.speechRate.rate!=this.data.speechRate.normal) {
         value = -10;
         while (value<=10 && this.data.speechRate.rate>this.calculateSpeechRate(value)) {
            value++;
         }
      }
      document.getElementById("voice-system-defaults-checkbox").checked = true;
      document.getElementById("speech-rate").disabled = true;
      document.getElementById("speech-rate").value = 0;
   }

   document.getElementById("speech-rate").value = value;
   if (this.data.speechRate.unit=="wpm") {
      document.getElementById("speech-rate-display").value = this.data.speechRate.rate+" wpm";
   } else {
      document.getElementById("speech-rate-display").value = this.data.speechRate.rate;
   }
}

ChangeSettings.prototype.updateData = function() {
   this.data.sound = document.getElementById("sound-checkbox").checked;
   this.data.highlight = document.getElementById("highlight-checkbox").checked;
   //this.data.selectText = document.getElementById("selecttext-checkbox").checked;
   this.data.focusFollow = document.getElementById("focusfollow-checkbox").checked;
   //this.data.debug = document.getElementById("debug-checkbox").checked;
   this.data.showImages = document.getElementById("show-images-checkbox").checked;
   //this.data.showSoundContextWindow = document.getElementById("sound-context-checkbox").checked;
   //this.data.generateSMIL = document.getElementById("generate-smil-checkbox").checked;
   this.data.quirks = document.getElementById("quirks-checkbox").checked;
   this.data.highlightTime = parseInt(document.getElementById("highlight-time").value);
   this.data.playbackPause = parseInt(document.getElementById("playback-pause-time").value)*500;
   this.data.readContent = document.getElementById("read-content-checkbox").checked;
   this.data.readUI = document.getElementById("read-ui-checkbox").checked;
   this.data.forceReadContent = document.getElementById("force-read-content-checkbox").checked;
   this.data.navigateByStructure = document.getElementById("navigate-by-structure-checkbox").checked;
   this.data.sidebarPause = document.getElementById("sidebar-pause-checkbox").checked;
   //this.data.readNotesInline = document.getElementById("notes-inline-checkbox").checked;
   this.data.cueSidebarOpen = document.getElementById("cue-sidebar-open-checkbox").checked;
   this.data.speechRate.preferredRate = document.getElementById("voice-system-defaults-checkbox").checked ?  null : this.calculateSpeechRate();
}

ChangeSettings.prototype.calculateSpeechRate = function(avalue) {
   var value = avalue ? avalue : document.getElementById("speech-rate").value;
   if (value==0) {
      return this.data.speechRate.normal;
   } else if (value<0) {
      var span = this.data.speechRate.normal-this.data.speechRate.minimum;
      var increment = span/10;
      var count = 10+value;
      return this.data.speechRate.minimum + (increment*count);
   } else {
      var span = this.data.speechRate.maximum-this.data.speechRate.normal;
      var increment = span/10;
      return this.data.speechRate.normal + (increment*value);
   }

}

ChangeSettings.prototype.testTTS = function() {
   var ttsMediator = Components.classes["@benetech.org/tts-mediator-service;1"].getService(Components.interfaces.nsITTSMediatorService);
   var engine = ttsMediator.findEngine(null);
   if (engine) {
      if (!document.getElementById("voice-system-defaults-checkbox").checked) {
         engine.rate = this.calculateSpeechRate();
      }
      var tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
      var ttsCallback = {
         thread: tm.currentThread,
         QueryInterface: function(aIID) {
            if (aIID.equals(Components.interfaces.nsITTSCallback) ||
               aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
               aIID.equals(Components.interfaces.nsISupports)) {
               return this;
            }
            throw Components.results.NS_NOINTERFACE;
         },
         onFinish: function() { 
            this.thread.dispatch({
               run: function() {
                  Console.log("Speech completed.");
               }
            },this.thread.DISPATCH_NORMAL);
         }
     };

      engine.speak("Welcome to the DAISY Book reader!",ttsCallback);
   } else {
      Console.log("Cannot find TTS engine.");
   }
}

ChangeSettings.prototype.onCancel = function() {
   setTimeout(function() {
      window.close();
   },10);
}

ChangeSettings.prototype.onApply = function() {
   this.updateData();
   this.data.onUpdate();
}
