Console = {
   preferencesService: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("daisy.client."),
   _initialized: false,
   _debug: false,
   service: Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
   log: function(message) {
      this.service.logStringMessage(message);
   },
   debug: function(message) {
      if (!this._initialized) {
         if (Console.preferencesService) {
            try {
               //Console.log(Console.preferencesService.prefHasUserValue("debug"));
               Console._debug = Console.preferencesService.getBoolPref("debug");
            } catch (ex) {
               // no preference
            }
         }
         this._initialized = true;
      }
      if (this._debug) {
         this.service.logStringMessage(message);
      }
   }
}

function makeCallback(callback) {
   var tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   return {
      thread: tm.currentThread,
      QueryInterface: function(aIID) {
         if (aIID.equals(Components.interfaces.nsITTSCallback) ||
            aIID.equals(Components.interfaces.nsISupports)) {
            return this;
         }
         throw Components.results.NS_NOINTERFACE;
      },
      onFinish: function() {
         this.thread.dispatch({
            run: function() {
               callback();
            }
         },this.thread.DISPATCH_NORMAL);
      }
  };
}

function FilePickerDialog() {
   var theClass = Components.classes["@benetech.org/daisy;1"];
   var theComponent = theClass.getService(Components.interfaces.nsISupports);
   this.service = theComponent.wrappedJSObject;
}

FilePickerDialog.XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";


FilePickerDialog.getInstance = function() {
   if (!FilePickerDialog.instance) {
      FilePickerDialog.instance = new FilePickerDialog();
   }
   return FilePickerDialog.instance;
}

FilePickerDialog.prototype.init = function(data) {
   this.data = data;
   document.getElementById("title").value = data.title;
   this.dirMenu = document.getElementById("dirs");
   this.filesTree = document.getElementById("files");
   this.matches = [];
   if (this.data.filters) {
      for (var i=0; i<this.data.filters.types.length; i++) {
         var patterns = this.data.filters.types[i].split(";");
         for (var j=0; j<patterns.length; j++) {
            if (patterns[j].substring(0,2)=="*.") {
               this.matches.push(new RegExp(patterns[j].substring(1)+"$"));
            } else {
               this.matches.push(new RegExp(patterns[j]));
            }
         }
      }
   }
   this.onDirectoryChange();
   var current = this;
   this.dirMenu.addEventListener("keypress",function(event) {
      if (event.keyCode!=13) {
         return;
      }
      if (current.inDirectoryChange) {
         return;
      }
      current.inDirectoryChange = true;
      var path = current.dirMenu.selectedItem.value;
      current.data.directory = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
      current.data.directory.initWithPath(path);
      current.onDirectoryChange();
      current.inDirectoryChange = false;
      if (current.uivoice) {
         current.uivoice.speak("Changed directory to "+current.dirMenu.selectedItem.label,makeCallback(function() { current.filesTree.focus(); }));
      }
   },false);

   var onFileSelect = function() {
      var leafName = current.filesTree.view.getCellText(current.filesTree.view.selection.currentIndex,current.filesTree.columns.getFirstColumn());
      var file = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(current.data.directory.path);
      file.append(leafName);
      current.data.file = file;
      if (file.isDirectory()) {
         current.inDirectoryChange = true;
         current.data.directory = file;
         current.onDirectoryChange();
          current.inDirectoryChange = false;
         if (current.uivoice) {
            current.uivoice.speak("Changed directory to "+current.dirMenu.selectedItem.label,makeCallback(function() { current.filesTree.onfocus(); }));
         }
      } else {
         current.onOpen();
      }
   };
   this.filesTree.ondblclick = onFileSelect;
   this.filesTree.onkeypress = function(event) {
      if (event.keyCode==13) {
         onFileSelect();
      }
   };
   this.filesTree.onselect = function() {
      if (current.inDirectoryChange) {
         return;
      }
      var leafName = current.filesTree.view.getCellText(current.filesTree.view.selection.currentIndex,current.filesTree.columns.getFirstColumn());
      var file = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
      file.initWithPath(current.data.directory.path);
      file.append(leafName);
      current.data.file = file;
      if (current.uivoice) {
         current.uivoice.speak(leafName,null);
      }
   }
   this.filesTree.onfocus = function() {
      var leafName = current.filesTree.view.getCellText(current.filesTree.view.selection.currentIndex,current.filesTree.columns.getFirstColumn());
      if (current.uivoice) {
         current.uivoice.speak("File List, "+leafName+" selected.",null);
      }
   }


   this.uivoice = new UIVoice();
   this.uivoice.init();
   this.uivoice.attach(document);
   this.uivoice.speak("Browse dialog, "+data.title,makeCallback(function() {
      current.filesTree.focus();
   }));

}

FilePickerDialog.prototype.onDirectoryChange = function() {
   var current = this;
   while (this.dirMenu.itemCount>0) {
      this.dirMenu.removeItemAt(0);
   }
   var currentDir = this.data.directory;
   while (currentDir) {
      var item = this.dirMenu.appendItem(currentDir.leafName,currentDir.path);
      currentDir = currentDir.parent;
   }
   this.dirMenu.selectedIndex = 0;
   var treechildren = this.filesTree.firstChild.nextSibling;
   var child = treechildren.firstChild;
   while (child) {
      var next = child.nextSibling;
      treechildren.removeChild(child);
      child = next;
   }
   var dirEntries = this.data.directory.directoryEntries;
   var count = 0;
   while (dirEntries.hasMoreElements()) {
      count++;
      var file = dirEntries.getNext().QueryInterface(Components.interfaces.nsIFile);
      if (file.leafName.charAt(0)==".") {
         //Console.debug("Skipping "+file.path);
         continue;
      }
      if (!file.exists()) {
         //Console.debug("Skipping "+file.path);
         continue;
      }
      if (this.matches.length>0 && !file.isDirectory()) {
         var match = false;
         for (var i=0; !match && i<this.matches.length; i++) {
            match = this.matches[i].test(file.leafName);
         }
         if (!match) {
            continue;
         }
      }
      var item = document.createElementNS(FilePickerDialog.XUL_NS,"treeitem");
      var row = document.createElementNS(FilePickerDialog.XUL_NS,"treerow");
      item.appendChild(row);
      var nameCell = document.createElementNS(FilePickerDialog.XUL_NS,"treecell");
      nameCell.setAttribute("label",file.leafName);
      nameCell.setAttribute("properties",file.isDirectory() ? "directory" : "file");
      row.appendChild(nameCell);
      var modDate = new Date(file.lastModifiedTime);
      var modCell = document.createElementNS(FilePickerDialog.XUL_NS,"treecell");
      modCell.setAttribute("label",modDate.toLocaleString());
      row.appendChild(modCell);
      var children = document.createElementNS(FilePickerDialog.XUL_NS,"treechildren");
      item.appendChild(children);
      treechildren.appendChild(item);
   }
   if (count>0) {
      this.filesTree.view.selection.select(0);
   }

}

FilePickerDialog.prototype.onCancel = function() {
   this.data.result = false;
   setTimeout(function() {
      window.close();
   },10);
}

FilePickerDialog.prototype.onOpen = function() {
   this.data.result = true;
   setTimeout(function() {
      window.close();
   },10);
}
