Console = {
   preferencesService: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("daisy.client."),
   _initialized: false,
   _debug: false,
   service: Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
   log: function(message) {
      this.service.logStringMessage(message);
   },
   debug: function(message) {
      if (!this._initialized) {
         if (Console.preferencesService) {
            try {
               //Console.log(Console.preferencesService.prefHasUserValue("debug"));
               Console._debug = Console.preferencesService.getBoolPref("debug");
            } catch (ex) {
               // no preference
            }
         }
         this._initialized = true;
      }
      if (this._debug) {
         this.service.logStringMessage(message);
      }
   }
}


function Settings() {
   this.password = null;
   var extManager = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
   // the path may use forward slash ("/") as the delimiter
   // returns nsIFile for the extension's install.rdf
   var id = "{6be5ef88-4289-44d1-81f2-097313ed640b}";
   this.opfFile = extManager.getInstallLocation(id).getItemFile(id,"content/galley.xhtml");
   this.galleyFile = extManager.getInstallLocation(id).getItemFile(id,"content/galley.xhtml");
   this.renderedFile = extManager.getInstallLocation(id).getItemFile(id,"content/rendered.xhtml");
   this.renderedPlayerFile = extManager.getInstallLocation(id).getItemFile(id,"content/rendered-player.xhtml");
   this.soundFlash1 = extManager.getInstallLocation(id).getItemFile(id,"content/soundmanager2/swf/soundmanager2.swf");
   this.soundFlash2 = extManager.getInstallLocation(id).getItemFile(id,"content/soundmanager2/swf/soundmanager2_flash9.swf");
   this.soundFile = extManager.getInstallLocation(id).getItemFile(id,"content/sound.xhtml");
   this.soundTest = extManager.getInstallLocation(id).getItemFile(id,"content/bass.mp3");
   this.notFoundSound = extManager.getInstallLocation(id).getItemFile(id,"content/notFound.mp3");
   this.searchWrappedSound = extManager.getInstallLocation(id).getItemFile(id,"content/wrapped.mp3");
   this.playButtonImage = extManager.getInstallLocation(id).getItemFile(id,"content/images/play-button.png");
   this.pauseButtonImage = extManager.getInstallLocation(id).getItemFile(id,"content/images/pause-button.png");
   this.helpOPF = extManager.getInstallLocation(id).getItemFile(id,"content/help/help.opf");
   var theClass = Components.classes["@benetech.org/daisy;1"];
   var theComponent = theClass.getService(Components.interfaces.nsISupports);
   this.service = theComponent.wrappedJSObject;
}

Settings.prototype._getPreferenceString = function(path,name) {
   var preferencesService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(path);
   if (preferencesService) {
      try {
         return preferencesService.getCharPref(name);
      } catch (ex) {
         // no preference
      }
   }
   return null;
}

Settings.prototype._setPreferenceString = function(path,name,value) {
   var preferencesService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(path);
   if (preferencesService) {
      try {
         return preferencesService.setCharPref(name,value);
      } catch (ex) {
         // no preference
      }
   }
   return null;
}

Settings.prototype._getPreferenceInt = function(path,name) {
   var preferencesService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(path);
   if (preferencesService) {
      try {
         return preferencesService.getIntPref(name);
      } catch (ex) {
         // no preference
      }
   }
   return null;
}

Settings.prototype._setPreferenceInt = function(path,name,value) {
   var preferencesService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(path);
   if (preferencesService) {
      try {
         return preferencesService.setIntPref(name,value);
      } catch (ex) {
         // no preference
      }
   }
   return null;
}

Settings.prototype._getPreferenceBool = function(path,name) {
   var preferencesService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(path);
   if (preferencesService) {
      try {
         return preferencesService.getBoolPref(name);
      } catch (ex) {
         // no preference
      }
   }
   return null;
}

Settings.prototype._setPreferenceBool = function(path,name,value) {
   var preferencesService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(path);
   if (preferencesService) {
      try {
         return preferencesService.setBoolPref(name,value);
      } catch (ex) {
         // no preference
      }
   }
   return null;
}

Settings.prototype.notify = function(title,message) {
   var data = {
      title: title,
      message: message
   };
   var notifyWindow = window.openDialog("chrome://daisy/content/notify.xul",'daisy-notify',"chrome,resizable,modal",data);
   notifyWindow.focus();
}

Settings.prototype.promptForPassword = function(title,retry) {
   var data = {
      prompt: retry ? "Incorrect password. Please enter the password for the book titled \""+title+"\" :" :
                      "Please enter the password for the book titled \""+title+"\" :"
   };
   var passwordWindow = window.openDialog("chrome://daisy/content/password.xul",'daisy-password',"chrome,resizable,modal",data);
   passwordWindow.focus();
   return {
      success: data.success,
      password: data.password
   };
}

Settings.prototype.ensureUnpacked = function(book,force) {
   if (force || (book.isLocal() && !this.service.isExtracted(book))) {
      var decoder = new ZipDecoder(book.zipFile.path);
      var encrypted = decoder.isEncrypted();
      Console.log("Extracting book "+book.title+" ...");
      if (encrypted && !this.password) {
         var result = this.promptForPassword(book.title,false);
         if (!result.success) {
            return false;
         } else {
            this.password = result.password;
         }
      }
      var targetDir = this.service.getBookDirectory(book);
      this.service.extract(book.zipFile,targetDir,this.password);
      if (!this.service.isExtracted(book)) {
         if (encrypted) {
            var result = this.promptForPassword(book.title,true);
            if (result.success) {
               this.password = result.password;
               this.service.extract(book.zipFile,targetDir,this.password);
               if (!this.service.isExtracted(book)) {
                  this.notify("Incorrect Password", "Cannot extract book "+book.title+" due to incorrect password.");
                  return false;
               }
            } else {
               return false;
            }
         } else {
            this.notify("Incomplete Operation","Cannot extract book "+book.title);
            return false;
         }
      }
   }
   return true;
}


