Sidebar.prototype = new Settings();
Sidebar.prototype.constructor = Sidebar;

Sidebar.XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
Sidebar.ios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
Sidebar.positionOffset = 80;
Sidebar.fileURL = function(file) {
   return Sidebar.ios.newFileURI(file);
}

Sidebar.fileFromURI = function(suri) {
   var uri = Sidebar.ios.newURI(suri,"UTF-8",null);
   if (uri.scheme=="file") {
      uri = uri.QueryInterface(Components.interfaces.nsIFileURL);
      return uri.file;
   }
   return null;
}


function Sidebar() {
   this._windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService(Components.interfaces.nsIWindowMediator);
   this.OS = Components.classes["@mozilla.org/xre/app-info;1"].getService(Components.interfaces.nsIXULRuntime).OS;
   var current = this;
   window.addEventListener('focus', function() {
      current.hasFocus = true;
   }, false);
   window.addEventListener('focus', function() {
      current.hasFocus = false;
   }, false);
}

Sidebar.getInstance = function() {
   if (!Sidebar.instance) {
      Sidebar.instance = new Sidebar();
   }
   return Sidebar.instance;
}

Sidebar.prototype.init = function() {
   this.service.active = true;
   try {
   this.playButtonImageURI = Sidebar.fileURL(this.playButtonImage).spec;
   this.pauseButtonImageURI = Sidebar.fileURL(this.pauseButtonImage).spec;
   Console.debug("Initializing sidebar");
   this.deck = document.getElementById("mode");
   this.mainWindow = this._windowMediator.getMostRecentWindow("navigator:browser");
   this.soundContext = this.mainWindow.document.getElementById("daisy-sound-context");
   this.showSoundContextWindow = this.soundContext.boxObject.height>0;
   this.initSoundManager();
   window.onresize = function() {
      var listWidth = window.innerWidth-30;
      document.getElementById("toc-booklist").style.maxWidth = listWidth+"px";
   }
   window.onresize();
   var current = this;
   window.onkeypress = function(event) {
      //Console.log("Sidebar: "+event.shiftKey+" "+event.altKey+" "+event.metaKey+" "+event.ctrlKey+" "+event.charCode);
      if (event.shiftKey && event.altKey && event.ctrlKey && event.charCode==68) {
         // shift-ctrl-alt-d toggles debug
         if (current.service.debug) {
            Console.log("Turning debug off.");
            current.service.debug = false;
         } else {
            Console.log("Turning debug on.");
            current.service.debug = true;
         }
      } else if (event.shiftKey && event.altKey && event.ctrlKey && event.charCode==83) {
         // shift-ctrl-alt-s toggles sound manager context
         if (current.showSoundContextWindow) {
            current.showSoundContextWindow = false;
            current.soundContext.style.height = "0px";
         } else {
            current.showSoundContextWindow = true;
            current.soundContext.style.height = "300px";
         }
      } else if ((event.ctrlKey && event.altKey && event.charCode==97) || (event.metaKey && event.altKey && event.charCode==97)) {
         var tabBrowser = current.mainWindow.getBrowser();
         var helper = tabBrowser.selectedTab.linkedBrowser.contentWindow.wrappedJSObject.helper;
         if (helper) {
            helper = helper.wrappedJSObject;
            helper.onMoveToActionBar();
         }
      }
   };
   var focusBooklist = false;
   if (!this.attached) {
      var tabBrowser = this.mainWindow.getBrowser();
      this.url = tabBrowser.selectedTab.linkedBrowser.contentWindow.location.href;
      var hashPos = this.url.indexOf("#");
      if (hashPos>0) {
         this.url = this.url.substring(0,hashPos);
      }
      Console.debug("tab url: "+this.url);
      var book = this.service.getBook(this.url);
      if (book) {
         Console.debug("OPF document detected.");
         if (tabBrowser.selectedTab.linkedBrowser.contentWindow.wrappedJSObject.helper) {
            // only call for reopening sidebar when on the same OPF document
            var helper = tabBrowser.selectedTab.linkedBrowser.contentWindow.wrappedJSObject.helper.wrappedJSObject;
            var context = helper.context;
            if (context) {
               Console.debug("Attaching to existing DAISY document: "+this.url);
               this.helper = helper;
               this.attach(this.url,context);
               if (this.bookSpec.contentWindow.daisyRendered) {
                  this.populateTOC(this.bookSpec);
                  this.update();
               }
            } else {
               Console.debug("Helper does not have a context yet, waiting...");
               var waitCount = 0;
               var waitForContext = function() {
                  if (helper.context) {
                     current.init();
                  } else if (waitCount<600) {
                     waitCount++;
                     setTimeout(waitForContext,100);
                  } else {
                     Console.log("Cannot get context from helper--render may have failed.");
                  }
               }
               waitForContext();
               return;
            }
         } else {
            Console.log("No helper in tab: "+tabBrowser.selectedTab.linkedBrowser.contentWindow);
         }
         this.booklist = document.getElementById("toc-booklist");
      } else {
         this.booklist = document.getElementById("booklist");
         this.deck.selectedIndex = 0;
         window.focus();
         focusBooklist = true;
      }
   } else {
      this.booklist = document.getElementById("toc-booklist");
   }
   this.populateBooklist();
   this.booklist.onselect = function() {
      if (current.service.readUI) {
         current.speak(current.booklist.selectedItem.label);
      }
   }
   if (document.defaultView.parent.DAISYOverlay.getInstance().useCount()==0) {
      var current = this;
      setTimeout(function() {
         current.onOpenCue(function() {
            if (focusBooklist) {
               current.booklist.focus();
            }
         });
      },250);
   } else {
      if (focusBooklist) {
         this.booklist.focus();
      }
   }
   } catch (ex) {
      Console.log(ex);
   }
}

Sidebar.prototype.cancelVoice = function(element) {
   if (!this.service.readUI) {
      return;
   }
   this.service.tts.cancel();
}

Sidebar.prototype.voice = function(element) {
   if (!this.service.readUI) {
      return;
   }
   var current = this;
   var id = element.getAttribute("id");
   var suffix = "";
   if (id=="booklist" || id=="toc-booklist" || id=="book-toc") {
      if (element.localName=="tree") {
         var pos = parseInt(current.bookSpec.toc.view.getCellValue(current.bookSpec.toc.view.selection.currentIndex,current.bookSpec.toc.columns.getFirstColumn()));
         suffix = ", "+current.bookSpec.tocData[pos].point.labels[0].text;
         if (current.bookSpec.toc.view.isContainer(current.bookSpec.toc.view.selection.currentIndex)) {
            suffix += ", item has children, ";
            if (current.bookSpec.toc.view.isContainerOpen(current.bookSpec.toc.view.selection.currentIndex)) {
               suffix += ", is open";
            } else {
               suffix += ", is closed";
            }
         }
      } else if (current.booklist.selectedItem) {
         suffix = ", "+current.booklist.selectedItem.label;
      }

   }
   var text = element.getAttribute("tooltiptext");
   if (element.localName=="button" || element.localName=="toolbarbutton") {
      text = "Button, "+text;
   }
   text += suffix;
   this.speak(text);
}

Sidebar.prototype.speak = function(text) {
   var lock = this.service.getLockFor("tts");
   var current = this;
   this.ownedLock = lock.acquire({
      exclusive: true,
      onRelinquish: function() {
         current.service.tts.cancel();
      }
   });
   this.service.tts.speak(text,{
      QueryInterface: function(aIID) {
         if (aIID.equals(Components.interfaces.nsITTSCallback) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports)) {
            return this;
         }
         throw Components.results.NS_NOINTERFACE;
      },
      onFinish: function() {
         if (current.ownedLock) {
            current.ownedLock.release();
            current.ownedLock = null;
         }
      }
   });
}

Sidebar.prototype.populateBooklist = function() {
   Console.debug("Populating recently opened book list");
   var current = this;
   this.booklist.ondblclick = function() {
      current.openDocument(current.booklist.selectedItem.value);
   };
   this.booklist.onkeypress = function(event) {
      if (event.keyCode==13) {
         current.openDocument(current.booklist.selectedItem.value);
      } else if (event.keyCode==8 || event.keyCode==46){
         current.onDeleteBookFromList();
      }
   };

   var db = this.service.database;
   var history = db.getBookHistory();
   while (this.booklist.itemCount>0) {
      this.booklist.removeItemAt(0);
   }
   for (var i=0; i<history.length; i++) {
      var url = history[i].url;
      if (history[i].position && history[i].position.length>0) {
         url = url+"#"+history[i].position;
      }
      Console.debug(history[i].title+" -> "+url);
      this.showBook(history[i].title,url);
   }
   if (history.length>0) {
      this.booklist.selectedIndex = 0;
   }
}

Sidebar.prototype.setDocumentTitle = function(title) {
   var desc = document.getElementById("document-title");
   DOMUtil.getInstance().clearChildren(desc);
   desc.appendChild(desc.ownerDocument.createTextNode(title));
}

Sidebar.prototype.attach = function(url,context) {
   Console.debug("Attach: "+url);
   var current = this;
   this.url = url;
   if (!this.deck) {
      var checker = function() {
         current.deck = document.getElementById("mode");
         if (!current.deck) {
            setTimeout(checker,100);
         } else {
            current.attach(url,context);
         }
      }
      checker();
      return;
   }
   /*
   var trySound = 0;
   var setSoundManager = function() {
      if (window.soundManager) {
         context.options.soundManager = window.soundManager;
         FlashBlockHack.restore();
      } else if (trySound<20) {
         trySound++;
         setTimeout(setSoundManager,100);
      } else {
         FlashBlockHack.restore();
      }
   };
   setSoundManager();
   */
   var tabBrowser = this._windowMediator.getMostRecentWindow("navigator:browser").getBrowser();
   this.bookSpec = {
      book: context.book,
      toc: document.getElementById("book-toc"),
      tocData: [],
      context: context,
      contentWindow: tabBrowser.selectedTab.linkedBrowser.contentWindow.wrappedJSObject,
      contentDocument: context.document
   }

   this.deck.selectedIndex = 1;

   this.setDocumentTitle(this.bookSpec.book.title);
   var treechildren = this.bookSpec.toc.firstChild.nextSibling;
   this.bookSpec.toc.onselect = null;
   DOMUtil.getInstance().clearChildren(treechildren);

   this.bookSpec.toc.onselect = function() {
      if (!current.tocSelectFromScroll && current.bookSpec.toc.view.selection.currentIndex>=0) {
         current.tocSelect = true;
         var pos = parseInt(current.bookSpec.toc.view.getCellValue(current.bookSpec.toc.view.selection.currentIndex,current.bookSpec.toc.columns.getFirstColumn()));
         var makeSelector = function(index) {
            return function() {
               if (index!=current.bookSpec.toc.view.selection.currentIndex) {
                  return;
               }
               current.tocSelectTimer = null;
               current.bookSpec.context.stop();
               current.bookSpec.context.owner.togglePlayButton(true);
               current.bookSpec.context.currentPoint = current.bookSpec.tocData[pos].point;
               // Since we are using the hash, we don't need to scroll
               //var navPoint = current.bookSpec.tocData[pos].point;
               //current.bookSpec.context.moveTo(navPoint.content.id,-Sidebar.positionOffset);
               //
               // Setting the hash has some affect on JAWS context but ultimately does
               // not navigate JAWS to the TOC item.
               var tabBrowser = current.mainWindow.getBrowser();
               var helper = tabBrowser.selectedTab.linkedBrowser.contentWindow.wrappedJSObject.helper.wrappedJSObject;
               helper.moveToLocation(current.bookSpec.tocData[pos].point.content.id);
               //current.bookSpec.contentWindow.location.hash = "#"+current.bookSpec.tocData[pos].point.content.id;
               // Note: scrolling causes the document to "flash"
               setTimeout(function() {
                  //current.bookSpec.contentWindow.scroll(0,current.bookSpec.contentWindow.scrollY-Sidebar.positionOffset);
                  helper.notifyTOCChange();
               },100);
               setTimeout(function() {
                  current.tocSelect = false;
               },250);
            }
         }
         setTimeout(makeSelector(current.bookSpec.toc.view.selection.currentIndex),500);
         if (current.service.readUI) {
            var text = current.bookSpec.tocData[pos].point.labels[0].text;
            //var text = current.bookSpec.toc.view.getCellText(current.bookSpec.toc.view.selection.currentIndex,current.bookSpec.toc.columns.getFirstColumn())
            if (current.bookSpec.toc.view.isContainer(current.bookSpec.toc.view.selection.currentIndex)) {
               text += ", item has children, ";
               if (current.bookSpec.toc.view.isContainerOpen(current.bookSpec.toc.view.selection.currentIndex)) {
                  text += ", is open";
               } else {
                  text += ", is closed";
               }
            }
            Console.debug("UI: tree item, text="+text);
            current.service.tts.cancel();
            current.speak(text);
         }
      }
      /*
      if (current.bookSpec.toc.view.selection.currentIndex>=0) {
         current.bookSpec.book.tocPosition = current.bookSpec.toc.view.selection.currentIndex
      }
      */
   }
   this.bookSpec.toc.onkeypress = function(event) {
      if (event.charCode==112) {
         // the 'p' key
         // This is disabled because it conflicts with moving to TOC entries that start with 'p'
         //current.onPlay();
      } else if (event.charCode==115) {
         // the 's' key
         // This is disabled because it conflicts with moving to TOC entries that start with 's'
         //current.bookSpec.context.stop();
      } else if (event.keyCode==13) {
         if (event.shiftKey) {
            current.bookSpec.context.owner.bookContext.togglePlay();
         }
         current.bookSpec.contentWindow.focus();
      }
   }
   if (this.bookSpec.contentWindow.daisyRendered) {
      this.update();
   }
   this.attached = true;

}

Sidebar.prototype.switchTo = function(url,context) {
   if (!this.deck) {
      return;
   }
   if (this.url==url) {
      return;
   }
   Console.debug("switchTo: "+url);
   var current = this;
   this.url = url;
   this.deck.selectedIndex = 1;
   var tabBrowser = this._windowMediator.getMostRecentWindow("navigator:browser").getBrowser();
   this.bookSpec = {
      book: context.book,
      toc: document.getElementById("book-toc"),
      tocData: [],
      context: context,
      contentWindow: tabBrowser.selectedTab.linkedBrowser.contentWindow.wrappedJSObject,
      contentDocument: context.document
   };
   this.booklist = document.getElementById("toc-booklist");
   this.booklist.onselect = function() {
      if (current.service.readUI && current.booklist.selectedItem) {
         current.speak(current.booklist.selectedItem.label);
      }
   }
   this.populateBooklist();

   this.setDocumentTitle(this.bookSpec.book.title);

   var treechildren = this.bookSpec.toc.firstChild.nextSibling;
   var onselect = this.bookSpec.toc.onselect;
   this.bookSpec.toc.onselect = null;
   DOMUtil.getInstance().clearChildren(treechildren);
   this.populateTOC(this.bookSpec);
   this.bookSpec.toc.onselect = onselect;
   //Console.log("switchTo()");
   this.updateNavigationPosition();
   this.updateContext(this.bookSpec.contentWindow.scrollY+100,true);
}

Sidebar.prototype.update = function() {
   if (!this.bookSpec || !this.attached) {
      return;
   }
   Console.debug("Sidebar.update:");
   this.updateNavigationPosition();
   this.updateContext(this.bookSpec.contentWindow.scrollY+100,true);
}

Sidebar.prototype.onAfterRender = function() {
   //Console.log("onAfterRender()");
   if (!this.bookSpec) {
      return;
   }
   this.populateTOC();
   this.update();
   /*  This shouldn't be necessary
   var current = this;
   setTimeout(function() {
      current.updateNavigationPosition();
   },500);
   setTimeout(function() {
      current.updateNavigationPosition();
   },1000);
   */
}

Sidebar.prototype.focus = function() {
   var current = this;
   Console.debug("focus: "+this.deck.selectedIndex+", booklist="+this.booklist);
   if (this.deck && this.deck.selectedIndex==1) {
      var tabbox = document.getElementById("toc-tabs");
      if (tabbox.selectedIndex==1) {
         setTimeout(function() {current.bookSpec.toc.focus();},100);
      } else {
         setTimeout(function() {current.booklist.focus();},100);
      }
   } else {
      setTimeout(function() {current.booklist.focus();},100);
   }
}

Sidebar.prototype.openBook = function(book) {
   if (!this.ensureUnpacked(book)) {
      return;
   }
   var entry = book.findByMediaType("application/oebps-package+xml");
   if (!entry) {
      entry = book.findByExtension("opf");
   }
   if (!entry) {
      throw "Cannot find OPF document in book manifest.";
   }
   var url = entry.resolve();
   var tab = this._newTab(url.spec);

}



Sidebar.prototype.destroy = function() {
   this.closed = true;
}


Sidebar.prototype.showBook = function(title,uri) {
   return this.booklist.appendItem(title,uri);
}

Sidebar.prototype.onPlay = function() {
   if (this.bookSpec.context.playContext) {
      this.bookSpec.context.stop();
      this.bookSpec.context.owner.togglePlayButton(true);
   } else {
      try {
         var pos = parseInt(this.bookSpec.toc.view.getCellValue(this.bookSpec.toc.view.selection.currentIndex,this.bookSpec.toc.columns.getFirstColumn()));
         var navPoint = this.bookSpec.tocData[pos].point;

         Console.debug("Calling play on context "+navPoint.content.id);
         this.bookSpec.context.play(navPoint);
         this.bookSpec.context.owner.togglePlayButton(false);
      } catch (ex) {
         Console.log(ex);
      }
   }
}

Sidebar.prototype.openDocument = function(url,forcenobook) {
   //alert("Open: "+url);
   var cleanurl = url;
   var hashPos = cleanurl.indexOf("#");
   if (hashPos>0) {
      cleanurl = cleanurl.substring(0,hashPos);
   }
   var tabBrowser = this._windowMediator.getMostRecentWindow("navigator:browser").getBrowser();
   for (var i=0; i<tabBrowser.tabContainer.itemCount; i++) {
      var tab = tabBrowser.tabContainer.getItemAtIndex(i);
      var href = tab.linkedBrowser.contentWindow.location.href;
      var hash = href.indexOf("#");
      if (hash>0) {
         href = href.substring(0,hash);
      }
      //alert(href+" vs "+cleanurl);
      if (href==cleanurl) {
         tabBrowser.selectedTab = tab;
         tab.linkedBrowser.contentWindow.focus();
         document.getElementById("toc-tabs").selectedIndex = 1;
         return;
      }
   }


   if (!forcenobook) {
      var book = this.service.getBook(url);
      if (!book) {
         book = this.service.newBook(url);
      }

      // If the file is a local file, check to see if it is extracted
      if (url.substring(0,5)=="file:") {
         var extracted = false;
         var file = Sidebar.fileFromURI(url);
         if (file.exists()) {

            // It exists, but maybe only the OPF file was extracted
            var dirEntries = file.parent.directoryEntries;
            var opfTest = /\.opf$/;
            var nonOPFFound = false;
            while (!nonOPFFound &&dirEntries.hasMoreElements()) {
               var bookFile = dirEntries.getNext().QueryInterface(Components.interfaces.nsIFile);
               if (!bookFile.leafName.match(opfTest)) {
                  nonOPFFound = true;
               }
            }
            extracted = nonOPFFound ? true : false;
         } else {
            // It doesn't exist, hopefully there is a zip file
            extracted = false;
         }
         if (!extracted) {
            var zipFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
            zipFile.initWithPath(file.parent.parent.path);
            zipFile.appendRelativePath(file.parent.leafName+".zip");
            var bks2File = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
            bks2File.initWithPath(file.parent.parent.path);
            bks2File.appendRelativePath(file.parent.leafName+".bks2");
            if (zipFile.exists() || bks2File.exists()) {
               var arcFile = zipFile.exists() ? zipFile : bks2File;
               var result = this.service.readZipPackaging(arcFile);
               if (result.document) {
                  book.zipFile = arcFile;
                  book.init();
                  book.loadPackaging(result.document,result.baseURI);
                  book.zipFile = arcFile;
                  Console.debug("Book manifest from "+arcFile.path+" loaded.");
                  if (!this.ensureUnpacked(book,true)) {
                     return;
                  }
               } else {
                  Console.debug("No OPF document in zip "+arcFile.path);
                  return;
               }
            } else {
               alert("Cannot locate document: "+url);
               return;
            }
         }
      }
   }

   if (tabBrowser.selectedTab && tabBrowser.selectedTab.linkedBrowser.contentWindow.location.href=="about:blank") {
      tabBrowser.selectedTab.linkedBrowser.contentWindow.location = url;
   } else {
      var tab = this._newTab(url);
   }
}

Sidebar.prototype.onDeleteBookFromList = function() {
   if (!this.booklist.selectedItem) {
      return;
   }
   var url = this.booklist.selectedItem.value;
   this.service.database.removeBookHistory(url);
   var index = this.booklist.selectedIndex;
   this.booklist.removeItemAt(index);
   if (this.booklist.itemCount>0) {
      if (index>=this.booklist.itemCount) {
         index--;
      }
      this.booklist.selectedIndex = index;
   }
}

Sidebar.prototype.populateTOC = function() {
   if (this.bookSpec.tocDone) {
      return;
   }
   this.bookSpec.tocDone = true;
   var treechildren = this.bookSpec.toc.firstChild.nextSibling;
   for (var i=0; i<this.bookSpec.book.navigation.navMap.children.length; i++) {
      this.addNavigationPoint(this.bookSpec.tocData,treechildren,this.bookSpec.book.navigation.navMap.children[i],"toc-item");
   }
}

Sidebar.prototype.constructNavigationPoint = function(navPoint,value,className) {
   var item = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul","treeitem");
   if (navPoint.children.length>0) {
      item.setAttribute("container","true");
   }
   var row = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul","treerow");
   row.setAttribute("properties",navPoint.className ? navPoint.className : className);
   item.appendChild(row);
   var name = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul","treecell");
   name.setAttribute("properties",navPoint.className ? navPoint.className : className);
   row.appendChild(name);
   name.setAttribute("label",navPoint.labels[0].text);
   name.setAttribute("value",value);
   return item;
}

Sidebar.prototype.addNavigationPoint = function(tocData,parent,navPoint,className) {
   tocData.push({
      point: navPoint
   });
   var item = this.constructNavigationPoint(navPoint,""+(tocData.length-1),className);

   var children = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul","treechildren");
   item.appendChild(children);

   for (var i=0; i<navPoint.children.length; i++) {
      this.addNavigationPoint(tocData,children,navPoint.children[i],className);
   }

   parent.appendChild(item);
   
   navPoint.item = item;
}

Sidebar.prototype.notifyAddNote = function(note) {
   var context = note.point;
   while (context.parent && !context.next) {
      context = context.parent;
   }
   if (!context.parent) {
      // we're at the end
      this.addNavigationPoint(this.bookSpec.tocData,this.bookSpec.toc.firstChild.nextSibling,note.point,"note");
   } else {
      context = context.next;
      // we have a next element
      var index = parseInt(context.item.firstChild.firstChild.getAttribute("value"));
      // splice in the new data
      this.bookSpec.tocData.splice(index,0,{ point: note.point});
      // update the tocdata position
      for (var i=index+1; i<this.bookSpec.tocData.length; i++) {
         this.bookSpec.tocData[i].point.item.firstChild.firstChild.setAttribute("value",""+i);
      }
      // construct tree item
      var item = this.constructNavigationPoint(note.point,""+index,"note");

      context.item.parentNode.insertBefore(item,context.item);
      note.point.item = item;
   }
   this.updateNavigationPosition();
}

Sidebar.prototype.updateContext = function(scrollY,expand) {
   Console.debug("Sidebar.updateContext:");
   if (this.tocSelect || !this.bookSpec || this.bookSpec.toc.view.rowCount==0) {
      return;
   }
   var pos = 0;
   var testPos = scrollY+Sidebar.positionOffset;
   for (; pos<this.bookSpec.tocData.length; pos++) {
      if (this.bookSpec.tocData[pos].scrollY>testPos) {
         break;
      }
   }
   pos--;
   Console.debug("TOC position: "+pos+" scrollY="+testPos);
   if (pos<0) {
      pos = 0;
   }
   if (this.bookSpec.toc.view.rowCount==0) {
      return;
   }
   var index = pos>=this.bookSpec.toc.view.rowCount ? this.bookSpec.toc.view.rowCount-1 : pos;
   var rowPos = parseInt(this.bookSpec.toc.view.getCellValue(index,this.bookSpec.toc.columns.getFirstColumn()));
   while (rowPos>pos && index>0) {
      index--;
      rowPos = parseInt(this.bookSpec.toc.view.getCellValue(index,this.bookSpec.toc.columns.getFirstColumn()));
   }
   Console.debug("Seeked to: "+rowPos);

   while (expand && rowPos!=pos) {
      if (this.bookSpec.toc.view.isContainerOpen(index)) {
         // we can't find the item
         Console.debug("Stopping at item row "+rowPos+" as we've found an open ancestor.");
         break;
      } else {
         Console.debug("index: "+index);
         this.bookSpec.toc.view.toggleOpenState(index);
         // seek forward in the tree to find all items before the position that are now open
         while (rowPos<pos && index<this.bookSpec.toc.view.rowCount) {
            index++;
            if (index<this.bookSpec.toc.view.rowCount) {
               rowPos = parseInt(this.bookSpec.toc.view.getCellValue(index,this.bookSpec.toc.columns.getFirstColumn()));
            }
         }
         if (rowPos!=pos) {
            // go back one if we haven't found the target
            index--;
            rowPos = parseInt(this.bookSpec.toc.view.getCellValue(index,this.bookSpec.toc.columns.getFirstColumn()));
         }
      }
   }
   this.tocSelectFromScroll = true;
   this.bookSpec.toc.view.selection.select(index);
   this.bookSpec.toc.treeBoxObject.ensureRowIsVisible(index);
   if (!this.bookSpec.context.playContext) {
      this.bookSpec.context.currentPoint = this.bookSpec.tocData[pos].point;
   }
   this.tocSelectFromScroll = false;

}

Sidebar.prototype.updateNavigationPosition = function() {
   if (!this.bookSpec) {
      return;
   }
   var missing = [];
   Console.debug("Updating navigation positions from document.");
   for (var i=0; i<this.bookSpec.tocData.length; i++) {
      var point = this.bookSpec.tocData[i].point;
      var target = this.bookSpec.contentDocument.getElementById(point.content.id);
      if (target) {
         this.bookSpec.tocData[i].scrollY = target.offsetTop;
         //Console.debug(point.content.id+" -> "+this.bookSpec.tocData[i].scrollY);
      } else {
         missing.push(point.content.id);
      }
   }
   if (missing.length>0) {
      Console.log("Cannot find navigation point(s): "+missing);
   }
}

Sidebar.prototype._newTab = function(url) {
  var tabBrowser = this._windowMediator.getMostRecentWindow("navigator:browser").getBrowser();
  var tab = tabBrowser.addTab(url);
  tabBrowser.selectedTab = tab;
  tab.linkedBrowser.contentWindow.focus();
  return tab;
}

Sidebar.prototype.resize = function() {
   //Console.log("resize()");
   if (this.bookSpec && this.bookSpec.contentWindow.daisyRendered) {
      this.updateNavigationPosition();
   }
}

Sidebar.prototype.onOpenBookshelf = function() {
   var bookshelfWindow = window.openDialog("chrome://daisy/content/bookshelf.xul",'daisy-bookshelf',"chrome,resizable",{});
   bookshelfWindow.focus();
}

Sidebar.prototype.onFocusTab = function(index) {
   var tabbox = document.getElementById("toc-tabs");
   for (var i=0; i<tabbox.tabs.itemCount; i++) {
      var tab = tabbox.tabs.getItemAtIndex(i);
      if (i==index) {
         tab.setAttribute("tabindex","1");
      } else {
         tab.removeAttribute("tabindex");
      }
   }
}


Sidebar.prototype.onSettings = function() {
   var current = this;
   var data = {
      debug: current.service.debug,
      showImages: current.service.showImages,
      focusFollow: current.service.focusFollow,
      highlight: current.service.highlight,
      selectText: current.service.smil.selectHighlight,
      sound: current.service.sound,
      generateSMIL: current.generateSMIL,
      highlightTime: current.service.textDuration/1000,
      playbackPause: current.service.playbackPause,
      quirks: current.service.allowQuirks,
      readContent: current.service.readContent,
      readUI: current.service.readUI,
      forceReadContent: current.service.forceReadContent,
      navigateByStructure: current.service.navigateByStructure,
      showSoundContextWindow: current.showSoundContextWindow,
      sidebarPause: current.service.sidebarPause,
      readNotesInline: current.service.readNotesInline,
      cueSidebarOpen: current.service.cueSidebarOpen,
      speechRate: {
         preferredRate: current.service.speechRate,
         hasTTS: current.service._tts ? true : false,
         unit: current.service._tts ? current.service._tts.rateUnit : null,
         rate: current.service._tts ? current.service._tts.rate : null,
         minimum: current.service._tts ? current.service._tts.minimumRate : null,
         normal: current.service._tts ? current.service._tts.normalRate : null,
         maximum: current.service._tts ? current.service._tts.maximumRate : null
      },
      onUpdate: function() {
         current.service.showImages = this.showImages;
         current.service.focusFollow = this.focusFollow;
         current.service.highlight = this.highlight;
         current.service.smil.selectHighlight = this.selectText;
         current.service.sound = this.sound;
         current.generateSMIL = this.generateSMIL;
         current.service.textDuration = this.highlightTime*1000;
         current.service.playbackPause = this.playbackPause;
         current.service.allowQuirks = this.quirks;
         current.service.readContent = this.readContent;
         current.service.readUI = this.readUI;
         current.service.forceReadContent = this.forceReadContent;
         current.service.navigateByStructure = this.navigateByStructure;
         current.service.sidebarPause = this.sidebarPause;
         current.service.readNotesInline = this.readNotesInline;
         current.service.cueSidebarOpen = this.cueSidebarOpen;
         current.showSoundContextWindow = this.showSoundContextWindow;
         current.service.speechRate = this.speechRate.preferredRate;
         /*
         if (current.showSoundContextWindow) {
            current.soundContext.style.height = "300px";
         } else {
            current.soundContext.style.height = "0px";
         }
         */
      }
   };

   this.settingsWindow = window.openDialog("chrome://daisy/content/settings.xul",'daisy-settings',"chrome,resizable",data);
}

Sidebar.prototype.browseForFile = function() {
   this.service.tts.cancel();

   var fp = null;
   var nsIFilePicker = Components.interfaces.nsIFilePicker;
   if (this.service.readUI) {
      fp = new FilePicker();
      fp.title = "Choose Book (zip or opf)";
   } else {
      fp = Components.classes["@mozilla.org/filepicker;1"].createInstance(nsIFilePicker);
      fp.init(window, "Choose Book (zip or opf)", nsIFilePicker.modeOpen);
   }
   fp.appendFilter("DAISY Book","*.zip;*.bks2;*.opf");
   if(fp.show() == nsIFilePicker.returnOK) {
      var opfTest = /\.opf$/;
      var zipTest = /\.zip$/;
      var bks2Test = /\.bks2$/;
      if (fp.file.leafName.match(opfTest)) {
         var url = Sidebar.fileURL(fp.file);
         this.openDocument(url.spec);
      } else if (fp.file.leafName.match(zipTest) || fp.file.leafName.match(bks2Test)) {
         var bookDir = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
         bookDir.initWithPath(fp.file.parent.path);
         var pos = fp.file.leafName.lastIndexOf(".zip");
         if (pos<0) {
            pos = fp.file.leafName.lastIndexOf(".bks2");
         }
         var dirName = fp.file.leafName.substring(0,pos)
         bookDir.appendRelativePath(dirName);
         if (bookDir.exists()) {
            var opfFile = this.service.findPackagingInDirectory(bookDir);
            var url = Sidebar.fileURL(opfFile);
            this.openDocument(url.spec);
         } else {
            var result = this.service.readZipPackaging(fp.file);
            if (result.document) {
               var book = this.service.newBook();
               book.init();
               book.loadPackaging(result.document,result.baseURI);
               book.zipFile = fp.file;
               this.openBook(book);
            } else {
               alert("This zip file does not seem to contain a DAISY book.");
            }

         }
      }
   }
}

Sidebar.prototype.onHelp = function() {
   var url = Sidebar.fileURL(this.helpOPF);
   this.openDocument(url.spec);
}

Sidebar.prototype.onBookshare = function() {
   this.openDocument("http://www.bookshare.org/",true);
}


Sidebar.prototype.onOpenCue = function(oncomplete) {
   if (!this.service.cueSidebarOpen) {
      return;
   }
   var extManager = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
   var extid = "{6be5ef88-4289-44d1-81f2-097313ed640b}"
   var autoPauseSound =  Sidebar.fileURL(extManager.getInstallLocation(extid).getItemFile(extid,"content/loaded.mp3")).spec;
   var soundContext = this.mainWindow.document.getElementById("daisy-sound-context");
   var soundWindow = soundContext.contentWindow;
   var current = this;
   var soundCount = 0;
   var tryCount = 0;
   var waitForSoundManager = function() {
      if (soundWindow.soundManager && soundWindow.soundManager._didInit && !current.inSoundManagerInit) {
         if (soundWindow.soundManager.supported()) {
            var sound = current.soundManager.createSound({
               id: 'sidebar-open-cue',
               url: autoPauseSound,
               autoPlay: false,
               onfinish: function() {
                  if (oncomplete) {
                     oncomplete();
                  }
               }

            });
            sound.load();
            var waitFor = function() {
               if (sound.readyState==3) {
                  sound.play();
               } else {
                  setTimeout(waitFor,10);
               }
            }
            waitFor();
         } else if (tryCount<2) {
            Console.log("soundManager didn't load, retrying...");
            tryCount++;
            soundCount = 0;
            current.initSoundManager(true);
            setTimeout(waitForSoundManager,100);
         } else {
            Console.log("soundManager is not supported.");
            current.skipSoundManager = true;
            current.writeContent();
         }
      } else if (soundCount<30) {
         soundCount++;
         setTimeout(waitForSoundManager,100);
      } else {
         Console.log("No soundManager instance could be found in sound context.");
         current.skipSoundManager = true;
         current.writeContent();
      }
   };
   waitForSoundManager();
}

Sidebar.prototype.initSoundManager = function(force) {

   /**
    * The sound manager is loaded into a hidden browser instance located in
    * the status bar.  On windows, this browser instance must be partially
    * visable for the flash to load properly.
    *
    * Once the document is loaded for the browser window and initialized, the
    * same sound manager instance is used for all books.  This keeps books
    * from playing audio over each other.
    */
   var extManager = Components.classes["@mozilla.org/extensions/manager;1"].getService(Components.interfaces.nsIExtensionManager);
   var id = "{6be5ef88-4289-44d1-81f2-097313ed640b}";
   var soundFile = extManager.getInstallLocation(id).getItemFile(id,"content/sound.xhtml");
   var soundContext = this.mainWindow.document.getElementById("daisy-sound-context");
   var soundDocument = Sidebar.fileURL(soundFile).spec;
   if (soundContext.contentWindow.location.href!=soundDocument || force) {
      if (this.OS=="WINNT" && soundContext.boxObject.height==0) {
         soundContext.style.height = "10px";
      }
      this.inSoundManagerInit = true;
      Console.debug("Loading soundManager2");
      /**
       * flash block is evil, so we have to disable it and restore it
       * after we're done.
       */
      FlashBlockHack.save();

      /**
       * we track the progress of the document loading via a WebProgressListener
       * instance.
       */
      var current = this;
      var myListener = {
         QueryInterface: function(aIID) {
            if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
               aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
               aIID.equals(Components.interfaces.nsISupports)) {
               return this;
            }
            throw Components.results.NS_NOINTERFACE;
         },

         onStateChange: function(aWebProgress, aRequest, aFlag, aStatus) {
            if(aFlag & Components.interfaces.nsIWebProgressListener.STATE_STOP && aFlag & Components.interfaces.nsIWebProgressListener.STATE_IS_DOCUMENT && aStatus==0) {
                // This fires when the load finishes

               /**
                * The document is now loaded.  We now need to wait for the
                * sound manger to initialize.  Once that is complete, we're done.
                */
               soundContext.removeProgressListener(myListener);
               var tryCount = 0;
               var waitForInit = function() {
                  if (soundContext.contentWindow.soundManager && soundContext.contentWindow.soundManager._didInit) {
                     current.inSoundManagerInit = false;
                     current.soundManager = soundContext.contentWindow.soundManager;
                     FlashBlockHack.restore();
                     soundContext.style.height = "0px";
                  } else if (tryCount<30) {
                     tryCount++;
                     setTimeout(waitForInit,100);
                  } else {
                     current.inSoundManagerInit = false;
                     FlashBlockHack.restore();
                     Console.log("Sound manager did not initialize.");
                     soundContext.style.height = "0px";
                  }
               }
               waitForInit();
            }
            return 0;
         },

         onLocationChange: function(aProgress, aRequest, aURI) { return 0; },
         onProgressChange: function() {return 0;},
         onStatusChange: function() {return 0;},
         onSecurityChange: function() {return 0;},
         onLinkIconAvailable: function() {return 0;}
      };
      soundContext.addProgressListener(myListener,Components.interfaces.nsIWebProgress.NOTIFY_STATE_DOCUMENT);
      soundContext.loadURI(soundDocument,null,null);
   }

}

/**
 * Flash block enable/disable code
 */

var FlashBlockHack = {
  prefs: Components.classes["@mozilla.org/preferences-service;1"]
         .getService(Components.interfaces.nsIPrefBranch)
         .QueryInterface(Components.interfaces.nsIPrefBranchInternal),
  valueSet: false,
  value: false,

  // Returns the value of the flashblock.enabled pref
  save: function() {
     if (this.prefs.prefHasUserValue("flashblock.enabled")) {
        this.value = this.prefs.getBoolPref("flashblock.enabled");
        this.valueSet = true;
        Console.debug("User value: "+this.value);
     } else {
        this.valueSet = false;
        Console.debug("flashblock.enabled is not set");
     }
     this.prefs.setBoolPref("flashblock.enabled", false);
  },


  // Sets the flashblock.enabled pref to the given boolean value
  restore: function() {
     Console.debug("Restoring flashblock.enabled");
     try {
        if (this.valueSet) {
           this.prefs.setBoolPref("flashblock.enabled", this.value);
        } else {
           this.prefs.clearUserPref("flashblock.enabled");
        }
     } catch (ex) {
     }
  }

}




var theSidebar = Sidebar.getInstance();



