
function UIVoice() {
   this.voiceable = [ "button", "checkbox", "toolbarbutton", "textbox", "scale", "listbox" ];
   this.labels = {};
   this.voiced = {};
   this.disabled = false;
}

UIVoice.codePointNames = [
   { start: 9, map: ["tab", "line feed"]},
   { start: 13, map: ["carriage return"]},
   { start: 32, map: [ "space", "exclamation", "double quote", "hash", "dollar", "percent", "ampersand", "single quote", "left paren", "right parent",
                       "asterik", "plus", "comma", "minus", "period", "slash"]},
   { start: 58, map: [ "colon", "semicolon", "less than", "equals", "greater than", "question mark", "at sign"]},
   { start: 91, map: [ "left square bracket", "backslash", "right square bracket", "circumflex", "underscore", "grave",]},
   { start: 123, map: [ "left curly bracket", "veritcal line", "right curly bracket", "tilde"]}

];

UIVoice.prototype.init = function() {
   var theClass = Components.classes["@benetech.org/daisy;1"];
   var theComponent = theClass.getService(Components.interfaces.nsISupports);
   this.service = theComponent.wrappedJSObject;
}

UIVoice.prototype.makeCallback = function(operation) {
   var tm = Components.classes["@mozilla.org/thread-manager;1"].getService(Components.interfaces.nsIThreadManager);
   var ttsCallback = {
      thread: tm.currentThread,
      QueryInterface: function(aIID) {
         if (aIID.equals(Components.interfaces.nsITTSCallback) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports)) {
            return this;
         }
         throw Components.results.NS_NOINTERFACE;
      },
      onFinish: function() {
         this.thread.dispatch({
            run: function() {
               operation();
            }
         },this.thread.DISPATCH_NORMAL);
      }
   };
   return ttsCallback;
}

UIVoice.prototype.speak = function(text,callback) {
   if (this.service.tts && this.service.readUI && !this.disabled) {
      this.cancelling = false;
      var lock = this.service.getLockFor("tts");
      var current = this;
      this.ownedLock = lock.acquire({
         exclusive: true,
         onRelinquish: function() {
            current.cancelling = true;
            current.cancel();
         }
      });
      this.service.tts.speak(text,this.makeCallback(function() {
         if (!current.cancelling && callback) {
            current.ownedLock.release();
            callback.onFinish();
         }
      }));
   }
}

UIVoice.prototype.cancel = function() {
   if (this.service.tts && this.service.readUI) {
      this.service.tts.cancel();
      if (this.ownedLock) {
         this.ownedLock.release();
         this.ownedLock = null;
      }
   }
}


UIVoice.prototype.attach = function(parent) {
   var uivoice = this;
   var current = parent.firstChild;
   while (current!=parent) {
      if (current.nodeType==1) {
         var isVoiceable = false;
         for (var i=0; !isVoiceable && i<this.voiceable.length; i++) {
            if (current.localName==this.voiceable[i]) {
               isVoiceable = true;
            }
         }
         if (isVoiceable) {
            current.onfocus = function(event) {
               uivoice.voice(this);
            }
            if (current.localName=="checkbox") {
               current.onkeypress = function(event) {
                  this.lastCharCode = event.charCode;
               }
               current.onkeyup = function(event) {
                  if (this.lastCharCode==32) {
                     uivoice.voiceCheckboxState(this);
                  }
               }
            } else if (current.localName=="scale") {
               current.onkeypress = function(event) {
                  if (event.keyCode==39) {
                     uivoice.voiceScaleChange(true,this.value);
                  } else if (event.keyCode==37) {
                     uivoice.voiceScaleChange(false,this.value);
                  }
               }
            } else if (current.localName=="listbox") {
               current.onkeypress = function(event) {
                  this.lastKeyCode = event.keyCode;
               }
               current.onkeyup = function(event) {
                  switch (this.lastKeyCode) {
                     case 37: // left
                     case 38: // up
                     case 39: // right
                     case 40: // down
                        if (this.selectedItem) {
                           uivoice.speak("item, "+this.selectedItem.label,null);
                        }
                        break;
                  }
               };
            } else if (current.localName=="textbox" && current.getAttribute("type")!="password") {
               current.onkeypress = function(event) {
                  if (!event.shiftKey) {
                     this.selecting = 0;
                  }
                  //Console.log("pos: "+this.selectionStart+", "+this.selectionEnd+", length="+this.value.length);
                  //Console.log("key: "+event.keyCode+", char:"+event.charCode);
                  if ((event.metaKey && event.charCode==122) || (event.ctrlKey && event.charCode==122)) {
                     uivoice.speak("undo",null);
                  } else if (event.ctrlKey && event.altKey && event.charCode==115) {
                     uivoice.speak(this.value,null);
                  } else if (!event.altKey && !event.ctrlKey && !event.metaKey && event.charCode>0) {
                     var text = uivoice.getCharacterName(String.fromCharCode(event.charCode))
                     uivoice.speak(text,null);
                  } else if (!event.altKey && !event.ctrlKey && !event.metaKey && event.keyCode>0) {
                     if (event.keyCode==13 || event.keyCode==10) {
                        // carriage returns get turned into line feeds
                        uivoice.speak("line feed",null);
                     } else if (event.keyCode==8 && this.value.length>0) {
                        if (this.selectionStart==this.selectionEnd) {
                           var text = uivoice.getCharacterName(this.value.substring(this.selectionStart-1,this.selectionStart));
                           uivoice.speak(text,null);
                        } else {
                           uivoice.speak("Selection deleted",null);
                        }
                     } else if (event.shiftKey && event.keyCode==37) {
                        // left select
                        if (this.selectionStart==this.selectionEnd) {
                           this.selecting = -1;
                        }
                        if (this.selecting<0) {
                           if (this.selectionStart>0) {
                              var text = uivoice.getCharacterName(this.value.substring(this.selectionStart-1,this.selectionStart));
                              uivoice.speak(text,null);
                           }
                        } else {
                           var text = uivoice.getCharacterName(this.value.substring(this.selectionEnd-1,this.selectionEnd));
                           uivoice.speak(text,null);
                        }
                     } else if (event.shiftKey && event.keyCode==39) {
                        // right select
                        if (this.selectionStart==this.selectionEnd) {
                           this.selecting = 1;
                        }
                        if (this.selecting<0) {
                           if (this.selectionStart>0) {
                              var text = uivoice.getCharacterName(this.value.substring(this.selectionStart,this.selectionStart+1));
                              uivoice.speak(text,null);
                           }
                        } else {
                           var text = uivoice.getCharacterName(this.value.substring(this.selectionEnd,this.selectionEnd+1));
                           uivoice.speak(text,null);
                        }
                     } else if (event.keyCode==37) {
                        if (this.selectionStart>0) {
                           var text = uivoice.getCharacterName(this.value.substring(this.selectionStart-1,this.selectionStart));
                           uivoice.speak(text,null);
                        }
                     } else if (event.keyCode==39) {
                        // right select
                        if (this.selectionStart<this.value.length) {
                           var text = uivoice.getCharacterName(this.value.substring(this.selectionStart,this.selectionStart+1));
                           uivoice.speak(text,null);
                        }
                     }

                  }
               }
            } else if (current.localName=="textbox") {
               // a password entry
               current.onkeypress = function(event) {
                  if (!event.shiftKey) {
                     this.selecting = 0;
                  }
                  if (event.metaKey && event.charCode==122) {
                     uivoice.speak("undo",null);
                  } else if (!event.altKey && !event.ctrlKey && !event.metaKey && event.charCode>0) {
                     uivoice.speak("star",null);
                  } else if (!event.altKey && !event.ctrlKey && !event.metaKey && event.keyCode>0) {
                     if (event.keyCode==8 && this.value.length>0) {
                        if (this.selectionStart==this.selectionEnd) {
                           uivoice.speak("delete",null);
                        } else {
                           uivoice.speak("Selection deleted",null);
                        }
                     }
                  }
               }
            }
         } else if (current.firstChild) {
            current = current.firstChild;
            continue;
         }
      }

      if (current.nextSibling) {
         current = current.nextSibling;
      } else {
         while (!current.nextSibling && current!=parent) {
            current = current.parentNode;
         }
         if (current!=parent) {
            current = current.nextSibling;
         }
      }
   }
}

UIVoice.prototype.voice = function(element) {
   var text = element.getAttribute("tooltiptext");
   if (!text || text=="") {
      text = element.getAttribute("label");
   }
   if (element.localName=="button" || element.localName=="toolbarbutton") {
      text = "Button, "+text;
   } else if (element.localName=="checkbox") {
      text = "Checkbox, "+text+" , "+(element.checked ? "checked" : "not checked");
   } else if (element.localName=="scale") {
      text = "Slider, "+text+" , value "+element.value;
   } else if (element.localName=="textbox") {
      text = "Text box, "+text+" , value "+element.value;
   } else if (element.localName=="listbox") {
      text = "list box, "+text;
      if (element.selectedItem) {
         text += element.selectedItem.label;
      }
   }
   this.speak(text,null);
}

UIVoice.prototype.voiceCheckboxState = function(element) {
   this.speak(element.checked ? "checked" : "not checked",null);
}

UIVoice.prototype.voiceScaleChange = function(increment,value) {
   this.speak((increment ? "increment to " : "decrement to ")+"value "+value,null);
}

UIVoice.prototype.getCharacterName = function(s) {
   var c = null;
   if (s.charCodeAt) {
      c = s.charCodeAt(0)
   } else {
      c = s;
   }
   for (var i=0; i<UIVoice.codePointNames.length; i++) {
      if (c<UIVoice.codePointNames[i].start) {
         return s;
      } else if (c<(UIVoice.codePointNames[i].start+UIVoice.codePointNames[i].map.length)) {
         return UIVoice.codePointNames[i].map[c-UIVoice.codePointNames[i].start];
      }
   }
   return s;
}